/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): Marc Dutoo, 
 * JOnAS 2.4 Murad Meghani (Murad.Meghani@compuware.com) killServer and stopServer
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.adm;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import javax.naming.InitialContext;
import javax.naming.Context;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NameClassPair;
import org.objectweb.common.RemoteObject;
import org.objectweb.common.Env;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.container.EJBServiceImpl;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.jtm.TransactionServiceImpl;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.registry.RegistryManager;
import org.objectweb.jonas_ejb.container.JContainer;
import org.objectweb.jonas.naming.NamingManager;

/**
 * This class implements a remote interface used for administering
 *      the server.
 */
public class Adm extends RemoteObject implements AdmInterface {

    static final int trace = Trace.DB_18;

    public static final String ADMNAME_SUFFIX = "_Adm";

    private ServiceManager sm = null;
    private EJBServiceImpl ejbserv  = null;
    private TransactionServiceImpl tm = null;
    private JProp jonasProperties = null;
    private InitialContext ictx = null;
    private boolean isNotEJB = false;   // = true if TMServer or Tomcat standalone

    /**
     * Adm constructor
     */
    public Adm(JProp jp, boolean isNotEJB) throws RemoteException, NamingException, ServiceException {
	Trace.outln(trace, "Adm.Adm() ");
	jonasProperties = jp;
	this.isNotEJB = isNotEJB;

	// Get Service references
	sm = ServiceManager.getInstance();
	if (isNotEJB == false) {
	    ejbserv = (EJBServiceImpl) sm.getEjbService();
	}
        try{
            tm = (TransactionServiceImpl) sm.getTransactionService();
        }catch(ServiceException e){
        }

	// Get an Initial Context and
	// registers the JOnAS Server Name to JNDI
	String name = null;
        name = jonasProperties.getValue("jonas.name", "jonas") + ADMNAME_SUFFIX;
	try {
	    if (Env.isDavid()) 
		ictx = NamingManager.getInstance().getCorbaInitialContext();
	    else
		ictx = NamingManager.getInstance().getInitialContext();

	    ictx.bind(name, this);
	    Trace.outln(trace, "rebind done " + name);
        } catch (NameAlreadyBoundException e) {
            Trace.errln("Cannot register  "+name+" already bound ");
	    throw e;
	} catch (NamingException e) {
	    Trace.errln("Cannot register "+name+" : " + e);
	    throw e;
	}
    }

    // ------------------------------------------------------------------
    //  AdmInterface implementation
    // ------------------------------------------------------------------

    /**
     * configure trace messages in EJB Server
     */
    public void setTraceLevel(String level) throws RemoteException {
	Trace.outln(trace, "Adm.setTraceLevel");
	Trace.setFilters(level);	
    }

    /**
     * configure trace output device
     */
    public void setTraceOutput(String out) throws RemoteException {
	Trace.outln(trace, "Adm.setTraceOutput");
	Trace.setOutput(out);
    }

    /**
     * configure trace header
     */
    public void setTraceHeader(boolean h) throws RemoteException {
	Trace.outln(trace, "Adm.setTraceHeader");
	Trace.setHeader(h);
    }

    /**
     * Create a container and load beans in it
     * @parameter fileName name of the ejb-jar or xml file
     */
    public void addBeans(String fileName) throws RemoteException {
	if (isNotEJB) return;
	Trace.outln(trace, "Adm.addBeans");
	ejbserv.createContainer(fileName);
    }

    /**
     * Remove the container identified by fileName and remove all beans in it
     * @parameter fileName name of the ejb-jar or xml file 
     */
    public void removeBeans(String fileName) throws RemoteException {
	if (isNotEJB) return;
	Trace.outln(trace, "Adm.removeBeans ", fileName);
	ejbserv.removeContainer(fileName);
    }

    /**
     * returns true if beans are already loaded in server.
     * @parameter fileName name of the ejb-jar or xml file
     */
    public boolean isLoaded(String fileName) throws RemoteException {
	Trace.outln(trace, "Adm.isLoaded");
	return (ejbserv.getContainer(fileName) != null);
    }

    /**
     * List beans of all JOnAS containers
     */
    public String [] listBeans() throws RemoteException {

	if (isNotEJB) return new String[0];

	Trace.outln(trace, "Adm.listBeans");
	JContainer[] lcont = ejbserv.listContainers();
	Vector lbeans = new Vector();
	for( int j = 0; j<lcont.length; j++) {
	    String[] lbn = lcont[j].listBeanNames();
	    String contname = lcont[j].getName();
	    for( int k = 0; k<lbn.length; k++){
		String s = new String(contname + ": " + lbn[k]);
		lbeans.addElement(s);
	    } 
	}
	String[] ret = new String[lbeans.size()];
	lbeans.copyInto(ret);

	return ret;
    }

    /**
     * List JNDI context
     */
    public Vector listContext() throws RemoteException {
	Trace.outln(trace, "Adm.listContext()");
	String name = null;
	Vector ret  = new Vector();
	try {
	    NamingEnumeration ne;

	    if (Env.isDavid()) {
		// list CosNaming and rmiregistry
		ne = NamingManager.getInstance().getCorbaInitialContext().list("");
		for (Enumeration e = ne; e.hasMoreElements(); ) {
		    name = ((NameClassPair) e.nextElement()).getName();
		    Trace.outln(trace, "    name found in context = "+ name);
		ret.addElement(name);
		}
	    }

	    ne = NamingManager.getInstance().getInitialContext().list("");
	    for (Enumeration e = ne; e.hasMoreElements(); ) {
		name = ((NameClassPair) e.nextElement()).getName();
		Trace.outln(trace, "    name found in context = "+ name);
		ret.addElement(name);
	    }
	} catch (Exception ex) {
	    throw new RemoteException("Cannot list JNDI context", ex);
	}
	return ret;
    }

    /**
     * List Environment
     */
    public Properties listEnv() {
	Trace.outln(trace, "Adm.listEnv() ");
	return jonasProperties.getFilesEnv();
    }

    /**
     * Stop the Server without stopping the JVM
     */
    public void stopServer() throws RemoteException {
	Trace.outln(trace, "Adm.stopServer() ");
	try {
	    sm.stopServices();
	} catch (ServiceException e) {
	    throw new RemoteException("Cannot stop services: ", e);
	}

    }
 

    /**
     * Stop  the Server and  stop the JVM
     */
    public void killServer() throws RemoteException {
	Trace.outln(trace, "Adm.killServer() ");
	stopServer();
	System.exit(0);
    }


    /**
     * To test if an Adm is reachable
     */
    public boolean ping() throws RemoteException {
	if (isNotEJB) {
	    Trace.outln(trace, "Adm.ping on StandAloneServer");
	} else {
	    Trace.outln(trace, "Adm.ping on EJBServer");
	}
	return isNotEJB;
    }

    /**
     * set the default value for transaction timeout
     */
    public void setTransactionTimeout(int timeout) throws RemoteException {
	Trace.outln(trace, "Adm.setTransactionTimeout");
        if(tm != null)
            tm.setCurrentTimeOut(new Integer(timeout));
    }

    /**
     * run the garbage collector
     */
    public void runGC() throws RemoteException {
	Trace.outln(trace, "Adm.runGC");	
	Runtime.getRuntime().gc();
    }
    
    /**
     * sync all entity instances outside transactions
     */
    public void syncAllEntities() throws RemoteException {
	if (isNotEJB) return;
	Trace.outln(trace, "Adm.syncAllEntities");
	ejbserv.syncAllEntities();
    }
}

 

