/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


/*----------------------------------------------------*/
/*                  ComboBoxList                      */
/*----------------------------------------------------*/

package org.objectweb.jonas.beandeployer.components;

import org.objectweb.jonas.beandeployer.Consts;
import org.objectweb.jonas.common.Trace;

import java.util.Vector;
import java.util.Enumeration;

import java.awt.GridBagConstraints;
import java.awt.Dimension;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import javax.swing.JComboBox;
import javax.swing.JComponent;

/**
 * Component to display a label with a combo box element.
 * @author Christophe GERMAIN.
 */
public class ComboBoxList extends DeployerComponent implements ActionListener {

    static private final int trace = Trace.DB_19;
    
    /** Specific component of ComboBoxList **/
    public JComboBox nameList;

    /** List of choices in the JComboBox **/
    private Vector list;

    /**
     * Builds the combo box.
     * @param labelName  the name of the label.
     * @param list       the list of all available choices.
     *                   This information is represented in a Vector.
     * @param obligatory true if the filling of the component is obligatory,
     *                   and false if it is optional.
     */
    public ComboBoxList(String labelName, Vector list, boolean obligatory) {
	super(labelName,obligatory);
	this.list = list;
	if (this.list.size() == 0)
	    this.list.add(Consts.EMPTY_STRING);
	init();
    }

    /**
     * Add a label and a list of choices in a combo box.
     */
    private void init(){
	addLabel();
	addComboBox(list);
    }

    /**
     * Add a JComboBox component in the panel.
     * @param list the list of choices in the JComboBox.
     */
    private void addComboBox(Vector list){
	this.nameList = new JComboBox(list);	
	//	nameList.setSelectedIndex(0);
	this.nameList.setPreferredSize(new Dimension(Consts.WITDH_COMBOBOX,
						     Consts.HEIGHT_COMBOBOX));
	this.nameList.setEditable(false);
	this.nameList.setMaximumRowCount(Consts.MAXIMUM_ROW_COUNT);
	this.nameList.addActionListener(this);
	this.add(nameList);
    }

    /**
     * Update the list of choices in the JComboBox.
     * @param newList the new list of choices in the JComboBox.
     */
    public void updateList(Vector newList){
	// tips to not have a line with empty string because of bad display on Solaris and Windows
	if (this.list.indexOf(Consts.DEFAULT_STRING) != -1)
	    this.nameList.removeItem(Consts.DEFAULT_STRING);

	for (Enumeration enum = newList.elements() ; enum.hasMoreElements() ;) {
	    String temp = (String)enum.nextElement();
	    if ( this.list.indexOf(temp) == -1 && temp != null)
		this.nameList.addItem(temp);
	}
 
    }

    /**
     * Obligatory method to use ActionListener.
     * @param e is the ActionEvent.
     */
    public void actionPerformed(ActionEvent e) {}

    /**
     * Get the specific JComponent, here it is JComboBox.
     * @return the specific component.
     */
    JComponent getSpecificJComponent() {
	return (JComponent)this.nameList;
    }
 
    /**
     * Set the Selected item of the combo box.
     * @param text is the new text of the component.
     */   
    public void setText(String text) {
	if ( list.indexOf(text) == -1 )
	    list.add(text);
	nameList.setSelectedIndex(list.indexOf(text));
    }

    /**
     * Get the Selected item of the combo box.
     * @return the value of the selected item of the JComboBox.
     */
    public String getText() {
	return (String)this.nameList.getSelectedItem();
    }

    /**
     * Set the name's list editable or not.
     * @param aFlag is true if the name's list must be editable.
     */
    public void setEditable(boolean aFlag){
	this.nameList.setEditable(aFlag);
    }

    /**
     * Print the selected item.
     * Method used only for testing the component.
     */
    public void print(){
	String temp =  this.nameList.getSelectedItem().toString();
	Trace.outln(trace, "ComboBoxList.print(): Selected item is "+temp);
    }
}
