/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): Xavier Spengler
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.common;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;
import java.util.Properties;
import java.util.StringTokenizer;
import org.objectweb.common.TraceCore;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_jms.TraceJms;
import org.objectweb.jonas_timer.TraceTimer;
import org.objectweb.jonas_tm.TraceTm;

/**
 *  This class implements the tracing facilities.
 */
public class Trace {
    public static final int DB_0 = (1<<0); 	// 
    public static final int DB_1 = (1<<1);  	// tests ejb
    public static final int DB_2 = (1<<2); 	// env
    public static final int DB_3 = (1<<3); 	// naming
    public static final int DB_4 = (1<<4);   	// jtm, timers
    public static final int DB_5 = (1<<5); 	// jtm -jta
    public static final int DB_6 = (1<<6); 	// implicit propagation of context
    public static final int DB_7 = (1<<7); 	// threads
    public static final int DB_8 = (1<<8); 	// connection manager
    public static final int DB_9 = (1<<9); 	// jdbc,ldap
    public static final int DB_10 = (1<<10);   	// jdbc_xa
    public static final int DB_11 = (1<<11); 	// trace connection pool (very verbose)
    public static final int DB_12 = (1<<12);    // container (Context)
    public static final int DB_13 = (1<<13); 	// xa resource
    public static final int DB_14 = (1<<14);   	// Container, Server
    public static final int DB_15 = (1<<15); 	// Generated classes
    public static final int DB_16 = (1<<16); 	// container
    public static final int DB_17 = (1<<17); 	// container (preinvoke/postinvoke)
    public static final int DB_18 = (1<<18); 	// 
    public static final int DB_19 = (1<<19); 	// tools
    public static final int DB_20 = (1<<20);  	// deployment
    public static final int DB_21 = (1<<21); 	// container (pool entitycontext)
    public static final int DB_22 = (1<<22); 	// jms
    public static final int DB_23 = (1<<23); 	// tests ejb (old tests)
    public static final int DB_24 = (1<<24); 	//
    public static final int DB_25 = (1<<25);  	// security context
    public static final int DB_26 = (1<<26);   	// jmx
    public static final int DB_27 = (1<<27);   	// 
    public static final int DB_28 = (1<<28);   	// resource adapter
    public static final int DB_29 = (1<<29);   	//   
    public static final int DB_30 = (1<<30); 	// 
    public static final int DB_31 = (1<<31);  	// Warnings
    public static final int DB_NONE = 0x0000;    
    public static final int DB_JTM = DB_4|DB_5|DB_6|DB_31;
    public static final int DB_EJB = DB_12|DB_14|DB_15|DB_16|DB_17|DB_21|DB_31;
    public static final int DB_JDBC = DB_8|DB_9|DB_10|DB_13|DB_31;
    public static final int DB_JONAS = DB_7|DB_JTM|DB_EJB|DB_JDBC;
    public static final int DB_ALL = 0xFFFFFFFF;
  
    private static LogWriter logwriter = new LogWriter(System.out);
    private static boolean header = false;
	
    // tempo before error message output
    private static final int TEMPO_DEFAULT = 1000;
    private static int mTempo = 0; // milliseconds
  
    private static int mLevel = 0; 
    private static String mPrefixString = "";

    // for jmx
    private static String traceOut = "out";

    // constructor is private to prevent client program from instanciating a 'Trace'
    private Trace() { };


    /**
     * Configure trace from trace.properties
     */
    public static void configure() {

	// reads trace.properties file
	Properties traceprop = new Properties();
	try {
	    traceprop.load(ClassLoader.getSystemResourceAsStream("trace.properties"));
	} catch (IOException e) {
	    System.err.println("Cannot read trace.properties: take default values");
	} catch (NullPointerException e) {
	    System.err.println("No trace.properties: take default values");
	}

	// gets the properties from file
	String out = traceprop.getProperty("trace.out", "out");
	String level = traceprop.getProperty("trace.level", "31");
	String header = traceprop.getProperty("trace.header", "false");
	String tempo = traceprop.getProperty("trace.tempo", "false");

	// configure trace
	boolean bHeader = header.equalsIgnoreCase("true");
	if (tempo.equalsIgnoreCase("true")) {
	    mTempo = TEMPO_DEFAULT;
	}
	configure(out, levelValue(level), bHeader);
    }


    /**
     * To initialize trace from parameters 
     * @deprecated
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param err    error device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in string format
     */
    public static void configure(String out, String err, String level, boolean withHeader) {
	configure(out, levelValue(level), withHeader);
    }

    /**
     * To initialize trace from parameters
     * @deprecated
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in int format
     * @param header with header if set to true, without header if set to false.
     */
    public static void configure(String out, String err, int level, boolean withHeader) {
	configure(out, level, withHeader);
    }

    /**
     * To initialize trace from parameters 
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param err    error device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in string format
     */
    public static void configure(String out, String level, boolean withHeader) {
	configure(out, levelValue(level), withHeader);
    }

    /**
     * To initialize trace from parameters
     * @param out    output device  "out": standard output; "err": standard error; other: filename
     * @param level  trace level in int format
     * @param header with header if set to true, without header if set to false.
     */
    public static void configure(String out, int level, boolean withHeader) {

	traceOut = out;
	header = withHeader;
	mLevel = level;
	configureOutput(out);
	logwriter.setHeader(header);
	logwriter.setMTempo(mTempo);
	configureServices();
    }

    /**
     * Sets the level only (for Adm)
     */
    public static void setFilters(String level) {
	outln("Setting trace filter to "+level);
	mLevel = levelValue(level);
	configureServices();
    }

    /**
     * Sets the output device for Trace (for Adm)
     */
    public static void setOutput(String out) {
	outln("Setting trace output to "+out);
	configureOutput(out);
	configureServices();
    }

    /**
     * Set/unset the header for Trace (for Adm)
     */
    public static void setHeader(boolean h) {
	outln("Setting trace header to "+h);
	header = h;
	logwriter.setHeader(header);
	configureServices();
    }


    /**
     * Sets the trace prefix
     */
    public static void setPrefix(String prefix) {
	mPrefixString = prefix+" ";
	logwriter.setPrefix(mPrefixString);
    }

    /**
     * Returns the integer value for a string trace level
     */
    public static int levelValue(String level) {
	int ret = 0;
	StringTokenizer tok = new StringTokenizer(level, ",.-");

	while (tok.hasMoreTokens()) {
	    String token = tok.nextToken().trim();
	    if (token.equalsIgnoreCase("all")) {
		return DB_ALL;
	    }
	    if (token.equalsIgnoreCase("none")) {
		return DB_NONE;
	    }
	    if (token.equalsIgnoreCase("jtm")) {
		ret = ret + DB_JTM;
		continue;
	    }
	    if (token.equalsIgnoreCase("ejb")) {
		ret = ret + DB_EJB;
		continue;
	    }
	    if (token.equalsIgnoreCase("jdbc")) {
		ret = ret + DB_JDBC;
		continue;
	    }
	    if (token.equalsIgnoreCase("jonas")) {
		ret = ret + DB_JONAS;
		continue;
	    }
	    Integer it = new Integer(token);
	    ret = ret + (1 << it.intValue());
	}
	return ret;
    }


    /**
     * Returns true if this trace level is set
     */
    public static boolean isLevelSet(int level) {
	return ((level & mLevel) != 0);
    }


    /**
     *  Sends the trace message to the trace output device if flag is configured
     * @param f   this flag must be configured to get the message on output
     * @param s   message to send
     */
    public static void out(int f, String s) {
	if ((f & mLevel) != 0)
	    out(s);
    }


    /**
     *  Sends the trace message to the trace output device if flag is configured
     * @param f   this flag must be configured to get the message on output
     * @param s   message to send
     */
    public static void outln(int f, String s) {
	if ((f & mLevel) != 0)
	    out(s + "\n");
    }

   /**
     *  Sends the trace message to the trace output device if flag is configured
     *  the message is build by concatenation of two parameters
     * @param f   this flag must be configured to get the message on output
     * @param s1  first part of message
     * @param s2  second part of message
     */
    public static void outln(int f, String s, Object  o) {
	if ((f & mLevel) != 0)
	    out(s +o+ "\n");
    }

    /**
     *  Sends the trace message to the trace output device if flag is configured
     *  the message is build by concatenation of two parameters
     * @param f   this flag must be configured to get the message on output
     * @param s1  first part of message
     * @param s2  second part of message
     */
    public static void outln(int f, String s, int  i) {
	if ((f & mLevel) != 0)
	    out(s +i + "\n");
    }

    /**
     *  Sends the trace message to the trace output device.
     * @param s   message to send
     */
    public static void outln(String s) {
	out(s + "\n");
    }

    /**
     *  Sends the trace message to the trace output device.
     * @param s   message to send
     */
    public synchronized static void out(String s) {
	logwriter.print(logwriter.header() + s);
        logwriter.flush();
        logwriter.tempo();
    }


    /**
     *  Sends the trace message to the trace error device.
     * @param s   message to send
     */
    public static void errln(String s) {
	out(s + "\n");
    }

    /**
     *  Sends the exception and the current stack to the trace error device.
     * @param e      exception to send
     * @param stack  dumps stack if set to true
     */
    public static void errln(Throwable e, boolean stack) {
	if (stack) {
	    e.printStackTrace(logwriter);
	    out("\n");
	} else {
	    errln(e.toString());
	}
    }
    

    /**
     *  Sends the exception to the trace error device.
     * @param e      exception to send
     */
    public static void errln(Throwable e) {
	errln(e, true);
    }


    /**
     *  Sends the stream to the trace error device.
     * @param f      stream to send
     */    
    public static void errln(InputStream f) {
	try {
	    InputStreamReader isr = new InputStreamReader(f);
	    BufferedReader br = new BufferedReader(isr);
	    String line;
	    while ((line = br.readLine()) != null) {
		errln(line);
	    }
	    isr.close();
	    br.close();
	} catch (IOException e) {
	    errln("Trace.errln(InputStream) ERROR: " + e);
	}
    }


    /**
     * Closes output and error devices
     */
    public static void close() {
	logwriter.close();
    }

    public static PrintWriter getLogWriter() {
	return (PrintWriter)logwriter;
    }

    // -------------------------------------------------------
    // private methods
    // -------------------------------------------------------

    private static void configureServices() {
	TraceTimer.setDebug(isLevelSet(Trace.DB_4));
	TraceTimer.setVerbose(isLevelSet(Trace.DB_31));
	TraceTimer.setLogWriter(logwriter);
	
	TraceCore.setDebug(isLevelSet(Trace.DB_19));
	TraceCore.setVerbose(isLevelSet(Trace.DB_31));
	TraceCore.setLogWriter(logwriter);

	TraceTm.setDebugJta(isLevelSet(Trace.DB_5));
	TraceTm.setDebugTm(isLevelSet(Trace.DB_4));
	TraceTm.setVerbose(isLevelSet(Trace.DB_31));
	TraceTm.setLogWriter(logwriter);

	TraceEjb.setDebugCtx(isLevelSet(Trace.DB_12));
	TraceEjb.setDebugThread(isLevelSet(Trace.DB_7));
	TraceEjb.setDebugServer(isLevelSet(Trace.DB_14));
	TraceEjb.setDebug(isLevelSet(Trace.DB_16));
	TraceEjb.setDebugInvoke(isLevelSet(Trace.DB_17));
	TraceEjb.setDebugPool(isLevelSet(Trace.DB_21));
	TraceEjb.setDebugSecurity(isLevelSet(Trace.DB_25));
	TraceEjb.setDebugGenIC(isLevelSet(Trace.DB_15));
	TraceEjb.setDebugJms(isLevelSet(Trace.DB_22));
	TraceEjb.setVerbose(isLevelSet(Trace.DB_31));
	TraceEjb.setLogWriter(logwriter);

	TraceJms.setDebug(isLevelSet(Trace.DB_22));
	TraceJms.setVerbose(isLevelSet(Trace.DB_31));
	TraceJms.setLogWriter(logwriter);

	// Problem of compilation order here.
	// DataBaseServiceImpl.traceJdbc();
    }

    /**
     * configure output device
     */
    private static void configureOutput(String out) {
	if (out.equalsIgnoreCase("err")) {
	    logwriter = new LogWriter(System.err);
	} else if (out.equalsIgnoreCase("out")) {
	    logwriter = new LogWriter(System.out);
	} else if (out != null) {
	    String s = out + ".trace";
	    try {
		logwriter = new LogWriter(new FileOutputStream(s));
	    } catch (FileNotFoundException e) {
		System.err.println("File not found : " + e);
	    } catch (IOException e) {
		System.err.println("IOexception :" + e);
	    }
	}
    }

    /**
     * MBean method
     * @return true if header set
     */
    public static Boolean getCurrentTraceHeader() {
	return new Boolean(header);
    }
    
    /**
     * MBean method
     * @return current trace out
     */   
    public static String getCurrentTraceOut() {
	return traceOut;
    }

}

