/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 *  sebastien.chassande@inrialpes.fr
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.log;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

import javax.naming.Context;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;

import org.objectweb.monolog.api.BasicLevel;
import org.objectweb.monolog.api.Level;
import org.objectweb.monolog.api.Monitor;
import org.objectweb.monolog.api.MonitorFactory;
import org.objectweb.monolog.api.MonitorSpec;

/**
 * This class is used to map the monolog api over the jonas log system.
 * The mapping between the monolog name and the jonas trace level must 
 * be define into a configuration property file.
 * example of mapping definition from file: 
 * 		mapping.toto.titi 28
 * 		mapping.toto.tutu 13
 * 		mapping.org.objectweb 13
 * 		defaultMapping 14
 */
public class MonologTraceFactory extends AbsServiceImpl implements MonitorFactory  {
	
    static {
	BasicLevel.DEBUG 	= 0;
	BasicLevel.INFO 	= 1;
	BasicLevel.WARN 	= 2;
	BasicLevel.ERROR 	= 3;
	BasicLevel.FATAL 	= 4;
		
	BasicLevel.LEVEL_DEBUG 	= new LevelImpl(BasicLevel.DEBUG);
	BasicLevel.LEVEL_INFO 	= new LevelImpl(BasicLevel.INFO);
	BasicLevel.LEVEL_WARN 	= new LevelImpl(BasicLevel.WARN);
	BasicLevel.LEVEL_ERROR 	= new LevelImpl(BasicLevel.ERROR);
	BasicLevel.LEVEL_FATAL 	= new LevelImpl(BasicLevel.FATAL);
    }

    private static MonitorFactory singleton = null;
	
    public static MonitorFactory getInstance() throws Exception {
	if (singleton==null) 
	    throw new Exception("Log service isn't started!");
	return singleton;
    }
    private static synchronized boolean setSingleton(MonitorFactory mf) {
	if (singleton==null) {
	    singleton=mf;
	    return true;
	} else {
	    return false;
	}
    }
	
    /**
     * This field of the properties file descirbes if the instance must be set
     * as the singleton class
     */
    private static final String SINGLETON_CLASS = "singleton";

    /**
     * This field of the properties file describes the name of the file
     * where the mapping between jonas trace output and monolog name
     * is specified
     */
    private static final String CONFIGURATION = "configuration";
	
    /**
     * This field of the configuration properties file describes the default
     * jonas trace output
     */
    private static final String DEFAULT_MAPPING = "defaultMapping";
	
    /**
     * This string is the prefix to describe the mapping between monolog 
     * topic and jonas trace output.
     */
    private static final String MAPPING = "mapping.";
	
	
    /**
     * This tools class is used to store per monolog topic 
     * - the jonas trace output
     * - the reference to the Monitor instance 
     */
    private class MStruct {
	public int trace;
	public Monitor monitor;
	public MStruct(int t, Monitor m) {
	    trace = t;
	    monitor = m;
	}
    }
	
    /**
     * Association between a monolog topic and a MStruct
     */
    private Hashtable name2mstruct = null;

    /**
     * The default jonas trace output
     */
    private int defaultMapping = 29;
	
    public MonologTraceFactory(String name) {
	super(name);
	name2mstruct = new Hashtable();
    }
	
	// IMPLEMENTATION OF ABSTRACT CLASS AbsServiceImpl //
    protected synchronized void doInit(Context ctx) throws ServiceException {
	// get the file configuration properties
	String propFileName = null;
	try {
	    propFileName = (String) ctx.lookup(CONFIGURATION);
	} catch (Exception e) {
	    throw new ServiceException("ERROR : Property "
				+CONFIGURATION+" not set in jonas.properties");
	}

	try {
	    String sc = (String) ctx.lookup(SINGLETON_CLASS);
	    if ("true".equals(sc)) {
		setSingleton((MonitorFactory) this);
	    }
	} catch (Exception e) {}

	Properties p = null;
	try {
	    JProp prop = new JProp(propFileName);
	    p = prop.getFilesEnv();
	} catch (Exception e) {
	    throw new ServiceException("ERROR : Properties file "
				       +propFileName+".proerties not found");
	}

	// read properties to load configuration (mapping between 
	// monolog name and jonas level
	Enumeration enum = p.propertyNames();
	while ( enum.hasMoreElements() ){
	    String key = (String) enum.nextElement();
	    String value = (String) p.getProperty(key);
	    try {
		if ( key.equals(DEFAULT_MAPPING) ) {
		    defaultMapping = new Integer(value).intValue();
				
		} else if ( key.startsWith(MAPPING) ) {
		    int t = new Integer(value).intValue();
		    name2mstruct.put(
				     key.substring(MAPPING.length(), key.length()),
				     new MStruct(t, null)
					 );

		} else {
		    throw new ServiceException("Properties file error : field"
					       +key+" unexpected");
		}

	    } catch (NumberFormatException e) {
		throw new ServiceException("Properties file error : field"
					   +key+" doesn't contain a right int value");
	    }
	}
    }
    
    protected void doStart() throws ServiceException {
	//nothing to do
    }
    
    protected synchronized void doStop() throws ServiceException {
	// destroy all monitor
	name2mstruct.clear();
    }


    // IMPLEMENTATION OF INTERFACE MonitorFactory //
    public synchronized Monitor getMonitor(String key) {
	MStruct ms = (MStruct) name2mstruct.get(key);
	if (ms==null) {
	    // any mapping has bean specifie, then create a new Monitor
	    // mapped to the default trace number
	    ms = new MStruct(defaultMapping, (Monitor) new MonologTrace(defaultMapping) );
	    name2mstruct.put(key, ms);
	} else if (ms.monitor==null) {
	    //The mapping has been specified 
	    ms.monitor = (Monitor) new MonologTrace(ms.trace);
	}
	return ms.monitor;
    }

    public Monitor getMonitor(MonitorSpec descriptor) {
	return null;
    }
}

/**
 * This class is the implementation of the Level interface from the 
 * monolog api. This class is exactly the class provided in 
 * org.objectweb.monolog.lib package
 */
class LevelImpl implements Level {
    int value; 
	
    public LevelImpl(int _value) {
	value = _value;
    }
	
    public boolean isComparableWith(Level o) {
	return o instanceof LevelImpl;
    }
	
    public int compareTo(Object o) {
	return value - ((Level) o).getIntValue();
    }

    public int getIntValue() { 
	return value;
    }
}

/**
 * This class implements the Monitor interface from monolog api. All log calls
 * are mapped on a jonas output thaht must be specified in parameter of the
 * constructor.
 */
class MonologTrace implements Monitor {
	
    int level = BasicLevel.INFO; 
    boolean isOn = true;
    private int trace = 0;
	
    protected MonologTrace(int t) {
	switch (t) {
	case 1: trace = Trace.DB_1; break;
	case 2: trace = Trace.DB_2; break;
	case 3: trace = Trace.DB_3; break;
	case 4: trace = Trace.DB_4; break;
	case 5: trace = Trace.DB_5; break;
	case 6: trace = Trace.DB_6; break;
	case 7: trace = Trace.DB_7; break;
	case 8: trace = Trace.DB_8; break;
	case 9: trace = Trace.DB_9; break;
	case 10: trace = Trace.DB_10; break;
	case 11: trace = Trace.DB_11; break;
	case 12: trace = Trace.DB_12; break;
	case 13: trace = Trace.DB_13; break;
	case 14: trace = Trace.DB_14; break;
	case 15: trace = Trace.DB_15; break;
	case 16: trace = Trace.DB_16; break;
	case 17: trace = Trace.DB_17; break;
	case 18: trace = Trace.DB_18; break;
	case 19: trace = Trace.DB_19; break;
	case 20: trace = Trace.DB_20; break;
	case 21: trace = Trace.DB_21; break;
	case 22: trace = Trace.DB_22; break;
	case 23: trace = Trace.DB_23; break;
	case 24: trace = Trace.DB_24; break;
	case 25: trace = Trace.DB_25; break;
	case 26: trace = Trace.DB_26; break;
	case 27: trace = Trace.DB_27; break;
	case 28: trace = Trace.DB_28; break;
	case 29: trace = Trace.DB_29; break;
	case 30: trace = Trace.DB_30; break;
	case 31: trace = Trace.DB_31; break;
	default: trace = Trace.DB_29; break;
	}
    }
	
	
    private Level convert(int l) {
	if (l==BasicLevel.DEBUG) {
	    return BasicLevel.LEVEL_DEBUG;

	} else if (l==BasicLevel.INFO) {
	    return BasicLevel.LEVEL_INFO;

	} else if (l==BasicLevel.WARN) {
	    return BasicLevel.LEVEL_WARN;

	} else if (l==BasicLevel.ERROR) {
	    return BasicLevel.LEVEL_ERROR;

	} else if (l==BasicLevel.FATAL) {
	    return BasicLevel.LEVEL_FATAL;
		
	} else {
	    return BasicLevel.LEVEL_DEBUG;
	}
    }
	
    /*
     * Set the current level of the logger
     */
    public void setIntLevel(int _level) {
	level = _level;
    }

    public void setLevel(Level l) {
	level = l.getIntValue();
    }
    /**
     *  Return the current Level of the logger
     */
    public int getCurrentIntLevel() {
	return level;
    }

    public Level getCurrentLevel() {
	return convert(level);
    }
	
    /**
     * Check if the level parameter are not filtered by the logger
     */
    public boolean isLoggable(int l) {
	return l>=level;
    }

    public boolean isLoggable(Level l) {
	return l.getIntValue() >= level;
    }

    /**
     * Is the monitor enabled
     */
    public boolean isOn() {
	return isOn;
    }
	
    /**
     * Log an object with a specific level. If the level parameter is
     * loggable the object is handled.
     */
    public void log(int l, Object o) {
	if ( isOn && l>=level) 
	    Trace.outln(trace, o.toString());
    }

    public void log(Level l, Object o) {
	if ( isOn && l.getIntValue()>=level) 
	    Trace.outln(trace, o.toString());
    }

    /**
     * Log an object and a trowable with a specific level. 
     */
    public void log(int l, Object o, Throwable t) {
	if ( isOn && l>=level) 
	    Trace.outln(trace, o.toString(), t.getMessage());
    }
	 
    public void log(Level l, Object o, Throwable t) {
	if ( isOn && l.getIntValue()>=level) 
	    Trace.outln(trace, o.toString(), t.getMessage());
    }
	
    /**
     * Log an object and a trowable with a specific level. This method 
     * permits to specify an object instance and a method.
     */
    public void log(int l, Object o, Throwable t, Object location, Object method) {
	if ( isOn && l>=level) 
	    Trace.outln(trace,location.toString()+"."+method.toString()+":"+o.toString(), 
			t.getMessage() );
    }

    public void log(Level l, Object o, Throwable t, Object location, Object method) {
	if ( isOn && l.getIntValue()>=level) 
	    Trace.outln(trace, 
			location.toString()+"."+method.toString()+":"+o.toString(), 
			t.getMessage() );
    }
    /**
     * Enable the monitor
     */
    public void turnOn() {
	isOn = true;
    }
	
    /**
     * Disable the monitor
     */
    public void turnOff() {
	isOn = false;
    }

}	
