/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 *
 * 00/06/15: Jeff Mesnil
 * 01/06/15: Regis Le Brettevillois - Libelis / JOnAS team - evidian
 *           Service as org.objectweb.jonas.service.Service
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.server;

import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import javax.naming.NamingException;
import org.objectweb.common.Env;
import org.objectweb.common.RemoteObject;
import org.objectweb.jonas.adm.Adm;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.container.EJBService;
import org.objectweb.jonas.dbm.DataBaseService;
import org.objectweb.jonas.jms.JmsService;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.registry.RegistryManager;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas_ejb.lib.Version;


/**
 * This class implements an JOnAS Server.
 */
public class Server {

    static private final int trace = Trace.DB_14;
    static private String SERVERNAME="JOnAS";

    private String serverName = null;

    // The Server singleton value
    private static Server unique = null;

    // The associated ServiceManager for JOnAS
    ServiceManager sm = null;

    // Associated Adm object
    Adm adm = null;


    /**
     * Private constructor to force only 1 instance.
     */
    private Server() {
	sm = ServiceManager.getInstance();
    }

    /**
     * Get the unique instance.
     * Create it at first call.
     */
    public static Server getInstance() {
	if (unique == null) {
	    unique = new Server();
	}
	return unique;
    }

    /**
     * Start the server
     */
    public void start() throws Exception {

	// Print jonas version and orb
	String orbs = null;
	if (Env.isJeremie()) {
	    orbs = "Jeremie.";
	} else if (Env.isSunRMI()){
	    orbs = "rmi.";
	} else if (Env.isDavid()) {
	    orbs = "David RMI.";
	}

	System.out.println("JOnAS Server, version "+Version.NUMBER+", running on " + orbs);

	// Read the JONAS properties
	JProp props = null;
	try {
	    props = new JProp();
	} catch (Exception e) {
	    throw new Exception("JOnAS: Cannot read the JOnAS properties ("+e+")");
	}
	// get the jonas.name property value:
	serverName = props.getValue("jonas.name", "jonas"); 

        // start the framework
	this.startFramework(props);

	// start services
	sm.startServices();

	// JMX: register ejbserver and trace
	try {
	    JmxServer serverMBean = new JmxServer(this, serverName, Version.NUMBER, orbs); 
	    ((JmxService)ServiceManager.getJmxService()).getJmxServer().registerMBean(serverMBean, JonasObjectName.ejbServer());
	    ((JmxService)ServiceManager.getJmxService()).getJmxServer().registerMBean(new JmxTrace(),JonasObjectName.traceService());
	} catch (ServiceException se) {
	    // Jmx Service not available, do nothing
	} catch (Exception e) {
	    throw e;
	}

	// TM Server or EJB Server ?
	boolean isNotEJB = false;
	try {
	    Service ejbservice = ServiceManager.getEjbService();
	} catch (ServiceException se) {
            isNotEJB = true;
	}

	// Creates the Adm instance for remote administration
	try {
	    adm =  new Adm(props, isNotEJB);
	} catch (Exception e) {
	    throw new Exception("Cannot create Adm ("+e+")");
	}

	Trace.outln(Trace.DB_31, SERVERNAME+" is ready");
    }
    
    void startFramework(JProp props) throws Exception {

	// RMI Security Manager
	if (System.getSecurityManager() == null) {
	    System.setSecurityManager(new RMISecurityManager());
	}

	// Init Trace system
	Trace.configure();

	// Init the ORB
	try {
	    // Port value zero: an anonymous port is choosen
	    int portNumber = Integer.parseInt(props.getValue("jonas.orb.port", "0"));
	    RemoteObject.setPort(portNumber);
	    Trace.outln(trace, "The 'jonas.orb.port' value is "+portNumber);
	} catch (NumberFormatException e) {
	    Trace.errln("JOnAS: Invalid 'jonas.orb.port' property value (default value used!)");
	}

    }

    /**
     * Stop the server without stopping the JVM
     */
    public void stop() throws Exception {
	if (adm != null) {
	    adm.stopServer();
	}
    }

    /**
     * Stop the server and stop the JVM (MBean method)
     */
    public void kill() throws Exception {
	if (adm != null) {
	    adm.killServer();
	}
    }


    /**
     * Server main routine.
     * No args.
     */
    public static void main(String args[]) {
	Server server = Server.getInstance();
	try {
	    server.start();
	} catch (Exception e) {
	    Trace.errln("JOnAS error: "+e.getMessage());
	    Trace.errln(SERVERNAME+" halting");
	    System.exit(2);
	}
    }

}
