/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.util.Hashtable;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.Transaction;
import javax.transaction.TransactionRolledbackException;

// ---------------------------------------------------------------------
// Context Switch object used internally by JBeanEntity to synchronize
// accesses to the entity context and thus to the entity bean instance.
// ---------------------------------------------------------------------
public class ContextSwitch {

    // bean related values
    private JContainerImpl container;
    private JBeanEntity ejbobject;
    private JEntityHome home;
    private boolean reentrant;
    private boolean shared;

    // ident is used mainly for debugging
    private static int counter = 0;
    private String ident;

    // unique EntityContext 
    private JContextEntity ectx = null;

    // nb of threads waiting for the Context
    private int waiters = 0;

    // nb of requests with no transaction.
    private int IHcount = 0;

    // True if ctx has been used outside transactions.
    private boolean dirty = false;

    // True if a TX need this instance currently used outside tx
    private boolean mustStore = false;

    // True if we must reload the bean state before use.
    private boolean mustLoad = false;

    // True when context/instance has been discarded (outside tx only)
    private boolean IHdiscarded = false;

    // current tx running on this instance and not yet committed.
    private Transaction currtx = null;

    /**
     * constructor
     */
    public ContextSwitch(JBeanEntity e) {
	ejbobject = e;
	home = (JEntityHome) ejbobject.getHome();
	container = home.getContainer();
	reentrant = home.isReentrant();
	shared = home.isShared();
	// generate new ident (debug)
	counter++;
	ident = "<"+counter+">";
    }

    /**
     * Get a context/instance (a new request just arrived)
     * @param tx - the Transaction object
     * @param activate true if must activate instance (normal case)
     */
    public synchronized JContextEntity getICtx(Transaction tx, boolean activate) throws RemoteException {

	if (tx == null) {
	    // ------------- IH begin ------------- 
	    TraceEjb.debugPool(ident+"getICtx IH");
	    // Wait until context is available.
	    while (currtx != null || (!reentrant && IHcount > 0)) {
		if (currtx != null) {
		    TraceEjb.debugPool(ident+"getICtx IH: WAIT end IT");
		} else {
		    TraceEjb.debugPool(ident+"getICtx IH: WAIT end other IH");
		}
		waiters++;
		try {
		    wait();
		} catch (Exception e) {
		    TraceEjb.error(ident+"getICtx IH:", e);
		    throw new RemoteException("EntityContext Synchronization problem", e);
		} finally {
		    waiters--;
		}
		TraceEjb.debugPool(ident+"getICtx IH: NOTIFIED");
	    }

	    if (ectx == null) {
		// No context mapped: get one from the pool.
		TraceEjb.debugPool(ident+"getICtx IH: get new context from the pool");
		ectx = ejbobject.getContextFromPool(tx);
		// Activate this instance (not if create)
		if (activate) {
		    ectx.activate(true);
		}
	    } else {
		ectx.reuseEntityContext(tx);
		if (shared) {
		    // reload the bean state, because for example bean is shared.
		    TraceEjb.debugPool(ident+"getICtx IH: reload bean state");
		    ectx.activate(false);
		    mustLoad = false;
		}
	    }
	    IHcount++;
	    dirty = true;

	} else {
	    // ------------- IT begin ------------- 
	    TraceEjb.debugPool(ident+"getICtx IT");

	    // wait instance/context available
	    while (dirty || (currtx != null && !tx.equals(currtx))) {
		// If dirty, must passivate before use.
		// Check this inside the loop in case other thread passed before me
		// The only risk is to register it one more for nothing.
		if (dirty) {
		    if (IHcount == 0) {
			// Cannot passivate here, because must be done outside tx context
			container.registerEJB(ejbobject);
		    } else {
			mustStore = true;
		    }
		}
		if (dirty) {
		    TraceEjb.debugPool(ident+"getICtx IT: WAIT passivation");
		} else {
		    TraceEjb.debugPool(ident+"getICtx IT: WAIT end IT");
		}
		waiters++;
		try {
		    wait();
		} catch (Exception e) {
		    TraceEjb.error(ident+"getICtx IT:", e);
		    throw new RemoteException("EntityContext Synchronization problem", e);
		} finally {
		    waiters--;
		}
		TraceEjb.debugPool(ident+"getICtx IT: NOTIFIED");
		// Check that meanwhile, the transaction has not been rolledback
		try {
		    switch (tx.getStatus()) {
		    case Status.STATUS_MARKED_ROLLBACK:
		    case Status.STATUS_ROLLEDBACK:
		    case Status.STATUS_ROLLING_BACK:
			TraceEjb.verbose(ident+"getICtx IT: transaction rolled back");
			throw new TransactionRolledbackException("rollback occured while waiting");
		    default:
			break;
		    } 
		} catch (Exception e) {
		    TraceEjb.verbose(ident+"getICtx IT: no more transaction");
		    throw new TransactionRolledbackException("rollback occured while waiting");
		}
	    }


	    if (ectx == null) {
		TraceEjb.debugPool(ident+"getICtx IT: get new context from the pool");
		// Get a new context from the pool
		ectx = ejbobject.getContextFromPool(tx);

		// Activate this instance (not if create)
		if (activate) {
		    ectx.activate(true);
		}
	    } else {
		ectx.reuseEntityContext(tx);
		if (mustLoad) {
		    // reload the bean state, because for example bean is shared.
		    TraceEjb.debugPool(ident+"getICtx IT: reload bean state");
		    ectx.activate(false);
		    mustLoad = false;
		}
	    }

	    // register this Context to release instance at commit
	    if (currtx == null) {
		TraceEjb.debugPool(ident+"getICtx IT: new tx, registerSynchronization");
		currtx = tx;
		try {
		    tx.registerSynchronization(ectx);
		} catch (RollbackException e) {
		    TraceEjb.verbose(ident+"getICtx IT: transaction has been marked rollbackOnly");
		    releaseICtx(false);
		    throw new TransactionRolledbackException("cannot register EntityContext");
		} catch (Exception e) {
		    releaseICtx(false);
		    TraceEjb.error(ident+"getICtx IT: registerSynchronization: ", e);
		    throw new RemoteException("registerSynchronization problem", e);
		}
	    }
	}
	return ectx;
    }

    /**
     * Release a context/instance
     * IH: end of request.
     * IT: end of transaction.
     * @param committed - false if transaction has been rolled back
     */
    public synchronized void releaseICtx(boolean committed) {

	if (currtx == null) {
	    // ------------- IH end ------------- 
	    TraceEjb.debugPool(ident+"releaseICtx IH");

	    IHcount--;
	    if (IHcount == 0) {
		if (mustStore) {
		    try {
			ectx.storeIfModified();
		    } catch (RemoteException e) {
			TraceEjb.verbose(ident+"releaseICtx:"+e);
		    }
		    mustStore = false;
		    dirty = false;
		}
		if (waiters > 0) {
		    TraceEjb.debugPool(ident+"releaseICtx IH: notify");
		    notifyAll();
		}
	    }
	} else {
	    // ------------- IT end ------------- 
	    TraceEjb.debugPool(ident+"releaseICtx IT");

	    currtx = null;

	    if (waiters > 0) {
		if (shared || !committed) {
		    mustLoad = true;
		}
		TraceEjb.debugPool(ident+"releaseICtx IT: notify");
		notifyAll();
	    } else {
		// If bean is shared, we implement commit option C.
		// If bean is not shared, commit option A is better.
		if (shared || !committed) {
		    ectx.passivate();
		    releaseInstanceContext();
		}
	    }
	}
    }

    /**
     * Discard context/instance
     * IH only because IT are managed by TM.
     * A discarded instance will not be passivated.
     */
    public synchronized void discardICtx() {
	TraceEjb.debugPool(ident+"discardICtx IH");

	IHdiscarded = true;
	IHcount--;
	if (IHcount == 0) {
	    releaseInstanceContext();
	}
    }

    /*
     * release instance/context (internal method)
     */
    private void releaseInstanceContext() {
	TraceEjb.debugPool(ident+"releaseInstanceContext");

	ejbobject.putContextInPool(ectx);
	ectx = null;
	IHdiscarded = false;
	dirty = false;
    }

    /**
     * try to passivate IH (called from swapper)
     * args passivation false if only store bean state.
     * @return true if OK, false if busy.
     */
    public synchronized boolean passivateIH(boolean passivation) {

	// If no context mapped, nothing to do
	if (ectx == null) {
	    return true;
	}
	// If the instance is busy, don't touch it.
	if (IHcount > 0 || currtx != null) {
	    TraceEjb.debugPool(ident+"passivateIH: busy");
	    TraceEjb.debugPool("IHcount="+IHcount+" currtx="+currtx+" waiters="+waiters);
	    return false;
	}
	TraceEjb.debugPool(ident+"passivateIH");

	// Store the instance state
	// -> only if instance not discarded
	if (! IHdiscarded) {
	    try {
		ectx.storeIfModified();
	    } catch (Exception e) {
		TraceEjb.error(ident+"error while storing bean state:", e);
	    }
	    mustStore = false;
	    dirty = false;
	}

	// passivate this instance if required.
	if (passivation) {
	    ectx.passivate();
	    releaseInstanceContext();
	}

	if (waiters > 0) {
	    TraceEjb.debugPool(ident+"passivateIH: notify");
	    notifyAll();
	}
	return true;
    }

    /**
     * @return ejb instance's state
     * state values are 0=in-tx, 1=out-tx, 2=idle, 3=passive.
     * we don't synchronize this method to avoid jadmin blocks
     */
    public int getState() {
	if (ectx == null) {
	    return 3;
	}
	if (currtx != null) {
	    return 0;
	}
	if (dirty) {
	    return 1;
	} else {
	    return 2;
	}
    }

    /* 
     * Return true if instance busy
     *
     */

    public boolean isBusy(){
	return (IHcount > 0 || currtx != null);
    }
}
