/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.security.Guard;
import javax.ejb.EJBContext;
import javax.ejb.EJBHome;
import javax.ejb.EJBObject;
import javax.ejb.Handle;
import javax.ejb.RemoveException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRolledbackException;
import org.objectweb.common.RemoteObject;

/**
 * This class implements javax.ejb.EJBObject class.
 * This class is extended by all the interposition generated classes 
 *  corresponding to an EJB.
 * It mainly implements two operations <preinvoke> and <postinvoke>
 *  that are called before and after the business method of the Bean.
 */
public abstract class JBean extends RemoteObject implements EJBObject {

    /**
     * @serial
     */
    TransactionManager tm = null;

    // Must be declared as "protected" because accessed by generated classes

    /**
     * @serial
     */
    protected JBeanHome home = null;

    /**
     * @serial
     */
    protected JContainerImpl container = null;

    /**
     * @serial
     */
    Guard methodGuard = null;

    /**
     * @serial
     */
    protected Handle handleEJBObject = null;

    /**
     * Constructor for a type of bean
     *
     * @param      EJBhome home
     * @exception  RemoteException
     */
    JBean(JBeanHome home) throws RemoteException {
	this.home = home;
	container = home.getContainer();
	methodGuard = container.getMethodGuard();
	tm = home.getTransactionManager();
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    // --------------------------------------------------------------------------

    /**
     * Obtains the enterprise Bean's home interface.
     *
     * @return    A reference to the enterprise Bean's home interface. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public synchronized EJBHome getEJBHome() throws RemoteException {
	return (EJBHome) home;  
    }

    /**
     * Obtains the primary key of the EJB object.
     *
     * @return    The EJB's object primary key.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    abstract public Object getPrimaryKey() throws RemoteException;
  

    /**
     * Removes the EJB object.
     *
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     * @exception RemoveException: The enterprise Bean or the container does not allow 
     *            destruction of the object.
     *
     */
    abstract public void remove() throws RemoteException, RemoveException;


    /**
     * Obtains a handle for the EJB object. The handle can be used at later time to re-obtain
     *  a reference to the EHB object, possibly in a different Java Virtual Machine
     *
     * @return    A handle for the EJB object. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    abstract public Handle getHandle() throws RemoteException;

    /**
     * Tests if a given EJB is identical to the invoked EJB object.
     *
     * @param     EJBObject obj - An object to test for identity with the invoked object. 
     * @return    True if the given EJB object is identical to the invoked object. 
     * @throws RemoteException: Thrown when the method failed due to a system-level failure.
     */
    abstract public boolean isIdentical(EJBObject obj) throws RemoteException;

    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------
 
    /**
     * Common postinvoke routine for Session or Entity
     * See "Container Provider Responsabilities" in EJB 1.1 spec.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugInvoke("JBean.postinvoke");

	// Check reference to EJBContext
	EJBContext ctx = rctx.getEJBContext();
	if (ctx == null) {
	    TraceEjb.error("null EJBContext in RequestCtx!");
	}

	// Commit the transaction started in preInvoke
	if (rctx.mustCommit()) {
	    if (rctx.getSysExc() != null) {
		TraceEjb.error("System Exception raised by request:", rctx.getSysExc());
		//rollback the transaction
		try {
		    tm.rollback();
		} catch (Exception e) {
		    TraceEjb.error("Exception during rollback:", e);
		}
	    } else {
		// try to commit the transaction, even in case of appl. Exc.
		// If rollbackonly, do not throw any runtime exception to the client.
		boolean rollb = ctx.getRollbackOnly();
		try {
		    if (rollb) {
			TraceEjb.debugInvoke("postinvoke: Rolling back transaction");
			tm.rollback();
		    } else {
			TraceEjb.debugInvoke("postinvoke: Committing transaction");
			tm.commit();
		    }
		} catch (RollbackException e) {
		    // Probably a pb while writing data on database.
		    throw new RemoteException("Container cannot commit a transaction", e);
		} catch (Exception e) {
		    TraceEjb.error("postinvoke: ", e);
		    throw new RemoteException("Container exception", e);
		}
	    }
	}

	// We got a system Exception in business method:
	// - log exception
	// - discard instance
	// - set client transaction rollback only
	// - throw remote exception
	if (rctx.getSysExc() != null) {
	    // Log system exception
	    TraceEjb.error("System Exception in business method:", rctx.getSysExc());

	    // If client transaction: set it rollback only and throw TransactionRolledBackException
	    Transaction currentTx = null;
	    try {
		currentTx = tm.getTransaction();
		if (currentTx != null) {
		    currentTx.setRollbackOnly();
		    throw new TransactionRolledbackException("System exception in business method");
		}
	    } catch (SystemException e) {
		TraceEjb.error("postinvoke: cannot set rollback only current tx:", e);
	    }
	    return;
	}

	// Resume client transaction suspended in preInvoke
	Transaction tx = rctx.getClientTx();
	if (tx != null) {
	    try {
		tm.resume(tx);
	    } catch (SystemException e) {
		TraceEjb.error("postinvoke: Cannot resume transaction");
	    }
	}
    }

    /**
     * EJBObject no longer accessible
     */
    public void noMoreInUse() {
	TraceEjb.debugInvoke("JBean.noMoreInUse");
	try {
	    if (unexportObject(this, true) == false) {
		TraceEjb.error("noMoreInUse: unexportObject failed");
	    }
	} catch (Exception e) {
	    TraceEjb.error("noMoreInUse: Object was not exported");
	}
    }
    
    /**
     * EJBObject accessible again
     * @return true if object has been exported successfully
     */
    public boolean reUseObject() {
	TraceEjb.debugInvoke("JBean.reUseObject");
	try {
	    exportObject(this, getPort());
	} catch (RemoteException e) {
	    TraceEjb.error("reUseObject: cannot export:"+e);
	    return false;
	}
	return true;
    }
}
