/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.server.Unreferenced;
import javax.ejb.EJBObject;
import javax.ejb.EJBException;
import javax.naming.Context;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionRolledbackException;

/**
 * JBeanEntity is the base class for all Entity Beans.
 */
abstract public class JBeanEntity extends JBean implements Unreferenced {

    /**
     * @serial
     */
    private Object primarykey = null;

    /**
     * ContextSwitch
     */
    private ContextSwitch ctxSwitch;

    /**
     * persistence is container managed
     * @serial
     */
    private boolean isContainerManaged;

    /**
     * Object is unreferenced (RMI only)
     */
    private boolean unReferenced;

    /**
     * constructor
     */
    public JBeanEntity(JEntityHome home, boolean cmp) throws RemoteException {
	super(home);
	isContainerManaged = cmp;
	unReferenced = false;
	ctxSwitch = new ContextSwitch(this);
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    //
    // getEJBHome	in JBean
    // getHandle	generated by GenIC
    // getPrimaryKey	ok
    // isIdentical	ok
    // remove		generated by GenIC
    // --------------------------------------------------------------------------

     /**
     * Obtains the primary key of the EJB object.
     *
     * @return    The EJB's object primary key.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public Object getPrimaryKey() throws RemoteException {
	if (primarykey == null) {
	    TraceEjb.debugInvoke("getPrimaryKey -> null");
	}
	return primarykey;
    }
 
    /**
     * Tests if a given EJB is identical to the invoked EJB object.
     *
     * @param     EJBObject obj - An object to test for identity with the invoked object. 
     * @return    True if the given EJB object is identical to the invoked object. 
     * @throws RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public boolean isIdentical(EJBObject obj) throws RemoteException {
	boolean ret = false;
	if (obj != null) {
	    ret = obj.equals(this);
	}
	return ret;
    }

    // -------------------------------------------------------------------
    // Unreferenced implementation
    // -------------------------------------------------------------------

    /*
     * Called by the RMI runtime sometime after the runtime determines that
     * the list of clients referencing the remote object becomes empty.
     * mark the object as unreferenced
     */
    public synchronized void unreferenced() {

	TraceEjb.debug("JBeanEntity.unreferenced");
	//TraceEjb.verbose(">>>>>>> JBeanEntity.unreferenced for home: "+home.getEjbName());

	// If the container has been removed by JonasAdmin: nothing to do.
	if (home.isUnregistered()) {
	    return;
	}

	// mark object as "unreferenced"
	unReferenced = true;

    }

    /**
     * Return true if the current instance is unreferenced
     *
     */

    public boolean isUnreferenced(){
	return unReferenced;
    }

    /**
     * Return true if thecurrent instance is busy
     *
     */

    public boolean isBusy(){
	return(ctxSwitch.isBusy());
    }

    // --------------------------------------------------------------------------
    // EntityContext management
    // --------------------------------------------------------------------------

    /**
     * Get a context from the pool
     */
    public JContextEntity getContextFromPool(Transaction tx) throws RemoteException {
	TraceEjb.debugInvoke("JBeanEntity: getContextFromPool");
	JEntityHome h = (JEntityHome)home;
	JContextEntity ec = (JContextEntity)h.reuseEJBCtx();
	if (ec == null) {
	    ec = h.createEntityContext(this, tx);
	} else {
	    ec.initEntityContext(this, tx);
	}
	return ec;
    }

    /**
     * put back a context in the pool
     */
    public void putContextInPool(JContextEntity ec) {
	TraceEjb.debugInvoke("JBeanEntity: putContextInPool");
	ec.razEntityContext();
	home.freeEJBCtx(ec);
    }

    /** 
     * Gets the Instance Context for a given transactional context
     * if it does not exist get a new instance from the pool
     *
     * @param tx - Transaction object corresponding to the current transaction
     * @return the JContextEntity object
     *
     */
    public JContextEntity getEntityContext(Transaction tx) throws RemoteException {
	if (TraceEjb.isDebugInvoke) TraceEjb.debugInvoke("JBeanEntity.getEntityContext: " + tx);

	return ctxSwitch.getICtx(tx, true);
    }
    
    /** 
     * Gets the Instance Context for the proper transactional context
     * if it does not exist a new Instance Context is created and activated
     * this routine is only (and must only)called in case of remove methods.
     *
     * @return the JContextEntity object
     * @throws javax.ejb.RemoveException in case of pb in activation
     *                during remove
     *
     */
    public JContextEntity getEntityContext() throws RemoteException,javax.ejb.RemoveException  {
	TraceEjb.debugInvoke("JBeanEntity.getEntityContext");
	JContextEntity ec = null;
	// For which transaction ?
	 Transaction tx = null;
	 try {
	     tx = tm.getTransaction();
	 } catch(SystemException e) {
	     TraceEjb.error("getEntityContext: cannot get transaction:", e);
	 }

	 // Activate the instance before removing it (because possible rollback)
	 try {
	     ec = ctxSwitch.getICtx(tx, true);
	 } catch(Exception e){
	     // Pb during activation we can't remove properly
	     throw new javax.ejb.RemoveException(e.toString());
	 }
	 return ec;
    }
    
    /** 
     * Gets a new Instance Context for the proper transactional context.
     * Used only for create methods, when object instance have just been created
     * and when no context can exist yet.
     * We don't use getEntityContext in case of create methods because we do not have
     * to activate the instance since object did not exist before.
     *
     * @return the JContextEntity object
     */
    public JContextEntity getNewEntityContext() throws RemoteException {
	TraceEjb.debugInvoke("JBeanEntity.getNewEntityContext");

	// For which transaction ?
	Transaction tx = null;
	try {
	    tx = tm.getTransaction();
	} catch(SystemException e) {
	    TraceEjb.error("getNewEntityContext: cannot get transaction:", e);
	}

	// No activate in case of create
	JContextEntity ec = ctxSwitch.getICtx(tx, false);
	ec.setNewInstance(true);
	return ec;
    }

    /**
     * Release Instance/Context
     */
    public void releaseContext(boolean committed) {
	TraceEjb.debugInvoke("JBeanEntity.releaseContext");
	ctxSwitch.releaseICtx(committed);
    }

    /**
     * Discard instance/Context used outside transactions
     */
    public void discardContext() {
	TraceEjb.debugInvoke("JBeanEntity.discardContext");
	ctxSwitch.discardICtx();
    }

    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    /**
     * Sets the primary Key of the Entity Bean 
     *
     * @param pk - primary key
     */
    public void setPrimaryKey(Object pk) {
	primarykey = pk;
    }


    /**
     * Returns the JEntityHome
     */
    public JEntityHome getHome() {
	return (JEntityHome)home;
    }

	
    /**
     * This method makes the container's job before the invocation of the business method
     * for a EntityBean
     * Here are treated transaction attributes in order to execute the business.
     *  method in the correct transactional context. A new transaction may be created.
     *  It is important that this method is reentrant, because instance of bean
     *  may be accessed from multiple transactions.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @return    the selected context of the entityBean 
     * 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     */
    public RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException {
	TraceEjb.debugInvoke("JBeanEntity.preinvoke");
	try {
	    // security controls
	    home.checkSecurity(methodSignature);

	    // Build a RequestCtx to save information about this request
	    // until the postinvoke time.
	    RequestCtx rctx = new RequestCtx(txattr);

	    // Set bean context for JNDI
	    rctx.setOldJNDICtx(home.setComponentContext());

	    // Process transaction attribute
	    container.checkTransactionIn(rctx);

	    // Get a proper instance/context for this request
	    // If exception, just call postinvoke and throw a RemoteException
	    try {
		JContextEntity ec = getEntityContext(rctx.getCurrentTx());
		rctx.setEJBContext(ec);
	    } catch (TransactionRolledbackException e) {
		// no need to go further if transaction will be rolled back anyway.
		TraceEjb.verbose("preinvoke exception: "+e);
		// Reset bean context for JNDI
		home.naming.setComponentContext(rctx.getOldJNDICtx());
		throw e;
	    } catch (RemoteException e) {
		TraceEjb.error("preinvoke exception: ", e);
		rctx.setSysExc(e);
		postinvoke(rctx);
		throw e;
	    } catch (Error e) {
		TraceEjb.error("preinvoke error: ", e);
		rctx.setSysExc(e);
		postinvoke(rctx);
		throw new RemoteException("Cannot activate instance", e);
	    } catch (Exception e) {
		TraceEjb.error("preinvoke: cannot get an EntityContext: ", e);
		rctx.setSysExc(e);
		postinvoke(rctx);
		throw new RemoteException("Cannot activate instance", e);
	    }
	    // return rctx to the generated code
	    return rctx;
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }	

    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugInvoke("JBeanEntity.postinvoke");
	try {
	    try {
		// common part between entity and session beans
		super.postinvoke(rctx);
	    } finally {
		// Reset bean context for JNDI
		home.naming.setComponentContext(rctx.getOldJNDICtx());
		// If no transaction, must release context (update count of requests)
		// Must be done even if super.postinvoke threw an exception
		if (rctx.getCurrentTx() == null) {
		    if (rctx.getSysExc() != null) {
			discardContext();
		    } else {
			releaseContext(true);
		    }
		}
	    }
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /** 
     * Synchronizes instance state.
     * this method should be called by the swapper (container),
     * either on timeout, or when a request implied in a transaction
     * occured on this instance.
     * @return boolean false if instance was busy
     */
    public boolean storeState(boolean passivation) {
	return ctxSwitch.passivateIH(passivation);
    }

    /**
     * Get instance state
     */
    public int getInstanceState() {
	return ctxSwitch.getState();
    }
}

