/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): 
 * JOnAS 2.1 Jun Inamori
 *           poolling of EJBObjects
 *
 * 02/01/15 Dean Jennings - List  instead of Vector for pooledinstances
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.security.Guard;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.EmptyStackException;
import java.util.Properties;
import javax.ejb.EJBHome;
import javax.ejb.EJBMetaData;
import javax.ejb.Handle;
import javax.ejb.HomeHandle;
import javax.ejb.RemoveException;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.InitialContext;
import javax.naming.Reference;
import javax.naming.StringRefAddr;
import javax.transaction.TransactionManager;

import org.objectweb.common.RemoteObject;
import org.objectweb.common.Env;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.enhydra.naming.ContainerNaming;


/**
 * This class implements javax.ejb.EJBHome interface.
 * This class must be extended by the EJBHome class generated for a given bean.
 * 
 * Contains all information related to the bean.
 */
public abstract class JBeanHome extends RemoteObject implements EJBHome, Ejb  {

    /**
     * @serial
     */
    private BeanDesc ejbDescriptor = null; 

    /**
     * @serial
     */
    private JBeanMetaData ejbMetaData = null; 

    /**
     * @serial
     */
    private String beanName = null;

    /**
     * @serial
     */
    private Properties ejb10Env = null;

    /**
     * @serial
     */
    private HomeHandle handleEJBHome = null;

    /**
     * @serial
     */
    private boolean isSession = true;

    /**
     * @serial
     */
    private String beanClassName = null;

 
    /**
     * @serial
     */
    boolean TxBeanManaged = false;
 

    /**
     * JNDI Context for the bean. 
     * @serial
     */
    protected Context JNDICtx = null;

    /**
     * @serial
     */
    ContainerNaming naming = null;

    /**
     * @serial
     */
    TransactionManager tm = null;

    /**
     * @serial
     */
    JContainerImpl container = null;

    /**
     * @serial 
     */
    Guard methodGuard = null;

    /**
     * @serial
     */
    protected List pooledinstances = Collections.synchronizedList(new ArrayList());

    /**
     * @serial
     */
    private boolean unregistered = false; // true when beans are removed from container

    // ------------------------------------------------------------------
    // Constructors
    // ------------------------------------------------------------------

    /**
     * Constructor for the base class of the specific generated Home object.
     */
    public JBeanHome(BeanDesc desc, String beanName, JContainerImpl container) throws RemoteException {

	if (TraceEjb.isDebug) TraceEjb.debug("JBeanHome constructor for "+ beanName);

	this.beanName   = beanName;
	this.container = container;
	tm = container.getTransactionManager();
	methodGuard = container.getMethodGuard();
	naming = container.getContainerNaming();

	ejbDescriptor = desc;
	BeanNaming bn = new BeanNaming(ejbDescriptor);
	beanClassName = bn.getFullDerivedBeanName();
	ejb10Env = desc.getEjb10Environment();

	try {
	    // Create a JNDI context for this bean home
	    Context ctx = naming.createEnvironmentContext(beanName);
	    // setup the bean environment in this context
	    container.setBeanEnvironment(ctx, desc);
	    JNDICtx = ctx;
	    register();
	} catch(NamingException e) {
	    throw new RemoteException("failed to register bean to jndi", e);
	}
    }

    // ------------------------------------------------------------------
    // EJBHome implementation
    // ------------------------------------------------------------------

    /**
     * Obtains the EJBMetaData interface for the enterprise Bean. 
     * @return The enterprise Bean's EJBMetaData
     */
    public EJBMetaData getEJBMetaData() throws RemoteException {
	TraceEjb.debug("JBeanHome.getEJBMetaData");
	return ejbMetaData;
    }

    /**
     * Removes an EJB object identified by its handle. 
     */
    abstract public void remove(Handle handle) throws RemoteException, RemoveException;

    /**
     * Removes an EJB object identified by its primary key.
     */
    abstract public void remove(Object primaryKey) throws RemoteException, RemoveException;

    /**
     * Obtain a handle for the home object. The handle can be used at later time to re-obtain a
     * reference to the home object, possibly in a different Java Virtual Machine.
     * 
     * @return A handle for the home object.
     * 
     * @exception java.rmi.RemoteException - Thrown when the method failed due to a system-level
     * failure.
     */
    public HomeHandle getHomeHandle() throws java.rmi.RemoteException {

	TraceEjb.debug("JBeanHome.getHomeHandle()");

	if (handleEJBHome == null) {
	    handleEJBHome = new HomeHandleImpl(getDeploymentDescriptor().getJndiName(), naming.getEnv());
	}
	return handleEJBHome;
    }

    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    private static InitialContext corbaIctx = null;

    /**
     * Returns the ejb-name of the bean
     */

    public String getEjbName() {
	return(beanName);
    }

    /**
     * returns an InitialContext to access CosNaming
     */
    static InitialContext getCorbaInitialContext() throws NamingException {
	if (corbaIctx == null) {
	    Properties properties = new Properties();
	    properties.put("java.naming.factory.initial", "com.sun.jndi.cosnaming.CNCtxFactory");
	    String host = System.getProperty("david.CosNaming.default_host", "localhost");
	    String port = System.getProperty("david.CosNaming.default_port", "14727");
	    properties.put("java.naming.provider.url", "iiop://" + host + ":" + port);
	    corbaIctx =  new InitialContext(properties);
	}
	return corbaIctx;
    }
    
    /**
     * register this bean to JNDI (rebind)
     */
    private void register() throws NamingException {
	String name = ejbDescriptor.getJndiName();
	
	if (Env.isDavid()) {
	    // register in CosNaming
	    getCorbaInitialContext().rebind(name, this);
	    
	    // register a reference of the home bean in the rmiregistry so
	    // EJBs can lookup other beans
	    Reference ref = new Reference (getClass().getName(), 
					   "org.objectweb.jonas_ejb.container.HomeFactory",
					       null);
	    ref.add(new StringRefAddr("bean.name", name));
	    naming.getInitialContext().rebind(name, ref);
	} else
	    // register in default InitialContext
	    naming.getInitialContext().rebind(name, this);
    }

    /**
     * unregister this bean in JNDI (unbind)
     */
    protected synchronized void unregister() throws NamingException {
	String name = ejbDescriptor.getJndiName();
	
	if (Env.isDavid()) {
	    // unregister in CosNaming
	    getCorbaInitialContext().unbind(name);
	} else
	    // unregister in default InitialContext
	    naming.getInitialContext().unbind(name);

	unregistered = true;
    }
            
    /**
     * return true if home is removed
     */
    protected synchronized boolean isUnregistered() {
	return unregistered;
    }

    public void setEJBMetaData(JBeanMetaData md) {
	TraceEjb.debug("JBeanHome.setEJBMetaData");	
	ejbMetaData = md;
    }


    /**
     * return true if the bean manages itself the transactions
     */
    public boolean isTxBeanManaged() {
	return TxBeanManaged;
    }

    // DeploymentDescriptor
    public BeanDesc getDeploymentDescriptor() {
	TraceEjb.debug("JBeanHome.getDeploymentDescriptor()");
	return ejbDescriptor;
    }


    /**
     * Returns the EJB 1.0 style environment associated with the Bean
     */
    public Properties getEjb10Environment() {
	if (TraceEjb.isDebug) TraceEjb.debug("JBeanHome.getEjb10Environment "+beanName);
	return ejb10Env;
    }

    /**
     * Gets the full name of the Enterprise Bean's Class Name
     * (Gives the derived class generated by GenIC, if it exists)
     */
    public String getEnterpriseBeanClassName(){
	return beanClassName;
    }

    /**
     * set the Component Context for JNDI environment
     */
    public Context setComponentContext() {
	return naming.setComponentContext(JNDICtx);
    }

    /**
     * reset old Component Context for JNDI environment
     */
    public void resetComponentContext(Context oldctx) {
	naming.setComponentContext(oldctx);
    }

    /**
     * get the Container
     */
    public JContainerImpl getContainer() {
	return container;
    }

    /**
     * Check Security before calling a method.
     * @param methodSignature String representing the method
     */
    public void checkSecurity(String methodSignature) throws RemoteException {
	if (TraceEjb.isDebug) TraceEjb.debug("JBeanHome.checkSecurity");
	// A "null" string means : no security control.
	if (methodSignature.equals("null")) {
	    return;
	}
	// methodGuard == null : no security service available
	if(methodGuard == null){
	    TraceEjb.debugSecurity("JBeanHome.checkSecurity:  no security service available");
	    return;
	}
	TraceEjb.debugSecurity("JBeanHome.checkSecurity()");
	try {
	    methodGuard.checkGuard(methodSignature);
	} catch (SecurityException e) {
	    throw new RemoteException("checkSecurity", e);
	}
    }
    
    /**
     * return the reference on the Transaction Manager
     */
    public TransactionManager getTransactionManager() {
	return tm;
    }

    /**
     * preinvoke for home methods
     */
    abstract protected RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException;
    
    /**
     * postinvoke for home methods
     */
    abstract protected void postinvoke(RequestCtx rctx) throws RemoteException;
    

    // ----------------------------------------------------------------------
    // Pool of instance/context (entity or session beans)
    // ----------------------------------------------------------------------

    /**
     * Get an instance/context from the pool
     */
    public JContext reuseEJBCtx() {
	JContext ec = null;
	    try {
		synchronized (pooledinstances) {
		    ListIterator i = pooledinstances.listIterator();
		    if (i.hasNext()) {
			ec = (JContext) i.next();
			i.remove();
		    }
		}
	    } catch(IndexOutOfBoundsException ex) {
		TraceEjb.debugPool("JBeanHome.reuseEJBCtx: Empty List");
		return ec;
    	}
	return ec;
    }

    /**
     * Return an instance/context to the pool
     */
    public void freeEJBCtx(JContext ec) {
	    pooledinstances.add(ec); // don't need to push just need to add. 
	                             //It doesn't matter what context you get as long as you get one.
	    if (TraceEjb.isDebugPool) 
		TraceEjb.debugPool("JBeanHome.freeEJBCtx: size=" + pooledinstances.size());
    }

    public int poolSize() {
	return pooledinstances.size();
    }

    /**
     * try to reduce the pool size (for scalability)
     */
    public int trimEJBPoolSize(int value) {
	int ret = 0;
	synchronized (pooledinstances) {
	    while (pooledinstances.size() > value) {
		ListIterator i = pooledinstances.listIterator();
		if (i.hasNext()) {
		    i.next();
		    i.remove();
		}
		ret++;
	    }
	}
	return ret;
    }

}
