 /*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 *  Christophe Ney: for making easier Enhydra integration
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.security.Identity;
import java.security.Principal;
import java.security.Guard;
import java.util.Properties;
import javax.ejb.EJBContext;
import javax.ejb.EJBHome;
import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ServerSession;
import javax.jms.Session;
import javax.jms.XASession;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRolledbackException;
import javax.transaction.UserTransaction;
import javax.transaction.xa.XAResource;
import org.objectweb.jonas_ejb.container.RequestCtx;
import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Generic interposed class for Message Driven Beans
 * This class presents these interfaces, depending on object reached:
 * ServerSession interface to the ServerSessionPool
 * MessageDrivenContext interface to the bean instance
 * MessageListener interface to the JMS Session
 * Runnable interface to the ThreadPool
 */
public class JBeanMessageDriven implements MessageListener, ServerSession, Runnable, MessageDrivenContext {

    protected Session sess = null;
    protected JMdbFactory mdbfact = null;
    protected MessageDrivenBean mdb = null;
    protected int txattr;	// TX_NOT_SUPPORTED, TX_REQUIRED  or TX_BEAN_MANAGED
    protected TransactionManager tm = null;
    protected ThreadPool thpool = null;
    
    // ----------------------------------------------------------------------------
    // Constructors
    // ----------------------------------------------------------------------------

    public JBeanMessageDriven(JMdbFactory mdbfact, Session sess, MessageDrivenBean mdb, ThreadPool thpool) {
	this.mdbfact = mdbfact;
	this.sess = sess;
	this.mdb = mdb;
	this.thpool = thpool;
	// keep these locally for efficiency.
	txattr = mdbfact.getTRansactionAttribute();
	tm = mdbfact.getTransactionManager();
    }

    // ----------------------------------------------------------------------
    // javax.jms.MessageListener implementation
    // ----------------------------------------------------------------------

    /**
     * A message has been received by the Session.
     * Basically, we have to do: preinvoke + onMessage + postInvoke.
     */
    public synchronized void onMessage(Message message) {
	TraceEjb.debugJms("JBeanMessageDriven: onMessage");
	
	RequestCtx rctx = preinvoke();
	if (rctx == null) {
	    TraceEjb.error("onMessage: preinvoke failed");
	    return;
	}
	try {
	    ((MessageListener)mdb).onMessage(message);
	} catch (RuntimeException e) {
	    rctx.setSysExc(e);
	    TraceEjb.error("onMessage: RuntimeException thrown by an enterprise Bean", e);
	} catch (Error e) {
	    rctx.setSysExc(e);
	    TraceEjb.error("onMessage: Error thrown by an enterprise Bean", e);
	} finally {
	    postinvoke(rctx);
	}
    }

    /**
     * This method makes the container's job before the invocation of the onMessage method.
     * Here are treated transaction attributes.
     * A message driven bean may be use bean-managed or container-managed 
     * demarcation for transactions. In case of container-managed only the NotSupported and Required
     * transaction attributes may be used.
     * There is no permission checking on a MessageDriven bean because there is no principal.
     * @return    the context of the RequestContext
     *            null if exception caught during preinvoke
     */
    private RequestCtx preinvoke() {
	TraceEjb.debugJms("JBeanMessageDriven: preinvoke");

	// Build a RequestCtx to save information about this request
	// until the postinvoke time.
	RequestCtx rctx = new RequestCtx(txattr);

	// Set bean context for JNDI
	rctx.setOldJNDICtx(mdbfact.setComponentContext());
	rctx.setEJBContext(this);

	// Start a Transaction if required by the deployment descriptor.	
	Transaction currtx = null;
	boolean mustCommit = false;
	if (txattr == MethodDesc.TX_REQUIRED ) {
	    try {
		tm.begin();
		currtx = tm.getTransaction();
		// Enlist XAResource : must be aware of a possible rollback
		if (sess instanceof XASession) {
		    currtx.enlistResource(((XASession)sess).getXAResource());
		} else {
		    TraceEjb.error("JBeanMessageDriven : Session should be an XASession ?");
		}
		mustCommit = true;
	    } catch (NotSupportedException e) {
		TraceEjb.error("JBeanMessageDriven : Cannot start a transaction: NotSupportedException");
		return null;
	    } catch (RollbackException e) {
		TraceEjb.error("JBeanMessageDriven : Rollback on enlistResource");
		return null;
	    } catch (SystemException e) {
		TraceEjb.error("JBeanMessageDriven: Cannot start a transaction:\n", e);
		return null;
	    }
	} else {
	    // Bean  managed transaction or Container managed transaction with NotSupported
	    // Nothing to do
	    TraceEjb.debugJms("JBeanMessageDriven: preinvoke no transaction");
	}

	// update RequestCtx
	rctx.setTransactions(mustCommit, null, currtx);
	return rctx;
    }

    /**
     * called from postinvoke. Should be merged with equivalent code in JBean.
     */
    private void checkTransactionAndException(RequestCtx rctx) throws RemoteException {
	// Check reference to EJBContext
	EJBContext ctx = rctx.getEJBContext();
	if (ctx == null) {
	    TraceEjb.error("null EJBContext in RequestCtx!");
	}
	// Commit the transaction started in preInvoke
	if (rctx.mustCommit()) {
	    // First delist XAResource
	    try {
		rctx.getCurrentTx().delistResource(((XASession)sess).getXAResource(), 
						   XAResource.TMSUCCESS);
	    } catch (IllegalStateException e) {
		TraceEjb.error("JBeanMEssageDriven postinvoke: IllegalStateException");
	    } catch (SystemException e) {
		TraceEjb.error("JBeanMEssageDriven postinvoke: SystemException: "+e);
	    }
	    if (rctx.getSysExc() != null) {
		TraceEjb.error("System Exception raised by request:", rctx.getSysExc());
		//rollback the transaction
		try {
		    tm.rollback();
		} catch (Exception e) {
		    TraceEjb.error("Exception during rollback:", e);
		}
	    } else {
		// try to commit the transaction, even in case of appl. Exc.
		// If rollbackonly, do not throw any runtime exception to the client.
		boolean rollb = ctx.getRollbackOnly();
		try {
		    if (rollb) {
			TraceEjb.verbose("postinvoke: Rolling back transaction");
			tm.rollback();
		    } else {
			TraceEjb.debugInvoke("postinvoke: Committing transaction");
			tm.commit();
		    }
		} catch (RollbackException e) {
		    // Probably a pb while writing data on database.
		    throw new RemoteException("Container cannot commit a transaction", e);
		} catch (Exception e) {
		    TraceEjb.error("postinvoke: ", e);
		    throw new RemoteException("Container exception", e);
		}
	    }
	}

	// We got a system Exception in business method:
	// - log exception
	// - discard instance
	// - set client transaction rollback only
	// - throw remote exception
	if (rctx.getSysExc() != null) {
	    // Log system exception
	    TraceEjb.error("System Exception in business method:", rctx.getSysExc());

	    // If client transaction: set it rollback only and throw TransactionRolledBackException
	    Transaction currentTx = null;
	    try {
		currentTx = tm.getTransaction();
		if (currentTx != null) {
		    currentTx.setRollbackOnly();
		    throw new TransactionRolledbackException("System exception in business method");
		}
	    } catch (SystemException e) {
		TraceEjb.error("postinvoke: cannot set rollback only current tx:", e);
	    }
	    return;
	}

	// Resume client transaction suspended in preInvoke
	Transaction tx = rctx.getClientTx();
	if (tx != null) {
	    try {
		tm.resume(tx);
	    } catch (SystemException e) {
		TraceEjb.error("postinvoke: Cannot resume transaction");
	    }
	}
    }

    /**
     * Verify that a bean transaction started in onMessage has been terminated
     * called from postinvoke
     * this code may be used by stateless session bean too.
     * @param     rctx context of the current request
     * @return false if a transaction was running
     */
    private boolean checkBeanTransaction(RequestCtx rctx) {

	if (rctx.getTxAttr() == MethodDesc.TX_BEAN_MANAGED) {
	    // tx initiated by the bean must be suspended in any case
	    Transaction tx = null;
	    try {
		tx = tm.suspend();
	    } catch(SystemException e) {
		TraceEjb.error("checkBeanTransaction: cannot suspend transaction:", e);
	    }
	    if (tx != null) {
		// Stateless Session and Message driven beans are not allowed to retain a
		// transaction between 2 requests.
		try {
		    TraceEjb.error("Rollback the bean managed transaction");
		    tx.rollback();
		} catch (SystemException e) {
		    TraceEjb.error("postinvoke: cannot rollback transaction:", e);
		}
		return false;
	    }
	}
	return true;
    }
    

    /**
     * This method is called after the invocation of the onMessage method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    private void postinvoke(RequestCtx rctx) {
	TraceEjb.debugJms("JBeanMessageDriven: postinvoke");
	try {
	    if (!checkBeanTransaction(rctx)) {
		// we must discard the instance
	    }
	    checkTransactionAndException(rctx);
	    mdbfact.resetComponentContext(rctx.getOldJNDICtx());
	} catch (RemoteException ex) {
	    // Il ne faut pas remonter une exception au client
	}
    }

    // ----------------------------------------------------------------------
    // javax.jms.ServerSession implementation
    // ----------------------------------------------------------------------

    /**
     * Return the ServerSession's Session. 
     * This must be a Session created by the same Connection which will be
     * dispatching messages to it. 
     * The provider will assign one or more messages to the Session and then call
     * start on the ServerSession.
     * @return the server session's session.
     *
     * @exception JMSException - if a JMS fails to get associated session 
     * for this serverSession due to some internal error.
     */
    public Session getSession() throws JMSException {
	TraceEjb.debugJms("JBeanMessageDriven: getSession");
	return sess;
    }

    /**
     * Cause the session's run method to be called to process messages that 
     * were just assigned to it.
     *
     * @exception JMSException - if a JMS fails to start the server session 
     * to process messages.
     */
    public void start() throws JMSException {
	TraceEjb.debugJms("JBeanMessageDriven: start");
	thpool.addRunnable(this);
    }

    // ----------------------------------------------------------------------
    // Runnable implementation
    // ----------------------------------------------------------------------

    /**
     * Process messages by calling run method on Session. When finished,
     * return the object in the pool.
     */
    public void run() {
	TraceEjb.debugJms("JBeanMessageDriven: run");

	sess.run();
	mdbfact.toThePool(this);
    }
 
    // ----------------------------------------------------------------------
    // javax.ejb.MessageDrivenContext implementation
    // ----------------------------------------------------------------------

    private static final String DISALLOWED_MSG=" is disallowed in a message driven bean";

   /**
     * Obtains the java.security.Identity of the caller. 
     * DEPRECATED
     * disallowed in messagedriven bean method because there is no security context
     * @exception java.lang.IllegalStateException
     */
    public Identity getCallerIdentity() {
	TraceEjb.error("getCallerIdentity()"+ DISALLOWED_MSG);
	throw new IllegalStateException ("getCallerIdentity()"+DISALLOWED_MSG);
    }

    /**
     * Obtain the java.security.Principal that identifies the caller.
     * throws a java.lang.IllegalStateException for message driven bean because there is no security context available 
     * (EJB v2.0, chapter 14.5.1)
     * @exception java.lang.IllegalStateException
     */
    public Principal getCallerPrincipal() {
	TraceEjb.error("getCallerPrincipal()"+ DISALLOWED_MSG);
	throw new IllegalStateException ("getCallerPrincipal()"+DISALLOWED_MSG);
    }
    
    /**
     * Deprecated. 
     * throws a java.lang.IllegalStateException for message driven bean because there is no security context available 
     * (EJB v2.0, chapter 14.5.1)* @exception java.lang.IllegalStateException
     */
    public boolean isCallerInRole(Identity role) {
	TraceEjb.error("isCallerInRole()"+DISALLOWED_MSG);
	throw new IllegalStateException ("isCallerInRole()"+DISALLOWED_MSG);
    }

    /**
     * Test if the caller has a given role.
     * throws a java.lang.IllegalStateException for message driven bean because there is no security context available 
     * (EJB v2.0, chapter 14.5.1)* @exception java.lang.IllegalStateException
     */
    public boolean isCallerInRole(java.lang.String roleLink) {
	TraceEjb.error("isCallerInRole()"+DISALLOWED_MSG);
	throw new IllegalStateException ("isCallerInRole()"+DISALLOWED_MSG);
    }

    /** 
     * Marks the current transaction for rollback.
     * Should be used only if the instance is associated with a transaction
     * Must throw java.lang.IllegalStateException if the instance is not associated with a transaction
     *
     */
    public void setRollbackOnly() {

	TraceEjb.debugCtx("JContextMessageDriven.setRollbackOnly() ");
	try {
	    tm.setRollbackOnly();
	} catch (IllegalStateException e) {
	    TraceEjb.error("Current thread not associated with transaction");
	    throw e;
	} catch (SystemException e) {
	    TraceEjb.error("setRollbackOnly unexpected exception:", e);
	}
    }
   /** 
     * Tests if the transaction has been marked for rollback only.
     */
    public boolean getRollbackOnly() {
	TraceEjb.debugCtx("JContextMessageDriven.getRollbackOnly() ");

	try {
	    if (tm.getTransaction() != null) {
		switch (tm.getStatus()) {
		case Status.STATUS_MARKED_ROLLBACK:
		case Status.STATUS_ROLLEDBACK:
		case Status.STATUS_ROLLING_BACK:
		    return true;
		default:
		    return false;
		} 
	    } else {
		TraceEjb.error("IllegalState Exception the bean is not associated in a transaction");
		throw new IllegalStateException("the message driven bean is not associated in a transaction");
	    }
	} catch (SystemException e) {
	    TraceEjb.error("Cannot get status:", e);
	    return false;
	}
    } 
 

    /**
     * Is disallowed. There is no home for message driven bean.
     */
    public EJBHome getEJBHome() {
	TraceEjb.error(" getEJBHome()"+DISALLOWED_MSG);
	throw new IllegalStateException ("getEJBHome()"+DISALLOWED_MSG);
    }

    /**
     * Deprecated. 
     */
    public Properties getEnvironment() {
	TraceEjb.error("getEnvironment is deprecated use Use the JNDI naming context java:comp/env");
	return new java.util.Properties();
    }

    /**
     * Obtains the transaction demarcation interface. 
     * @return The UserTransaction interface that the enterprise bean instance 
     *         can use for transaction demarcation.  
     * @exception IllegalStateException: Thrown if the instance container does 
     *            not make the UserTransaction interface available to the 
     *            instance (only the enterprise beans with the TX_BEAN_MANAGED 
     *            transaction attribute are allowed to use the UserTransaction 
     *            interface). 
     */
    public UserTransaction getUserTransaction() throws IllegalStateException {

	TraceEjb.debugCtx("JContextMessageDriven.getUserTransaction() ");

	if (!mdbfact.isTxBeanManaged()) {
	    throw  new IllegalStateException("This bean is not allowed to use UserTransaction interface");
	}
	return (UserTransaction) tm;
    }

}
