/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * Jun Inamori: Unreferenced interface for  poolling of EJBObjects - JOnAS 2.1
 * Markus Karg: Novell port - jonas-2-2-4
 * Dean Jennings : Optimization
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.Unreferenced;
import java.util.*;
import javax.ejb.Handle;
import javax.ejb.RemoveException;
import javax.transaction.Transaction;
import javax.transaction.SystemException;

import org.objectweb.jonas.thds.ThreadData;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_timer.TimerEvent;
import org.objectweb.jonas_timer.TimerEventListener;
import org.objectweb.jonas_timer.TimerManager;

/**
 * JBeanSession is the base class for all Session Beans
 */
public abstract class JBeanSession extends JBean implements Unreferenced, TimerEventListener {

    /**
     * Saved connectionList for this instance. The real connectionList
     * is maintained in a ThreadLocal variable in ThreadData.
     * Used only for session stateful methods (?), if they keep
     * connection along several calls, not always in same thread.
     * This list must not be shared between all instances.
     * This info is transient because JDBC connection have not to be
     * saved on disk at passivation time : They should be closed in
     * ejbPassivate().
     */
    protected List connectionList = Collections.synchronizedList(new ArrayList());


    /**
     * unique SessionContext for this Session bean
     * @serial
     */
    protected JContextSession sctx = null;

    /**
     * true if remove has been called on this instance.
     * actual remove will be done in postinvoke of remove method.
     */
    protected boolean isMarkedRemoved = false;

    protected TimerEvent mytimer = null;

    /**
     * Constructor
     */
    public JBeanSession(JSessionHome home) throws RemoteException {
	super(home);
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    //
    // getEJBHome	in JBean
    // getHandle	ok
    // getPrimaryKey	ok
    // isIdentical	in subclasses
    // remove		generated
    // --------------------------------------------------------------------------

    /**
     * Obtains a handle for the EJB object. The handle can be used at later time to re-obtain
     * a reference to the EJB object, possibly in a different JVM.
     *
     * @return    A handle for the EJB object. 
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public Handle getHandle() throws RemoteException {
	TraceEjb.debugInvoke("JBeanSession.getHandle");

	if (handleEJBObject == null) {
	    handleEJBObject = new SessionHandleImpl(this);
	}
	return (Handle) handleEJBObject;
    }

    /**
     * Obtains the primary key of the EJB Object
     * must throw RemoteException for sessionBean
     * (5.5 Spec EJB1.1)
     */
    public Object getPrimaryKey() throws RemoteException {
	TraceEjb.error("getPrimaryKey: No primary Key for Session Bean");
	throw new RemoteException("No primary Key for Session Bean");
    }

    // --------------------------------------------------------------------------
    // Methods called from the code generated by GenIC
    // Be sure to change GenIC accordingly before changing these methods.
    // --------------------------------------------------------------------------

   /**
     * This method makes the container's job before the invocation of the business method.
     * Here are treated transaction attributes in order to execute the business
     * method in the correct transactional context. A new transaction may be created.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     * @return    the request context
     */
    public RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSession.preinvoke");

	// security controls
	home.checkSecurity(methodSignature);

	// Build a RequestCtx to save information about this request
	// until the postinvoke time.
	RequestCtx rctx = new RequestCtx(txattr);

	// Set bean context for JNDI
	rctx.setOldJNDICtx(home.setComponentContext());

	// Associates the connection list of this bean to the current thread.
	List oldList = ThreadData.getInstance().setConnectionList(connectionList);
	// save old value in RequestContext to reset it at postinvoke.
	rctx.setOldConnectionList(oldList);

	// must init first SessionContext (sctx)
	rctx.setEJBContext(getSessionContext());
	if (sctx == null) {
	    // If null, a nullPointerException would be thrown in business method:
	    // Better to reject now, with an explicit error message.
	    TraceEjb.debugInvoke("JBeanSession.preinvoke: sctx is null");
	    throw new RemoteException("preinvoke failed: No Context");
	}

	if (txattr == MethodDesc.TX_BEAN_MANAGED) {
	    // Bean managed transactions
	    // Suspend client transaction if exists
	    try {
		Transaction cltx = tm.suspend();
		if (cltx != null) {
		    TraceEjb.debugInvoke("JBeanSession.preinvoke: suspending tx");
		    rctx.setClientTx(cltx);
		}
	    } catch (SystemException e) {
		TraceEjb.error("Exception in suspend:", e);
		throw new RemoteException("preinvoke failed");
	    }
	} else {
	    // Container managed transactions
	    container.checkTransactionIn(rctx);
	}

	return rctx;
    }

    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSession.postinvoke");

	super.postinvoke(rctx);
	
	// Reset bean context for JNDI
	home.naming.setComponentContext(rctx.getOldJNDICtx());

	// Save connectionList for next requests and reset the ThreadLocal
	// with value before this request.
	List oldList = rctx.getOldConnectionList();
	connectionList = ThreadData.getInstance().setConnectionList(oldList);
    }

    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------

    /*
     * set the connection list (used for create only)
     */
    public void setMyConnectionList(List list) {
	connectionList = list;
    }

    /*
     * EJBObject accessible again
     */
    public synchronized boolean reUseObject() {
	if (super.reUseObject() == true) {
	    isMarkedRemoved = false;
	    return true;
	} else {
	    return false;
	}
    }

    /*
     * This EJBObject is no more used : Return it to the pool.
     * This may happen when bean is no more referenced, or at remove.
     */
    protected void releaseEJB() {
	TraceEjb.debugInvoke("JBeanSession.releaseEJB");
	if (sctx != null) {
	    TraceEjb.error("JBeanSession.releaseEJB: ctx not null!");
	}
	noMoreInUse();
	((JSessionHome)home).freeEJBObject(this);
    }

    /*
     * remove all about this EJB
     * called by SessionHome.unregister
     */
    public abstract void removeInstance();

    /**
     * returns the unique SessionContext.
     */
    public abstract JContextSession getSessionContext() throws RemoteException;

    public void startTimer(int timeout) {
	mytimer = TimerManager.getInstance().addTimer(this, timeout, null, false);
    }

    public void stopTimer() {
	if (mytimer != null) {
	    mytimer.stop();
	}
    }
}

