/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.Unreferenced;
import java.util.Vector;
import javax.ejb.EJBObject;
import javax.ejb.RemoveException;
import javax.naming.Context;
import javax.transaction.Transaction;
import javax.transaction.SystemException;

import org.objectweb.jonas_ejb.deployment.api.MethodDesc;

/**
 * This class is a specification of the JBeanSession class for
 * bean session stateful. It is instenciated by the generated code
 * during creation. It is abstract, because remove method is generated.
 */
public abstract class JBeanSessionSf extends JBeanSession {

    /**
     * true is the corresponding instance implements SessionSynchronization interface
     * @serial
     */
    protected boolean isSessionSynchro;

    /**
     * Ident of the Context associated to this EJBObject (for passivation)
     * @serial
     */
    private int ctxId;

    /**
     * Constructor
     */
    public JBeanSessionSf(JSessionHome home, boolean sync) throws RemoteException {
	super(home);
	isSessionSynchro = sync;
	ctxId = -1;
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    // --------------------------------------------------------------------------

    /**
     * Tests if a given EJB is identical to the invoked EJB object.
     *
     * @param     EJBObject obj - An object to test for identity with the invoked object. 
     * @return    True if the given EJB object is identical to the invoked object. 
     * @throws RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public boolean isIdentical(EJBObject obj) throws RemoteException {

	boolean ret = false;
	if (obj != null) {
	    // In case of Stateful session bean, compares the two EJBObjects
	    ret = obj.equals(this);
	}
	return ret;
    }

    // -------------------------------------------------------------------
    // Unreferenced implementation
    // -------------------------------------------------------------------

    /*
     * Called by the RMI runtime sometime after the runtime determines that
     * the list of clients referencing the remote object becomes empty.
     */
    public synchronized void unreferenced() {

	TraceEjb.debug("JBeanSessionSf.unreferenced");

	// If the container has been removed by JonasAdmin: nothing to do.
	if (home.isUnregistered()) {
	    return;
	}

	JContextSessionSf sc = (JContextSessionSf) sctx;
	if (sc == null) {
	    // unreferenced is called even if the object is already unexported.
	    // so, in that case, just return.
 	    TraceEjb.debug("unreferenced: null session context");
	    return;
	} else {
	    if (sc.inTransaction()) {
		// If we are inside a Tx, don't touch it.
		TraceEjb.debug("JBeanSessionSf.unreferenced: Session Bean inside Tx");
		return;
	    }
	}

	// Force a remove of the session bean
	forceSessionRemove();
    }

    // ===============================================================
    // TimerEventListener implementation
    // ===============================================================

    /**
     * The session timeout has expired
     */
    public synchronized void timeoutExpired(Object arg) {
	TraceEjb.debug("stateful session timeout expired");
	// No need to stop it now.
	mytimer = null;

	// If the container has been removed by JonasAdmin: nothing to do.
	if (home.isUnregistered()) {
	    return;
	}

	// If this session is involved in a transaction, we cannot remove
	// it now. Just set a flag that will be checked in AfterCompletion.
	JContextSessionSf sc = (JContextSessionSf) sctx;
	if (sc == null) {
	    TraceEjb.verbose("timeoutExpired: null session context");
	} else {
	    if (sc.inTransaction()) {
		TraceEjb.verbose("timeoutExpired: Stateful Session Bean inside Tx");
		sc.setTimeoutFlag();
		return;
	    }
	}

	// Force a remove of the session bean
	forceSessionRemove();
    }

    // --------------------------------------------------------------------------
    // Methods called from the code generated by GenIC
    // Be sure to change GenIC accordingly before changing these methods.
    // --------------------------------------------------------------------------

   /**
     * This method makes the container's job before the invocation of the business method.
     * Here are treated transaction attributes in order to execute the business.
     *  method in the correct transactional context. A new transaction may be created.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     * @return    the context of the SessionBean 
     *
     */
    public RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSf.preinvoke");
	try {
	    RequestCtx rctx = super.preinvoke(txattr, methodSignature);
	    JContextSessionSf sc = (JContextSessionSf) sctx;

	    if (txattr == MethodDesc.TX_BEAN_MANAGED) {
		// Resume transaction started previously in the bean
		try {
		    if (sc.beanAssociatedTransaction != null) {
			TraceEjb.debugInvoke("JBeanSessionSf.preinvoke: resuming tx");
			tm.resume(sc.beanAssociatedTransaction);
		    } 
		} catch (SystemException e) {
		    TraceEjb.error("Exception in resume:", e);
		    // We should resume cltx here ???
		    throw new RemoteException("preinvoke failed");
		}
	    } else {
		sc.isInContainerTransaction = rctx.mustCommit();
	    }

	    // Synchronization 
	    // Only stateful Session bean with container-managed transaction may implements
	    // the SessionSynchronization interface.
	    // 
	    if (isSessionSynchro) {
		sc.considerRequest();
	    }
	    return rctx;
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSf.postinvoke");
	try {
	    JContextSessionSf sc = (JContextSessionSf) sctx;

	    if (rctx.getTxAttr() == MethodDesc.TX_BEAN_MANAGED) {
		// tx initiated by the bean must be suspended in any case
		Transaction tx = null;
		try {
		    tx = tm.suspend();
		} catch(SystemException e) {
		    TraceEjb.error("postinvoke: cannot suspend transaction:", e);
		}
		// keep it for resume in next request
		// or reinit if committed.
		sc.beanAssociatedTransaction = tx;
	    }

	    // common part
	    super.postinvoke(rctx);
	
	    // If this session bean has been removed:
	    // - return the EJBContext in the pool
	    // - return the EJBObject in the pool
	    if (isMarkedRemoved) {
		releaseSessionContext();
		releaseEJB();
	    }
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * remove this EJBObject
     * sctx has been allocated in preinvoke.
     */
    public void removeEJB() throws RemoveException {
	TraceEjb.debugCtx("JBeanSessionSf.removeEJB");

	if (sctx == null) {
	    TraceEjb.error("removeEJB: No SessonContext!");
	    return;
	}
	JContextSessionSf sc = (JContextSessionSf) sctx;
	if (sc.inTransaction()) {
	    TraceEjb.error("Can't remove session in a transaction"); 
	    throw new RemoveException("Cannot remove a Session bean participating in a transaction");
	}
	isMarkedRemoved = true;
	sctx.removeMyInstance();
    }

    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------

    /**
     * remove this EJBObject, because unreferenced, timeout, or remove.
     * - assumes that this session is outside any transaction.
     */
    public void forceSessionRemove() {
	TraceEjb.debugCtx("JBeanSessionSf.forceSessionRemove");

	if (sctx != null) {
	    try {
		sctx.removeMyInstance(); // call ejbRemove on instance.
	    } catch (RemoveException e) {
		TraceEjb.error("exception on ejbRemove:"+e);
	    } 
	    releaseSessionContext(); // Release Context/Instance
	}
	releaseEJB();		// Release EJBObject
    }

    /**
     * removeInstance - called when the container is unloaded.
     */
    public synchronized void removeInstance() {
	JContextSessionSf sc = (JContextSessionSf) sctx;
	// If we are inside a Tx, don't touch it.
	if (sc != null && sc.inTransaction()) {
	    TraceEjb.verbose("removeInstance: Session Bean inside Tx");
	    return;
	}

	// Force a remove of the session bean
	try {
	    // call ejbRemove on instance.
	    if (sctx != null) {
		sctx.removeMyInstance();
	    }
	    // Release Context/Instance
	    releaseSessionContext();
	} catch (Exception ex){
	    TraceEjb.error("removeInstance unexpected exception : "+ex);
	}

	// EJBObject no longer usable
	noMoreInUse();
    }

    /**
     * returns the unique SessionContext. 
     * If not yet associated, find one in the pool.
     */
    public JContextSession getSessionContext() throws RemoteException {

	if (sctx == null) {
	    // Stateful: If no session context, it may have been passivated
	    // ask SessionPool to retrieve it
	    if (ctxId == -1) {
		TraceEjb.error("JBeanSessionSf: getSessionContext: no ctxId");
		return null;
	    }
	    sctx = container.statefulSessionPool.getContext(ctxId);
	}
	return sctx;
    }


    /**
     * Creates a new SessionContext
     * this is called only when a new Session bean is created.
     */
    public JContextSession getNewSessionContext() throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSf: getNewSessionContext");

	// No context should be associated already
	if (sctx != null) {
	    TraceEjb.error("JBeanSessionSf: getNewSessionContext: already a context!");
	}

	// Always creates a new one (specs EJB: newInstance must be called)
	JSessionHome h = (JSessionHome) home;
	sctx = h.createSessionContext(this, isSessionSynchro);

	// keep the Context in the Pool
	JContextSessionSf sc = (JContextSessionSf) sctx;
	ctxId = container.statefulSessionPool.addContext(sc);
	return sctx;
    }

    /**
     * Release the Session Context
     * Don't pool Stateful Session Bean Instances.
     */
    public void releaseSessionContext() {
	TraceEjb.debugInvoke("JBeanSessionSf: releaseSessionContext");

	container.statefulSessionPool.releaseContext(ctxId);
	ctxId = -1;
	if (sctx == null) {
	    TraceEjb.verbose("JBeanSessionSf: releaseSessionContext sctx=null");
	    return;
	}
	sctx = null;
    }

}
