/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.Unreferenced;
import java.util.Vector;
import javax.ejb.EJBObject;
import javax.ejb.RemoveException;
import javax.naming.Context;
import javax.transaction.Transaction;
import javax.transaction.SystemException;

import org.objectweb.jonas_ejb.deployment.api.MethodDesc;

/**
 * This class is a specification of the JBeanSession class for
 * bean session stateless. It is instenciated by the generated code
 * during creation. It is abstract, because remove method is generated.
 */
public abstract class JBeanSessionSl extends JBeanSession {

    /**
     * Constructor
     */
    public JBeanSessionSl(JSessionHome home) throws RemoteException {
	super(home);
    }

    // --------------------------------------------------------------------------
    // EJBObject implementation
    // --------------------------------------------------------------------------

    /**
     * Tests if a given EJB is identical to the invoked EJB object.
     *
     * @param     EJBObject obj - An object to test for identity with the invoked object. 
     * @return    True if the given EJB object is identical to the invoked object. 
     * @throws RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public boolean isIdentical(EJBObject obj) throws RemoteException {

	boolean ret = false;
	if (obj != null) {
	    // In case of Stateless session bean, compares the EJBHome associated to
	    // the two EJBObjects
	    ret = obj.getEJBHome().getEJBMetaData().getHomeInterfaceClass().getName().equals(this.getEJBHome().getEJBMetaData().getHomeInterfaceClass().getName());
	}
	return ret;
    }

    // -------------------------------------------------------------------
    // Unreferenced implementation
    // -------------------------------------------------------------------

    /*
     * Called by the RMI runtime sometime after the runtime determines that
     * the list of clients referencing the remote object becomes empty.
     */
    public synchronized void unreferenced() {

	TraceEjb.debug("JBeanSessionSl.unreferenced");

	// If the container has been removed by JonasAdmin: nothing to do.
	if (home.isUnregistered()) {
	    return;
	}

	// Should not be in use now.
	if (sctx != null) {
	    TraceEjb.verbose("Unreferenced: Stateless Session Bean in use!");
	    return;
	}

	// Force a remove of the session bean
	// No need to free context/instance because it's already done.
	releaseEJB();
    }

    // ===============================================================
    // TimerEventListener implementation
    // ===============================================================

    /**
     * The session timeout has expired
     */
    public synchronized void timeoutExpired(Object arg) {
	TraceEjb.verbose("stateless session timeout expired");
	// No need to stop it now.
	mytimer = null;

	// If the container has been removed by JonasAdmin: nothing to do.
	if (home.isUnregistered()) {
	    return;
	}

	// Session should never be used because this method is synchronized
	if (sctx != null) {
	    TraceEjb.verbose("timeoutExpired: Stateless Session Bean in use!");
	    return;
	}

	// Force a remove of the session bean
	// No need to free context/instance because it's already done.
	releaseEJB();
    }

    // --------------------------------------------------------------------------
    // Methods called from the code generated by GenIC
    // Be sure to change GenIC accordingly before changing these methods.
    // --------------------------------------------------------------------------

   /**
     * This method makes the container's job before the invocation of the business method.
     * Here are treated transaction attributes in order to execute the business.
     *  method in the correct transactional context. A new transaction may be created.
     *
     * @param     transAttribute Transaction attribute of the method that will be invoked.
     * @exception RemoteException: Thrown when the method failed due to a system-level failure. 
     *
     * @return    the context of the SessionBean 
     *
     */
    public RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSl.preinvoke");
	try {
	    return super.preinvoke(txattr, methodSignature);
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * This method is called after the invocation of the business method.
     *
     * @param     rctx context of the current request
     * @exception RemoteException: Thrown when the method failed due to a system-level failure.
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSl.postinvoke");
	try {
	    if (rctx.getTxAttr() == MethodDesc.TX_BEAN_MANAGED) {
		// tx initiated by the bean must be suspended in any case
		Transaction tx = null;
		try {
		    tx = tm.suspend();
		} catch(SystemException e) {
		    TraceEjb.error("postinvoke: cannot suspend transaction:", e);
		}
		if (tx != null) {
		    // Stateless Session beans are not allowed to retain a
		    // transaction between 2 requests.
		    try {
			TraceEjb.error("Rollback stateless session bean transaction");
			tx.rollback();
		    } catch (SystemException e) {
			TraceEjb.error("postinvoke: cannot rollback transaction:", e);
		    }
		}
	    }

	    // common part
	    super.postinvoke(rctx);
	    
	    // In case of stateless session bean:
	    // return the EJBContext in the pool after every request
	    releaseSessionContext();
	    
	    // If this session bean has been removed:
	    // - return the EJBObject in the pool
	    if (isMarkedRemoved) {
		releaseEJB();
	    }
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * remove this EJBObject
     * sctx has been allocated in preinvoke.
     */
    public void removeEJB() throws RemoveException {
	TraceEjb.debugCtx("JBeanSessionSl.removeEJB");
	if (sctx == null) {
	    TraceEjb.error("removeEJB: No SessonContext!");
	}
	isMarkedRemoved = true;
    }

    // --------------------------------------------------------------------------
    // Other methods
    // --------------------------------------------------------------------------

    /**
     * removeInstance - called when the container is unloaded.
     */
    public synchronized void removeInstance() {
	// If busy, don't touch it.
	if (sctx != null) {
	    TraceEjb.verbose("removeInstance: Session Bean busy!");
	    return;
	}

	// Force a remove of the session bean
	noMoreInUse();
    }

    /**
     * returns the unique SessionContext. 
     // If not yet associated, find one in the pool (general case if preinvoke)
     */
    public JContextSession getSessionContext() throws RemoteException {

	if (sctx == null) {
	    getNewSessionContext();
	}
	return sctx;
    }

    /**
     * Creates a new SessionContext
     * this is called only when a new Session bean is created.
     */
    public JContextSession getNewSessionContext() throws RemoteException {
	TraceEjb.debugInvoke("JBeanSessionSl: getNewSessionContext");

	// No context should be associated already
	if (sctx != null) {
	    TraceEjb.error("JBeanSessionSl: getNewSessionContext: already a context!");
	}

	// Try to get a Context from the HomePool.
	// If no Context available, creates a new one
	JSessionHome h = (JSessionHome) home;
	sctx = (JContextSession) h.reuseEJBCtx();
	if (sctx == null) {
	    sctx = h.createSessionContext(this, false);
	} else {
	    sctx.initSessionContext(this);
	}
	return sctx;
    }


    /**
     * Release the Session Context and put it back in the pool
     */
    public void releaseSessionContext() {
	if (sctx == null) {
	    TraceEjb.verbose("JBeanSessionSl: releaseSessionContext sctx=null");
	} else {
	    TraceEjb.debugInvoke("JBeanSessionSl: releaseSessionContext");
	    home.freeEJBCtx(sctx);
	    sctx = null;
	}
    }

}
