/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): Jeff Mesnil
 *   Christophe Ney: for making easier Enhydra integration
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.RemoteException;
import java.security.Guard;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.Vector;
import java.util.Properties;
import javax.ejb.EJBException;
import javax.naming.Context;
import javax.naming.LinkRef;
import javax.naming.StringRefAddr;
import javax.naming.NamingException;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.TransactionRequiredException;
import org.objectweb.common.Env;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_ejb.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityBmpDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmpDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmpDesc;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_jms.api.JmsManager;
import org.enhydra.naming.ContainerNaming;

/**
 * This class represents an EJB container.
 *
 * A container is where an enterprise Bean object lives. 
 * Several EJB can be installed in a single container.
 * In a  JOnAS container are installed all beans from the same ejb-jar file.
 * 
 * For each EJB installed, the container provides a home interface and
 * makes this interface available in the JNDI name space.
 */
public class JContainerImpl implements JContainer, JContainerImplMBean {

    private Swapper swapper;
    private ClassLoader loader = null;
    private ContainerNaming naming = null;
    private TransactionManager tm = null;
    private JmsManager jms = null;

    public String containerName; // Container name 
    public String fileName; // Container file name x.jar or x.xml
    public Vector entityList = new Vector();
    public Vector sessionList = new Vector();
    public Vector mdbList = new Vector();
    private Guard methodGuard = null;
    private Guard roleGuard = null;
    private PrincipalFactory principalFactory = null;

    public StatefulSessionPool statefulSessionPool;

    // Pool of threads for Message Driven Beans
    // We use only 1 pool for all containers.
    private static final int DEFAULT_THREAD_POOL_SIZE = 10;
    int threadpoolsize = DEFAULT_THREAD_POOL_SIZE;
    static ThreadPool threadpool = null;

    /**
     * constructor
     */
    public JContainerImpl(String name, String file, ClassLoader ld) {
	containerName = name;
	fileName = file;
	loader = ld;
	swapper = new Swapper(this);
	swapper.start();
	statefulSessionPool = new StatefulSessionPool();
    }

    // ---------------------------------------------------------------
    // JContainer  Interface 
    // ---------------------------------------------------------------

    /**
     * Return an array of bean names
     */
    public String [] listBeanNames() {
	TraceEjb.debugServer(getName(),":listBeanNames");
	String ret[] = new String[entityList.size() + sessionList.size() + mdbList.size()];
	Vector namelist = new Vector();
	JBeanHome bbh = null;
	Enumeration enum = entityList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    namelist.addElement(bbh.getDeploymentDescriptor().getEjbName());
	}
	enum = sessionList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    namelist.addElement(bbh.getDeploymentDescriptor().getEjbName());
	}

	enum = mdbList.elements();
	while (enum.hasMoreElements()) {
	    JMdbFactory jmdbf = (JMdbFactory) enum.nextElement();
	    namelist.addElement(jmdbf.getEjbName());
	}
	namelist.copyInto(ret);
	return ret;
    }

    /**
     * Return the name of the container
     */
    public String getName() {
	return containerName;
    }

    /**
     * set the Thread pool size.
     */
    public void setThreadPoolSize(int sz) {
	threadpoolsize = sz;
    }

    /**
     * Return the file name of the container (.xml or .jar)
     */
    public String getFileName() {
	return fileName;
    }

    public void setMethodGuard (Guard methodGuard) {
	if (TraceEjb.isDebugSecurity)
	    TraceEjb.debugSecurity(getName()+": setMethodGuard("+methodGuard+")");
	this.methodGuard = methodGuard;
    }
    
    public void setRoleGuard (Guard roleGuard) {
	if (TraceEjb.isDebugSecurity)
	    TraceEjb.debugSecurity(getName()+": setRoleGuard("+roleGuard+")");
	this.roleGuard = roleGuard;
    }

    public Guard getMethodGuard() {
	return methodGuard;
    }

    public Guard getRoleGuard() {
	return roleGuard;
    }

    public ClassLoader getClassLoader() {
	if (loader == null) {
	    return ClassLoader.getSystemClassLoader();
	} else {
	    return loader;
	}
    }
    public void setPrincipalFactory(PrincipalFactory pf){
	principalFactory = pf;
    }

    public PrincipalFactory getPrincipalFactory(){
	return principalFactory;
    }


    /**
     * Remove  beans in JOnAS container.
     */
    public void  remove() {
	TraceEjb.debugServer(getName(),": remove");

	// Stop the swapper.
	swapper.stopIt();

	// synchronize all entity beans before removing them.
	// must be done here, not in the swapper.
	syncAll();


	
	// remove all entities
	Enumeration enum = entityList.elements();
	while (enum.hasMoreElements()) {
	    JBeanHome bbh = (JBeanHome) enum.nextElement();
	    try {
		bbh.unregister();
		System.out.println(getName()+": "+ bbh.getEjbName() + " no longer available");
	    } catch (Exception e) {
		TraceEjb.error(getName()+": remove:", e);
	    }
	}
	entityList.removeAllElements();

	// remove all sessions
	enum = sessionList.elements();
	while (enum.hasMoreElements()) {
	    JSessionHome bbh = (JSessionHome) enum.nextElement();
	    try {
		bbh.unregister();
		System.out.println(getName()+": "+ bbh.getEjbName() + " no more available");
	    } catch (Exception e) {
		TraceEjb.error(getName()+": removeBeans:", e);
	    }
	}
	sessionList.removeAllElements();

	// remove all message driven beans
	enum = mdbList.elements();
	while (enum.hasMoreElements()) {
	    JMdbFactory jmdbf = (JMdbFactory) enum.nextElement();
	    try {
		jmdbf.unregister();
		System.out.println(getName()+": "+ jmdbf.getEjbName() + " no more available");
	    } catch (Exception e) {
		TraceEjb.error(getName()+": removeBeans:", e);
	    }
	}
	mdbList.removeAllElements();
	loader = null;

	// run the garbage collector to get free memory
	Runtime.getRuntime().gc();

    }

    /**
     * register an Entity that must be passivated
     */
    public void registerEJB(JBeanEntity ejb) {
	TraceEjb.debugServer(getName(),": registerEJB");

	swapper.addEntity(ejb);
    }

    /**
     * register a Home
     */
    public void registerHome(JEntityHome home) {
	TraceEjb.debugServer(getName(),": registerHome");

	swapper.addHome(home);
    }

    /**
     * Take into account the swapping time for the bean.
     * @args time in seconds  (time = 0 no time out)
     */
    public void setSwapTime(int t) {
	if (TraceEjb.isDebugServer) TraceEjb.debugServer(getName()+": setSwapTime "+t);
	if(t > 0)
	    swapper.setTimeout(1000*t);
    }

    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     * Try to passivate all entity bean instances
     */
    public void syncAll() {
	TraceEjb.debugServer(getName(),": syncAll");
	JEntityHome h = null;
	Enumeration enum = entityList.elements();
	while (enum.hasMoreElements()) {
	    h = (JEntityHome) enum.nextElement();
	    h.passivateAllIH(false);
	}
    }

    /**
     * Process Transaction Attribute before calling a business method
     */
    public void checkTransactionIn(RequestCtx rctx) throws RemoteException {
	TraceEjb.debugServer(getName(),": checkTransactionIn");
	
	boolean mustCommit = false;
	Transaction cltx = null;
	int txattr = rctx.getTxAttr();
	if (TraceEjb.isDebugServer) 
	    TraceEjb.debugServer(getName()+": Transaction Attribute = " + MethodDesc.getTxAttributeName(txattr));

	// First of all, get the current transaction
	Transaction currtx = null;
	try {
	    currtx = tm.getTransaction();
	} catch (SystemException e) {
	    TraceEjb.error(getName()+": SystemException while getting transaction:\n", e);
	}

	// Raises exception if "Never" and we are in a transaction
	if (txattr == MethodDesc.TX_NEVER && currtx != null) {
	    TraceEjb.verbose(getName()+": Never and transaction not null");
	    throw new RemoteException("Never attribute = caller must not be in a transaction");
	}

	// Raises exception if "Mandatory" and we are not in a transaction
	if (txattr == MethodDesc.TX_MANDATORY && currtx == null) {
	    TraceEjb.verbose(getName()+": Mandatory and not in transaction");
	    throw new TransactionRequiredException("Mandatory attribute = caller must be in a transaction");
	}

	// Suspend transaction if "NotSupported" or "RequiresNew"
	if (currtx != null &&
	    (txattr == MethodDesc.TX_REQUIRES_NEW || txattr == MethodDesc.TX_NOT_SUPPORTED)) {
	    try {
		cltx = tm.suspend();
		currtx = null;
	    } catch (SystemException e) {
		TraceEjb.error(getName()+": Cannot suspend transaction:\n", e);
		throw new RemoteException("Cannot suspend transaction", e);
	    }
	}

	// Start a new transaction in 2 cases:
	// - "RequiresNew" attribute
	// - "Required" and no current transaction
	// This transaction will be closed at postinvoke.
	if (txattr == MethodDesc.TX_REQUIRES_NEW || 
	    (txattr == MethodDesc.TX_REQUIRED && currtx == null)) {
	    try {
		tm.begin();
		mustCommit = true;
		currtx = tm.getTransaction();
	    } catch (NotSupportedException e) {
		TraceEjb.error(getName()+": Cannot start a transaction: NotSupportedException");
		throw new RemoteException("Nested Transactions Not Supported");
	    } catch (SystemException e) {
		TraceEjb.error(getName()+": Cannot start a transaction:\n", e);
		throw new RemoteException("Cannot start a transaction: SystemException");
	    }
	}

	// update RequestCtx
	rctx.setTransactions(mustCommit, cltx, currtx);
    }


    /**
     * Set the bean environment
     * args ctx Context for this bean
     * args dd Bean Deployment Descriptor
     * throws NamingException if could not rebind objects
     */
    public void setBeanEnvironment(Context ctx, BeanDesc dd) throws NamingException {

	// Set bean context
	Context ctxold = naming.setComponentContext(ctx);

	// Bean Environment
	EnvEntryDesc [] envt = dd.getEnvEntryDesc();
	for (int i = 0; i < envt.length; i++) {
	    // get information in descriptor
	    String name = envt[i].getName();
	    Object obj = envt[i].getValue();
	    // register object in JNDI
	    if (TraceEjb.isDebugServer) 
		TraceEjb.debugServer(getName()+": Binding object "+name+" -> "+obj);
	    ctx.rebind(name, obj);
	}

	// Resource References
	ResourceRefDesc [] resref = dd.getResourceRefDesc();
	for (int i = 0; i < resref.length; i++) {
	    // get information in descriptor
	    String name = resref[i].getName();
	    String resname = resref[i].getJndiName();
	    // build the LinkRef that will be registered:
	    // FactoryClassName = null, size = 1, refAddr = resname.
	    LinkRef lref = new LinkRef(resname);

	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in rmiregistry (and not in CosNaming)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
	    }

	    // register object in JNDI
	    if (TraceEjb.isDebugServer) 
		TraceEjb.debugServer(getName()+": Linking resource "+name+" -> "+resname);
	    ctx.rebind(name, lref);
	}

	// Resource Environment References
	ResourceEnvRefDesc [] resEnvref = dd.getResourceEnvRefDesc();
	for (int i = 0; i < resEnvref.length; i++) {
	    // get information in descriptor
	    String name = resEnvref[i].getName();
	    String resname = resEnvref[i].getJndiName();
	    LinkRef lref = new LinkRef(resname);

	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in rmiregistry (and not in CosNaming)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "RmiRegistry"));
	    }
    
	    if (TraceEjb.isDebugServer) 
		TraceEjb.debugServer(getName()+": Linking resource environment "+name
				     +" -> "+resname);
	    ctx.rebind(name, lref);
	}

	// EJB References
	EjbRefDesc [] ejbref = dd.getEjbRefDesc();
	for (int i = 0; i < ejbref.length; i++) {
	    // get information in descriptor
	    String name = ejbref[i].getName();
	    String ejbname = ejbref[i].getJndiName();
	    LinkRef lref = new LinkRef(ejbname);
	    
	    // if we are using David add a parameter to the link to tell that the
	    // object must be found in the CosNaming (and not in rmiregistry)
	    // this info is used by CompNamingContext.lookup()
	    if (Env.isDavid()) {
		lref.add(new StringRefAddr("TargetNaming", "CosNaming"));
	    }

	    if (TraceEjb.isDebugServer) 
		TraceEjb.debugServer(getName()+": Linking ejb "+name+" -> "+ejbname);
	    ctx.rebind(name, lref);
	}

	// Reset bean context
	naming.setComponentContext(ctxold);
    }

    /**
     * set the Transaction Manager.
     * Called by the EJB Server when starting ths service.
     */
    public void setTransactionManager(TransactionManager tm) {
	this.tm = tm;
    }

    /**
     * return the Transaction Manager
     * used internally by all the EJB Container classes.
     */
    public TransactionManager getTransactionManager() {
	return tm;
    }

    /**
     * set the ContainerNaming object
     * Called by the EJB Server when starting ths service.
     */
   public void setContainerNaming(ContainerNaming naming) {
	this.naming = naming;
   }
  
    /**
     * return the ContainerNaming object
     * used internally by all the EJB Container classes.
     */
    public ContainerNaming getContainerNaming() {
	return naming;
    }

    /**
     * set the JmsManager object
     * Called by the EJB Server when starting ths service.
     */
   public void setJmsManager(JmsManager jms) {
	this.jms = jms;
   }

    /**
     * return the JmsManager object
     * used internally by all the EJB Container classes.
     */
    public JmsManager getJmsManager() {
	return jms;
    }

    /**
     * Adds Session beans in container.
     * @param desc  DD.
     */
    public Ejb addSessionBean(SessionDesc desc) throws Exception {
	TraceEjb.debugServer(getName(),": addSessionBean ");
	// Build the JOnAS generated names for that bean.
	BeanNaming beanNames = new BeanNaming(desc);
	// Loads the specific EJBHome class
	String ejbHomeClassName = beanNames.getFullWrpHomeName();
	Class ejbHomeClass = null;
	try {
	    ejbHomeClass = getClassLoader().loadClass(ejbHomeClassName);
	} catch (ClassNotFoundException e) {
	    TraceEjb.error(getName()+" Cannot load "+ ejbHomeClassName);
	    throw e;
	}
	if (TraceEjb.isDebugServer)
	    TraceEjb.debugServer(getName()+": "+ejbHomeClassName+" loaded");

	// Creates JSessionHome instance by calling the appropriate constructor
	Object ejbHomeInstance  = null;
	try {
	    int nbParams = 3;
	    Class paramTypes[] = new Class[nbParams];  // constructor argument types
	    Object paramObjects[] = new Object[nbParams]; // constructor argument values
	    paramTypes[0] = org.objectweb.jonas_ejb.deployment.api.SessionDesc.class;
	    paramObjects[0] = (Object)desc;
	    // Second parameter
	    paramTypes[1] = java.lang.String.class;
	    paramObjects[1] = (Object)desc.getEjbName();
	    // Third parameter
	    paramTypes[2] = org.objectweb.jonas_ejb.container.JContainerImpl.class;
	    paramObjects[2] = (Object)this;
	    Constructor constructor = ejbHomeClass.getConstructor(paramTypes);
	    ejbHomeInstance = constructor.newInstance(paramObjects);
	} catch (InvocationTargetException te) {
	    TraceEjb.error("    InvocationTargetException : " + te.getTargetException());
	    throw te;
	} catch (Exception e) {
	    TraceEjb.error("    Call of " + ejbHomeClassName + " constructor : " + e);
	    throw e;
	}
	sessionList.addElement(ejbHomeInstance);
	return (Ejb)ejbHomeInstance;
    }

    /**
     * Adds Message Driven  beans in container.
     * @param desc  DD.
     */
    public Ejb addMessageDrivenBean(MessageDrivenDesc desc) throws Exception {
	TraceEjb.debugServer(getName(),": addMessageDrivenBean ");
	// We need a ThreadPool only for MDBs
	if (threadpool == null) {
	    threadpool = new ThreadPool(threadpoolsize);
	}
	JMdbFactory jmdbf = new JMdbFactory(desc, this, threadpool);
	mdbList.addElement(jmdbf);
	return jmdbf;
    }

    /**
     * Adds Entity bean in container.
     * @param desc  DD.
     */
    public Ejb addEntityBean(EntityDesc desc) throws Exception {
	TraceEjb.debugServer(getName(),": addEntityBean ");		
	// Build the JOnAS generated names for that bean.
	BeanNaming beanNames = new BeanNaming(desc);
		
	// Finds the DataSource associated to the bean
	// Useful for Entity with Container Managed Persistence.
	String dataSourceName = null;
	if (desc instanceof EntityJdbcCmpDesc) {
	    dataSourceName = ((EntityJdbcCmpDesc)desc).getDatasourceJndiName();
	}
	Object dataSource = null;

	Context context = naming.getInitialContext();
	if (context == null) {
	    TraceEjb.error(getName()+": JOnAS Server configuration error!");
	    TraceEjb.error("Cannot get initial context");
	    throw new EJBException("Cannot get initial context");
	}
	if (dataSourceName != null) {
	    try {
		dataSource = context.lookup(dataSourceName);
	    } catch(NamingException e) {
		TraceEjb.error(getName()+": JOnAS Server configuration error!");
		TraceEjb.error("Datasource '"+dataSourceName+"' is not known by the EJB server.");
		TraceEjb.error("Exception on lookup:"+e);
		TraceEjb.error("Please check the 'jonas.service.dbm.datasources' property in the 'jonas.properties' file.");
		throw e;
	    }
	}
	// Loads the specific EJBHome class
	String ejbHomeClassName = beanNames.getFullWrpHomeName();
	Class ejbHomeClass = null;
	try {
	    ejbHomeClass = getClassLoader().loadClass(ejbHomeClassName);
	} catch (ClassNotFoundException e) {
	    TraceEjb.error(getName()+" Cannot load "+ ejbHomeClassName);
	    throw e;
	}
	if (TraceEjb.isDebugServer)
	    TraceEjb.debugServer(getName()+": "+ejbHomeClassName+" loaded");
	
	
	// Creates JEntityHome instance by calling the appropriate constructor
	Object ejbHomeInstance  = null;
	
	try {
	    int nbParams = 4;
	    Class paramTypes[]   = new Class[nbParams];  // constructor argument types
	    Object paramObjects[] = new Object[nbParams]; // constructor argument values
	    
	    paramTypes[0] = org.objectweb.jonas_ejb.deployment.api.EntityDesc.class;
	    paramObjects[0] = (Object)desc;
	    
	    // Second parameter
	    paramTypes[1]       = java.lang.String.class;
	    paramObjects[1]     = (Object)desc.getEjbName();
	    
	    // Third parameter
	    paramTypes[2] = org.objectweb.jonas_ejb.container.JContainerImpl.class;
	    paramObjects[2] = (Object)this;

	    // Fourth parameter
	    paramTypes[3]   = java.lang.Object.class;
	    paramObjects[3] = (Object)dataSource;

	    Constructor constructor = ejbHomeClass.getConstructor(paramTypes);
	    ejbHomeInstance = constructor.newInstance(paramObjects);
	} catch (InvocationTargetException te) {
	    TraceEjb.error("    InvocationTargetException : " + te.getTargetException());
	    throw te;
	} catch (Exception e) {
	    TraceEjb.error("    Call of " + ejbHomeClassName + " constructor : " + e);
	    throw e;
	}
	
	// Keep the new EJBHome in a entity list
	entityList.addElement(ejbHomeInstance);
	// Set the passivation timeout
	setSwapTime(desc.getPassivationTimeout());
	return(Ejb)ejbHomeInstance;
    }

    /**
     * Adds beans in container.
     *
     * @param desc  Deployment Descriptor
     */
    public Ejb addBean(BeanDesc dd) throws Exception {
	Ejb  mEjbInstance = null;
	String beanName = dd.getEjbName();
	if (dd instanceof SessionDesc) {
	    mEjbInstance = addSessionBean((SessionDesc)dd);
	} else if (dd instanceof MessageDrivenDesc) {
	    mEjbInstance = addMessageDrivenBean((MessageDrivenDesc)dd);
	} else if (dd instanceof EntityDesc) {
	    mEjbInstance = addEntityBean((EntityDesc)dd);
	} else {
	    throw new EJBException("Bad Descriptor Type for "+beanName);
	}
	System.out.println(getName()+": "+ beanName + " available");
	return mEjbInstance;
    }


    /**
     * MBean method
     * @return Integer Number of Bean type currently in This container
     */
    public Integer getCurrentNumberOfBeanType(){
	return new Integer(entityList.size() + sessionList.size() + mdbList.size());
    }

    /**
     * MBean method
     * @return Integer Number of BMP type currently in This container
     */
    public Integer getCurrentNumberOfBMPType(){
	JBeanHome bbh;
	int totalCount = 0;
	Enumeration enum = entityList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    if (bbh.getDeploymentDescriptor() instanceof EntityBmpDesc) totalCount++;
	}
	return new Integer(totalCount);
    }

    /**
     * MBean method
     * @return Integer Number of CMP type currently in This container
     */
    public Integer getCurrentNumberOfCMPType(){
	JBeanHome bbh;
	int totalCount = 0;
	Enumeration enum = entityList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    if (bbh.getDeploymentDescriptor() instanceof EntityCmpDesc) totalCount++;
	}
	return new Integer(totalCount);
    }

    /**
     * MBean method
     * @return Integer Number of SBF type currently in This container
     */
    public Integer getCurrentNumberOfSBFType(){
	JBeanHome bbh;
	int totalCount = 0;
	Enumeration enum = sessionList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    if (bbh.getDeploymentDescriptor() instanceof SessionStatefulDesc) totalCount++;
	}
	return new Integer(totalCount);
    }

    /**
     * MBean method
     * @return Integer Number of SBL type currently in This container
     */
    public Integer getCurrentNumberOfSBLType(){
		JBeanHome bbh;
	int totalCount = 0;
	Enumeration enum = sessionList.elements();
	while (enum.hasMoreElements()) {
	    bbh = (JBeanHome) enum.nextElement();
	    if (bbh.getDeploymentDescriptor() instanceof SessionStatelessDesc) totalCount++;
	}
	return new Integer(totalCount);
    }

    /**
     * MBean method
     * @return Integer Number of MDB type currently in This container
     */
    public Integer getCurrentNumberOfMDBType() {
	return new Integer(mdbList.size());
    }

    /**
     * MBean method
     * @return String The name of this container
     */    
    public String getContainerName(){
	return getName();
    }

    /**
     * MBean Entity method
     * @return passivation time out for this container
     */
    public long getSwapTime() {
	return swapper.getTimeout();
    }

    /**
     * Management method
     * @return all ejb ObjectName using this datasource (with  there container name)
     */
    public Set getDataSourceDependence(String dsName) {
	HashSet result = new HashSet();
	// entity dependence 
	result.addAll(beansDependence(entityList.elements(), dsName, true));
	// session dependence 
	result.addAll(beansDependence(sessionList.elements(), dsName, true));
	// driven bean dependence
	result.addAll(beansDependence(mdbList.elements(), dsName, true));

	return result;

    }

    /**
     * Manangement method
     * @return all ejb ObjectName using this destination (with  there container name)
     */
    public Set getJmsDestinationDependence(String dsName){
	HashSet result = new HashSet();
	// entity dependence 
	result.addAll(beansDependence(entityList.elements(), dsName, false));
	// session dependence 
	result.addAll(beansDependence(sessionList.elements(), dsName, false));
	// driven bean dependence
	result.addAll(beansDependence(mdbList.elements(), dsName, false));
	return result;
    }

    /**
     * Management method
     * @return all ejb ObjectName using this Connection Factory (with  there container name)
     */ 
     public Set getJmsConnectionFactoryDependence(String cfName){
	 HashSet result = new HashSet();
	// entity dependence 
	result.addAll(beansDependence(entityList.elements(), cfName, true));
	// session dependence 
	result.addAll(beansDependence(sessionList.elements(), cfName, true));
	// driven bean dependence
	result.addAll(beansDependence(mdbList.elements(), cfName, true));
	return result;
     }

    /**
     * Management method
     * @return set of ejb ebjname string name and type using this ds
     * @param numeration of bean
     * @param String dependence jndi name
     * @param boolean true for resource-ref false for ressource-env-ref 
     */
    private Set beansDependence(Enumeration enum, String rName, boolean isResRef) {
	HashSet result = new HashSet();
	Ejb me;
	while (enum.hasMoreElements()) {
	    me = (Ejb) enum.nextElement();
	    BeanDesc meDesc = me.getDeploymentDescriptor();
	    String ejbType = null;
	    if (me instanceof JBeanHome) {
		 if (meDesc instanceof EntityBmpDesc) ejbType = "ejbbmp";
		 if (meDesc instanceof EntityCmpDesc ) ejbType = "ejbcmp";
		 if (meDesc instanceof SessionStatefulDesc) ejbType = "ejbsbf";
		 if (meDesc instanceof SessionStatelessDesc) ejbType = "ejbsbl";       
	    } else {
		ejbType = "ejbmdb";
	    }
	    if (isResRef) {// for Ressource-ref
		try {
		    ResourceRefDesc[] rrDesc = meDesc.getResourceRefDesc();
		    int i=0;
		    boolean exit = false;
		    while ((!exit)&&(i < rrDesc.length)) {
			if ((rrDesc[i].getJndiName().equals(rName))) {
			    Properties toAdd = new Properties();
			    toAdd.setProperty("type", ejbType);
			    toAdd.setProperty("fname", getFileName());
			    toAdd.setProperty("name", me.getDeploymentDescriptor().getEjbName());
			    result.add(toAdd);
			    exit = true;
			}
			i++;
		    }

		    // for cmp bean data source dependence
		    if ((ejbType == "ejbcmp")&&(!exit)) {
			if (rName.equals(((EntityJdbcCmpDesc)meDesc).getDatasourceJndiName())) {
			    Properties toAdd = new Properties();
			    toAdd.setProperty("type", ejbType);
			    toAdd.setProperty("fname", getFileName());
			    toAdd.setProperty("name", me.getDeploymentDescriptor().getEjbName());
			    result.add(toAdd);
			    exit = true;
			}
		    }
		} catch (Exception e) {
		    TraceEjb.error(getName()+": bean dependence:", e);
		}
	    } else { // for Ressource-env-ref
		try {
		    ResourceEnvRefDesc[] rerDesc = meDesc.getResourceEnvRefDesc();
		    int i=0;
		    boolean exit = false;
		    while ((!exit)&&(i < rerDesc.length)) {
			if ((rerDesc[i].getJndiName().equals(rName))) {
			    Properties toAdd = new Properties();
			    toAdd.setProperty("type", ejbType);
			    toAdd.setProperty("fname", getFileName());
			    toAdd.setProperty("name", me.getDeploymentDescriptor().getEjbName());
			    result.add(toAdd);
			    exit = true;
			}
			i++;
		    }
		    if ((ejbType == "ejbmdb")
			&&(rName.equals(((MessageDrivenDesc)meDesc).getDestinationJndiName()))) {
			// add this ejb  for his associate destination
			Properties toAdd = new Properties();
			toAdd.setProperty("type", ejbType);
			toAdd.setProperty("fname", getFileName());
			toAdd.setProperty("name", me.getDeploymentDescriptor().getEjbName());
			result.add(toAdd);
		    }
		} catch (Exception e) {
		    TraceEjb.error(getName()+": bean dependence:", e);
		}
	    }	
	}
	return result;
    }   
}
