/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container; 

import java.rmi.RemoteException;
import javax.ejb.EntityBean;
import javax.ejb.EJBObject;
import javax.ejb.EJBException;
import javax.ejb.RemoveException;
import javax.naming.Context;
import javax.ejb.EntityContext;
import javax.transaction.Transaction;
import javax.transaction.Status;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;

/**
 * This class implements javax.ejb.EntityContext interface.
 *
 * This class implements the  context associated to an EJBObject Entity
 * in order to allow concurrent access from multiple transactions 
 */
public class JContextEntity extends JContext implements EntityContext, Synchronization {

    // don't change for all Contexts
    private EntityBean enterpriseBean; // the bean instance
    
    // must be reinitialized before use (at each transaction)
    Transaction beanCoord;	// Transaction related to this synchronization
    private JBeanEntity ejbObject; // the remote object
    boolean ismarkedremoved = false;     // true between a remove and the commit 
    boolean isnewinstance = false;	// true if just created in this transaction

    // ------------------------------------------------------------------
    // constructors
    // ------------------------------------------------------------------

    /**
     * Constructs an EntityContext 
     * 
     * @param home the enterprise bean's home interface
     * @param eb   the Enterprise Bean
     */
    public JContextEntity(JBeanHome home, EntityBean eb) {
	super(home);
	TraceEjb.debugCtx("JContextEntity constructor");
	enterpriseBean = eb;
    }
  
    // ------------------------------------------------------------------
    // EntityContext implementation
    // ------------------------------------------------------------------

    /**
     * Obtains a reference to the EJB object that is currently associated with the instance. 
     *  An instance of enterprise Bean can call this method at anytime between the
     *   ejbCreate() and ejbRemove() methods, including from within the ejbCreate() and
     *   ejbRemove() methods.
     *  An instance can use this method, for example, when it wants to pass a reference to 
     *   itself in a method argument or result.
     * @return The EJB object currently associated with the instance. 
     * @exception IllegalStateException: Thrown if the instance invokes this 
     *            method while the instance is in a state that does not allow the 
     *            instance to invoke this method.
     */
    public  EJBObject getEJBObject() throws IllegalStateException {
	TraceEjb.debugCtx("JContextEntity.getEJBObject");

	if (ismarkedremoved || ejbObject == null)
	    throw new IllegalStateException("EJB Object is removed");
	return ejbObject;
    }


    /**
     * Obtains the primary key of the EJB object that is currently associated 
     *  with this instance. 
     * An instance of an entity enterprise Bean can call this method only when the 
     *  instance is associated with an EJB object identity, i.e. in the ejbActivate, 
     *  ejbPassivate, ejbPostCreate method, ejbRemove, ejbLoad, ejbStore, and 
     *  business methods. 
     * Note: The result of this method is the same as the result of 
     *  getEJBObject().getPrimaryKey(). 
     *
     * @return  The EJB object currently associated with the instance. 
     *
     * @exception  IllegalStateException  Thrown if the instance invokes this 
     *             method while the instance is in a state that does not allow 
     *             the instance to invoke this method.     
     */
    public Object getPrimaryKey() throws IllegalStateException {
	TraceEjb.debugCtx("JContextEntity.getPrimaryKey ");
	try{
	    Object pk = ejbObject.getPrimaryKey();
	    if (pk == null)
		throw new IllegalStateException();
	    return pk;
	} catch(java.rmi.RemoteException e) {
	    TraceEjb.error("JContextEntity.getPrimaryKey() RemoteException caught :" , e);
	    throw new IllegalStateException();
	}
    }

    // -------------------------------------------------------------------
    // Synchronization implementation
    // -------------------------------------------------------------------

   /** 
     * This beforeCompletion method is called by the transaction 
     * manager prior to the start of the transaction completion process.
     * This method executes in the transaction context of the calling 
     * thread.
     * The Bean's state must be stored on the persistent storage before
     * the completion of the transaction.
    */
    public void beforeCompletion() {

	TraceEjb.debugCtx("JContextEntity.beforeCompletion");

	// First check that the primary key is not null.
	// We must avoid accessing the DataBase when ejbCreate has failed.
	if (isnewinstance) {
	    try {
		if (ejbObject.getPrimaryKey() == null) {
		    // We do not force abort creation in order
		    // to be able to catch the CreateException
		    // otherwise CreateException will be overloaded
		    // in postinvoke by "cannot commit exception"
		    return;
		}
	    } catch (RemoteException e) {
		return;
	    }
	}

	// Set the JNDI component context (for ejbStore)
	Context bnctx = home.setComponentContext();

	try {
	    storeIfModified();
	} catch (RemoteException e) {
	    // Could not store bean data: set transaction rollback only
	    abortTransaction();
	} finally {
	    home.resetComponentContext(bnctx);
	}
    }


    /** 
     * The afterCompletion method is called by the transaction 
     * manager after the transaction is committed or rolled back. 
     * This method executes without a transaction context.
     *
     * @param status The status of the transaction completion.
     */
    public void afterCompletion(int status) {
	if (TraceEjb.isDebugCtx)
	    TraceEjb.debugCtx("JContextEntity.afterCompletion: Status= "+status);

	// Just check that we knew we were in a transaction (debug)
	if (beanCoord == null) {
	    TraceEjb.error("JContextEntity.afterCompletion: no tx");
	}

	boolean committed = (status == Status.STATUS_COMMITTED);

	// We must remove the EJBObject in these 2 cases:
	// 1. object has been removed and tx has been committed
	// 2. object has been created but tx has been rolledback
	if (ismarkedremoved && committed || isnewinstance && !committed) {
	    TraceEjb.debugCtx("JContextEntity.afterCompletion: remove object");
	    try {
		((JEntityHome)home).removeEJBObject(ejbObject.getPrimaryKey());
	    } catch (RemoteException e) {
		TraceEjb.error("JContextEntity.afterCompletion : removeEJBObject fails:", e);
	    }
	    // mark remote object as removed to avoid using it anymore
	    // until the garbage collector dispose of it.
	    ejbObject.noMoreInUse();
	}

	// release the instance/context
	if (ejbObject == null) {
	    TraceEjb.error("JContextEntity.afterCompletion : null ejbobject!");
	    return;
	}

	// passivation will be decided by ejbObject, depending on commit
	// option that can be applied (A or C).
	// we must be in the correct component context for ejbPassivate
	Context bnctx = home.setComponentContext();
	ejbObject.releaseContext(committed);
	home.resetComponentContext(bnctx);
    }


    // -------------------------------------------------------------------
    // Other methods
    // -------------------------------------------------------------------

    /**
     * Reinit Context for reuse
     * Sets the reference to the EJBObject currently associated with this instance.
     *
     * @param pk - the EJB object currently associated with the instance.
     */
    public void initEntityContext(JBeanEntity remote, Transaction tx) {

	// reinit Context for reuse
	ejbObject = remote;
	beanCoord = tx;
    }

    /**
     * Raz Context before freeing it. This is mainly for garbage collector.
     */
    public void razEntityContext() {
	ejbObject = null;
	beanCoord = null;
	ismarkedremoved = false;
	isnewinstance = false;
    }
   
    /**
     * reuse EntityContext for another transaction.
     * optimization: don't pass it by the pool.
     */
    public void reuseEntityContext(Transaction tx) {
	beanCoord = tx;
	ismarkedremoved = false;
	isnewinstance = false;
    }

   /**
     * Returns the bean instance of this context
     * Used in the generated classes to retrieve the instance
     *
     * @return the bean instance
     *
     */
    public EntityBean getInstance() {
	if (enterpriseBean == null) {
	    TraceEjb.error("getInstance: null!");
	}
	return enterpriseBean;
    }

    /**
     *  Persistence: write data on storage
     */
    public void storeIfModified() throws RemoteException {

	if (ismarkedremoved) {
	    TraceEjb.debugCtx("JContextEntity.storeIfModified: marked removed");
	    return;
	}
	TraceEjb.debugCtx("JContextEntity.storeIfModified");

	try {
	    // this method can fail if database serializes transactions
	    // This may do nothing if bean has not been modified (isModified optimization)
	    enterpriseBean.ejbStore();
	} catch(RemoteException e) {
	    TraceEjb.error("JContextEntity.storeIfModified : ", e);
	    throw e;
	} catch (EJBException e) {
	    TraceEjb.verbose("JContextEntity.storeIfModified raised EJBException");
	    throw new RemoteException("Exception while storing data", e.getCausedByException());
	} catch(RuntimeException e) {
	    TraceEjb.error("JContextEntity.storeIfModified : ", e);
	    throw new RemoteException("Exception while storing data", e);
 	} catch(Error e) {
	    TraceEjb.error("JContextEntity.storeIfModified : ", e);
	    throw new RemoteException("Error while storing data", e);
	}
    }

    /*
     * abort current transaction if an exception occured while
     * accessing the bean state (load, store, ...)
     */
    private void abortTransaction() {

	TraceEjb.debugCtx("JContextEntity.abortTransaction");

	// If we are inside a transaction, it must be rolled back
	if (beanCoord != null) {
	    try {
		beanCoord.setRollbackOnly();
	    } catch (SystemException e) {
		TraceEjb.error("JContextEntity: Cannot setRollbackOnly");
	    }
	}
    }

    /**
     * passivate this instance
     */
    public void passivate() {
	TraceEjb.debugCtx("JContextEntity.passivate");

	try {
	    enterpriseBean.ejbPassivate();
	} catch(Exception e) {
	    TraceEjb.error("JContextEntity.passivate : ejbPassivate failed", e);
 	} catch(Error e) {
	    TraceEjb.error("JContextEntity.passivate : ejbPassivate error", e);
	}
    }

   /** 
    * Activate instance.
    *
    * @throws RemoteException if a system problem occured.
    */
    public void activate(boolean doactivate) throws RemoteException {
	TraceEjb.debugCtx("JContextEntity.activate");

	try {
	    if (doactivate) {
		// Call ejbActivate
		enterpriseBean.ejbActivate();
	    }
	    // Load bean state from database
	    enterpriseBean.ejbLoad();
	} catch(RemoteException e) {
	    TraceEjb.error("JContextEntity.activate: ", e);
	    throw e;
	} catch(RuntimeException e) {
	    TraceEjb.error("JContextEntity.activate: ", e);
	    throw new RemoteException("Got exception in activate:", e);
	} catch(Error e) {
	    TraceEjb.error("JContextEntity.activate : ", e);
	    throw new RemoteException("Got error in activate:", e);
	}
    }


    /**
     * Set the flag new instance 
     *
     * @param flg - true if we are creating instance
     *
     */
    public void setNewInstance(boolean flg) {
	TraceEjb.debugCtx("JContextEntity.setNewInstance");
	isnewinstance = flg;
    }

    /**
     * remove the bean instance associated to this context
     * this is the generic part of the generated remove method
     */
    public void remove() throws RemoveException, RemoteException {
	TraceEjb.debugCtx("JContextEntity.remove");

	// Access the database to remove the instance
	// If exception inside a transaction, rollback will be done
	// in postinvoke (rctx is aware of syst. exc.)
	enterpriseBean.ejbRemove();

	// If no transaction, release objects now
	// If transaction, just Mark instance as removed
	if (beanCoord == null) {
	    try {
		((JEntityHome)home).removeEJBObject(ejbObject.getPrimaryKey());
	    } catch (RemoteException e) {
		TraceEjb.error("JContextEntity.remove : removeEJBObject fails:", e);
	    }
	    ejbObject.noMoreInUse();
	}
	ismarkedremoved = true;
    }

}
