/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.io.Serializable;
import java.rmi.RemoteException;
import javax.ejb.SessionBean;
import javax.ejb.SessionSynchronization;
import javax.ejb.EJBObject;
import javax.ejb.RemoveException;
import javax.ejb.SessionContext;
import javax.naming.Context;
import javax.transaction.Transaction;
import javax.transaction.Synchronization;
import javax.transaction.Status;
import javax.transaction.SystemException;

/**
 * This class extends JContextSession in case of Stateful Session Bean.
 */
public class JContextSessionSf extends JContextSession implements Synchronization {

    // Transaction related to the synchronization object
    transient Transaction beanCoord = null; 

    // Bean-managed Transaction currently associated with bean instance
    transient Transaction beanAssociatedTransaction = null;  

    // true = in container transaction
    transient boolean isInContainerTransaction = false; 

    // true if first request not done
    transient boolean firstRequest; 

    /**
     * the bean instance implements SessionSynchronization interface
     * @serial
     */
    boolean flgsynchronized = false;

    /**
     * the bean instance has received a timeout
     * @serial
     */
    boolean flgtimeout = false;

    // ------------------------------------------------------------------
    // constructors
    // ------------------------------------------------------------------

    /**
     * Constructs a JContextSessionSf
     */
    public JContextSessionSf(JBeanHome home, SessionBean sb, boolean sync) {
	super(home, sb);
	flgsynchronized = sync;
    }


    // -------------------------------------------------------------------
    // Synchronization implementation
    // -------------------------------------------------------------------

   /** 
     * This beforeCompletion method is called by the transaction 
     * manager prior to the start of the transaction completion process.
     * This method executes in the transaction context of the calling 
     * thread.
     */
    public void beforeCompletion() {
	TraceEjb.debugCtx("JContextSessionSf.beforeCompletion()");

	SessionSynchronization ssbean = null;
	if ((ssbean = getSynchronized()) != null) {
	    // Set the JNDI bean context
	    Context bnctx = home.setComponentContext();
	    try {
		ssbean.beforeCompletion();
	    } catch (RemoteException e) {
		TraceEjb.error("JContextSessionSf.beforeCompletion exception:", e);
	    } finally {
		home.resetComponentContext(bnctx);
	    }
	}
    }

    /** 
     * The afterCompletion method is called by the transaction 
     * manager after the transaction is committed or rolled back. 
     * This method executes without a transaction context.
     *
     * @param status The status of the transaction completion.
     */
    public void afterCompletion(int status) {
	TraceEjb.debugCtx("JContextSessionSf.afterCompletion() ");

	SessionSynchronization ssbean = null;

	if ((ssbean = getSynchronized()) != null) {
	    // Set the JNDI bean context
	    Context bnctx = home.setComponentContext();
	    try {
		ssbean.afterCompletion(status == Status.STATUS_COMMITTED);
	    } catch (RemoteException e) {
		TraceEjb.error("JContextSessionSf.afterCompletion exception:", e);
	    } finally {
		home.resetComponentContext(bnctx);
	    }
	}
	beanCoord = null;
	firstRequest = true;

	// The timeout has expired during the transaction. We can now remove
	// this session.
	if (flgtimeout) {
	    TraceEjb.verbose("JContextSessionSf: timeout expired during the transaction");
	    flgtimeout = false;
	    ((JBeanSessionSf)ejbObject).forceSessionRemove();
	}
    }


    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    public void initSessionContext(JBeanSession remote) {
	super.initSessionContext(remote);
	firstRequest = true;
	flgtimeout = false;
    }

    public void setTimeoutFlag() {
	flgtimeout = true;
    }

    /**
     * Raz Context before freeing it. This is mainly for garbage collector.
     */
    public void razSessionContext() {
	beanCoord = null;
	beanAssociatedTransaction = null;
	isInContainerTransaction = false;
    }

    /**
     * Returns the bean instance viewed like an implementation of SessionSynchronization interface
     *
     * @return the SessionSynchronization bean instance
     *         null if the bean instance doesn't implement SessionSynchronization
     */
    public SessionSynchronization getSynchronized() {
	if (flgsynchronized)
	    return (SessionSynchronization)sessionBean;
	else
	    return null;
    }

    /**
     * This method considers the incoming bean request, and manages the 
     *  synchronization jobs to be done for the bean. 
     * Nothing is to be done outside a transaction scope.
     * If the bean was not yet involved in a transaction, registers the 
     *  synchro object in the TM, and calls the bean afterBegin() method.
     * If the bean was already involved in a transaction, and a request 
     *  arrives for an other transaction, then the thread is suspended, 
     *  otherwise the request is taken into account.
     * This method is not called if the instance don't implement SessionSynchronization
     */
    public void considerRequest() throws RemoteException {
	TraceEjb.debugCtx("JContextSessionSf.considerRequest");

	Transaction currentCoord = null;
	try {
	    currentCoord = tm.getTransaction();
	} catch (SystemException e) {
	    TraceEjb.verbose("considerRequest: Cannot get transaction");
	    throw new RemoteException("Cannot get transaction");
	}
	if (currentCoord == null) {
	    TraceEjb.debugCtx("considerRequest: currentCoord == null");
	    // A Session Bean cannot be called in different transactions
	    if (beanCoord != null) {
		TraceEjb.error("The Session Bean cannot be called in a different context");
		throw new RemoteException("The Session Bean cannot be called in a different context");
	    }
	    // No transaction => no synchronization.
	    return;
	}

	// If this is not the first request for this Tx, return.
	// Note that if the request is called outside any Tx, it's like
	// if each request were in a separate Tx.
	if (!firstRequest) {
	    // A Session Bean cannot be called in different transactions
	    if (beanCoord.equals(currentCoord) == false) {
		TraceEjb.verbose("Bean already engaged in an other transaction");
		throw new RemoteException("The Session Bean cannot be called in a different context");
	    }		
	    return;
	}

	// just to remember that we are participating in a transaction.
	beanCoord = currentCoord;
	firstRequest = false; // just to remember that next requests are not the first request.
	// register this object as a synchronization
	try {
	    beanCoord.registerSynchronization(this);
	} catch (Exception e) {
	    TraceEjb.error("considerRequest: registerSynchronization(this) : ", e);
	    return;
	}

	// Calls afterBegin on SessionSynchro interface implemented by the bean
	SessionSynchronization ssbean = null;
	if ((ssbean = getSynchronized()) != null) {
	    try {
		ssbean.afterBegin();
	    } catch (Exception e) {
		TraceEjb.error("considerRequest: afterBegin() : ", e);
	    } 
	} else {
	    TraceEjb.error("considerRequest: not synchronized???");
	}
    }

    /**
     * return true if the bean is participating in a transaction
     */
    public boolean inTransaction() {
	return ((beanCoord != null) && ! isInContainerTransaction);
    }

}
