/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * Markus Karg: Novell port.
 *
 * Kok Arjan: Fix the bug of the Application Exception thrown in ejbCreate 
 *            and not received by the client.
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.Enumeration;
import java.util.Hashtable;
import java.rmi.RemoteException;
import javax.ejb.EntityBean;
import javax.naming.Context;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionRolledbackException;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmpDesc;


/**
 * Class inherited by all the Home object for entity beans
 */
abstract public class JEntityHome extends JBeanHome {
    
    /**
     * @serial
     */
    private Hashtable ejbobjects = new Hashtable();

    /**
     * @serial
     */
    private Object datasource = null;

    /**
     * @serial
     */
    private boolean isContainerManaged;

    /**
     * true if the bean is Reentrant: i.e. instances can be accessed concurrently
     * by different threads.
     * @serial
     */
    private boolean isReentrant;

    /**
     * true if bean is shared with other applications
     * @serial
     */
    private boolean isShared;

    /**
     * @serial
     */
    protected String primaryKeyClassName = null;

    /**
     * @serial
     */
    protected int instanceCount = 0;

    /**
     * max cache size
     * @serial
     */
    protected int maxCacheSize;

    /**
     * min pool size
     * @serial
     */
    protected int minPoolSize;


    /**
     * constructor
     * The DataSourceImpl param value may be null
     */
    public JEntityHome(EntityDesc desc, String beanName, JContainerImpl cont, Object ds) 
	throws RemoteException {

	super(desc, beanName, cont);

	if (TraceEjb.isDebug) TraceEjb.debug("JEntityHome constructor for "+ beanName);

	this.isContainerManaged = (desc instanceof EntityCmpDesc);
	this.isReentrant = desc.isReentrant();
	this.isShared = desc.isShared();
	this.maxCacheSize = desc.getCacheMax();
	this.minPoolSize = desc.getPoolMin();
	this.datasource = ds;
	this.primaryKeyClassName = desc.getPrimaryKeyClass().getName();

	JBeanMetaData bbdm = new JBeanMetaData(this,
					       desc.getHomeClass(),
					       desc.getRemoteClass(),
					       desc.getPrimaryKeyClass(),
					       false,
					       false);
	this.setEJBMetaData(bbdm);

	// Populate the pool of Contexts
	for (int i = 0; i < minPoolSize; i++) {
	    addEntityContext();
	}
    }


    /**
     * Add the EJBObject in the list kept in the Home. The primary key
     * is used as a key in a hashtable to retrieve it later.
     * The callers must be synchronized
     *
     * @param ejbObject - the EJBObject
     */
    public void putEJBObject(JBeanEntity ejbObject) {

	TraceEjb.debug("JEntityHome.putEJBObject");
	
	// Get the primary key: should always be OK.
	Object pk;
	try {
	    pk = ejbObject.getPrimaryKey(); 
	} catch(RemoteException e) {
	    TraceEjb.error("putEJBObject: cannot get pk");
	    return;
	}
	if (!ejbobjects.containsKey(pk)) {
	    ejbobjects.put(pk, ejbObject);
	}
    }


    /**
     * Returns the EJBObject to which the specified primary key is mapped
     * The callers must be synchronized
     *
     * @param  pk - the primary key 
     * @return the value to which the key is mapped in this hashtable; 
     *         null if the key is not mapped to any value in this hashtable.
     */
    public JBeanEntity getEJBObject(Object pk) {

	if (TraceEjb.isDebug) TraceEjb.debug("JEntityHome.getEJBObject " +pk);

	JBeanEntity bb = null;
	bb = (JBeanEntity) ejbobjects.get(pk);
	if (bb == null) {
	    if (TraceEjb.isDebug) 
		TraceEjb.debug("There is no EJBObject for this primary key " +pk);
	}
	// Do not check isMarkedRemoved() anymore: not really useful and it
	// leads to memory leaks
	return bb;
    }

    /**
     * Remove EJBObject from the list
     * @args pk the Primary Key of this EJBObject
     */
    public void removeEJBObject(Object pk) {

	if (pk == null) {
	    if (TraceEjb.isDebug)
		TraceEjb.debug("JEntityHome.removeEJBObject: pk = null");
	    return;
	}
	TraceEjb.debug("JEntityHome.removeEJBObject");

	try {
	    synchronized(ejbobjects) {
		ejbobjects.remove(pk);
	    }
	} catch (Exception e) {
	    TraceEjb.error("Could not remove EJBObject: ", e);
	}
    }


    /**
     * Returns the DataSource associated to the bean
     * JOnAS dependant, used mainly by generated classes.
     *
     * @return  the DataSource associated to the bean or null if there is no associated DataSource
     *
     */
    public Object getDataSource() {
	TraceEjb.debug("JEntityHome.getDataSource");
	if (datasource == null) {
	    TraceEjb.error("JOnAS configuration error:");
	    TraceEjb.error("Datasource is null in EntityHome");
	}
	return datasource;
    }


    /**
     * Create a new couple Context+Instance
     * @params ejbobject The ejbObject owning this context
     * @params tx The transaction associated to it
     *
     * @returns ctx The new Entity Context
     */
    public JContextEntity createEntityContext(JBeanEntity ejbo, Transaction tx) throws RemoteException {
	TraceEjb.debug("JEntityHome.createEntityContext");

	JContextEntity ctx = null;
	EntityBean eb = null;
	try {
	    // create instance + context
	    eb = createNewInstance();
	    ctx = new JContextEntity(this, eb);
	    instanceCount++;
	    // If too much instances, ask swapper to clean memory.
	    if (maxCacheSize != 0 && instanceCount > maxCacheSize) {
		container.registerHome(this);
	    }
	} catch (Exception e) {
	    TraceEjb.error("createEntityContext: Failed to create new instance:", e);
	    throw new RemoteException("Container cannot create a new instance", e);
	}
	// Init the Context and give it to the instance.
	ctx.initEntityContext(ejbo, tx);
	eb.setEntityContext(ctx);
	return ctx;
    }

    /**
     * addEntityContext in pool of available instances
     * This is used when bean is loaded to speed up bean create.
     */
    public void addEntityContext() throws RemoteException {
	TraceEjb.debug("JEntityHome.addEntityContext");

	JContextEntity ctx = null;
	EntityBean eb = null;
	try {
	    // create instance + context
	    eb = createNewInstance();
	    ctx = new JContextEntity(this, eb);
	    instanceCount++;
	    // put it in the pool.
	    ctx.razEntityContext();
	    eb.setEntityContext(ctx);
	    freeEJBCtx(ctx);
	} catch (Exception e) {
	    TraceEjb.error("addEntityContext: ", e);
	    throw new RemoteException("Container cannot add new instance:", e);
	}
    }

    /**
     * Get a Context+Instance for finder methods
     * This instance is picked from the pooled instances.
     */
    public synchronized JContextEntity getAvailable() throws RemoteException {
	JContextEntity ctx = (JContextEntity)reuseEJBCtx();
	if (ctx == null) {
	    ctx = createEntityContext(null, null);
	}
	return ctx;
    }

    /**
     * preinvoke for entity home methods
     * Entity beans must be container managed for transations,
     * so we needn't to test this.
     */
    protected RequestCtx preinvoke(int txattr, String methodSignature) throws RemoteException {
	TraceEjb.debug("JEntityHome.preinvoke");
	try {
	    // security controls
	    checkSecurity(methodSignature);

	    // Build a RequestCtx to save information about this request
	    // until the postinvoke time.
	    RequestCtx rctx = new RequestCtx(txattr);

	    // Set bean context for JNDI
	    rctx.setOldJNDICtx(setComponentContext());

	    // Process transaction attribute
	    container.checkTransactionIn(rctx);
	    return rctx;
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * postinvoke for Entity home methods
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debug("JEntityHome.postinvoke");
	try {
	    // Commit or Rollback transaction started in preinvoke
	    if (rctx.mustCommit()) {
		if (rctx.getSysExc() != null) {
		    // rollback tx
		    try {
			TraceEjb.error("System exc. => Rollback the transaction.");
			tm.rollback();
		    } catch (Exception e) {
			TraceEjb.error("Exception during rollback:", e);
		    }
		} else {
		    // try to commit the transaction, even in case of appl. Exc.
		    // If rollbackonly, do not throw any runtime exception to the client.
		    boolean rollb = false;
		    JContextEntity ec = (JContextEntity) rctx.getEJBContext();
		    if( ec != null)
			rollb = ec.getRollbackOnly();
		    try {
			if (rollb) {
			    TraceEjb.debug("postinvoke: Rolling back transaction");
			    tm.rollback();
			} else {	
			    TraceEjb.debug("postInvoke: Committing transaction");
			    tm.commit();
			}
		    } catch (RollbackException e) {
			// Probably a pb while writing data on database.
			throw new RemoteException("Container cannot commit a transaction", e);
		    } catch (Exception e) {
			TraceEjb.error("postInvoke: ", e);
			throw new RemoteException("Container exception", e);
		    }
		}
	    }

	    // We got a system Exception in a method from home:
	    // - Log system Exception
	    // - discard instance (???)
	    // - set client transaction rollback only
	    // - throw remote exception
	    if (rctx.getSysExc() != null) {
		// Log system exception
		TraceEjb.error("postinvoke: System Exception in a home method:", rctx.getSysExc());
		
		// If client transaction: set it rollback only and throw TransactionRolledBackException
		Transaction currentTx = null;
		try {
		    currentTx = tm.getTransaction();
		    if (currentTx != null) {
			currentTx.setRollbackOnly();
			throw new TransactionRolledbackException("System exception in business method");
		    }
		} catch (SystemException e) {
		    TraceEjb.error("postinvoke: cannot set rollback only current tx:", e);
		}
		return;
	    }
	
	    // Reset bean context for JNDI
	    naming.setComponentContext(rctx.getOldJNDICtx());
	    
	    // Resume transaction context of the caller
	    Transaction cltx = rctx.getClientTx();
	    if (cltx != null) {
		try {
		    tm.resume(cltx);
		} catch (SystemException e) {
		    TraceEjb.error("postinvoke: cannot resume tx:\n", e);
		}
	    }
	    
	    // If no transaction, we must release the EntityContext
	    if (rctx.getCurrentTx() == null) {
		JContextEntity ec = (JContextEntity) rctx.getEJBContext();
		if (ec != null) {	// No entity context for finder methods!
		    JBeanEntity ejbobject = null;
		    try {
			ejbobject = (JBeanEntity) ec.getEJBObject();
			if (ejbobject == null) {
			    TraceEjb.error("postinvoke: ejbobject == null");
			} else {
			    ejbobject.releaseContext(true);
			}
		    } catch (IllegalStateException e) {
			// nothing to do if object deleted.
		    }
		}
	    }
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * Synchronizes all instances that are used outside a transaction.
     * @return int count of instances that where busy.
     */
    public int passivateAllIH(boolean passivation) {
	TraceEjb.debug("JEntityHome.passivateAllIH");
	int busy = 0;
	JBeanEntity ejb = null;
	Enumeration enum = null;
	synchronized(ejbobjects) {
	    enum = ejbobjects.elements();
	}

	// Set the JNDI component context (for ejbStore)
	Context bnctx = setComponentContext();
	int freecount = 0;
	while (enum.hasMoreElements()) {
	    ejb = (JBeanEntity) enum.nextElement();
	    // sanity check (but should never occur!)
	    if (ejb == null) {
		TraceEjb.error("JEntityHome "+getEjbName()+"error: ejbobjects holds a null ref");
		TraceEjb.error("We continue to see if it happens several times");
		//throw new RuntimeException("Internal error in JEntityHome");
		continue;
	    }
	    if (ejb.storeState(passivation) == false) {
		busy++;
	    }
	    /*
	     if (ejb.isUnreferenced() ){
		 if(ejb.isBusy()){
		    // busy don't touch it
		   TraceEjb.verbose("Cannot freeEJBObject "+ejb.toString());
		 }else{
		    // free EJBObjects
		     freeEJBObject(ejb);
		     freecount++;
		}
	    }
	    */
	}

	// Reset old value for component context
	resetComponentContext(bnctx);

	return busy;
    }

    /**
     * Reduce number of instances in memory
     */
    public void reduceCache() {
	TraceEjb.debug("JEntityHome.reduceCache");
 	int count = passivateAllIH(true);
	int poolsz = minPoolSize - count;
	if (poolsz < 0) 
	    poolsz = 0;
	count = trimEJBPoolSize(poolsz);
	instanceCount -= count;
    }

    /*
     * Release  instance of EJBObject
     * -> called by the Swapper on timeout via passivateAllIH
     *    when it find unreferenced instances
     * Not used with Jeremie no dgc
     */

    public synchronized void freeEJBObject(JBeanEntity sejb) {
	TraceEjb.debug("JEntityHome.freeEJBObject");
	sejb.noMoreInUse();
	try{
	    Object pk = sejb.getPrimaryKey();	
	    removeEJBObject(pk);
	}catch(RemoteException e) {
	    TraceEjb.error("freeEJBObject: cannot get pk");
	}

    }


    /**
     * Return true if bean is reentrant
     */
    public boolean isReentrant() {
	return isReentrant;
    }

    /**
     * Return true is bean is shared
     */
    public boolean isShared() {
	return isShared;
    }

    /* 
     * must be generated in the interposition class
     */
    protected abstract EntityBean createNewInstance() throws Exception;

    /**
     * @return current count of instances (pool + cache)
     */
    public int getCacheSize() {
	return instanceCount;
    }

    /**
     * @return min-pool-size
     */
    public int getMinPoolSize() {
	return minPoolSize;
    }

    /**
     * @return max-cache-size
     */
    public int getMaxCacheSize() {
	return maxCacheSize;
    }

    /**
     * Take a dump of current entity counters and return them
     */
    public EntityCounters getEntityCounters() {
	EntityCounters ec = new EntityCounters();
	Enumeration enum = null;
	JBeanEntity ejb;
	synchronized(ejbobjects) {
	    enum = ejbobjects.elements();
	}
	while (enum.hasMoreElements()) {
	    ejb = (JBeanEntity) enum.nextElement();
	    switch (ejb.getInstanceState()) {
	    case 0:
		ec.inTx++;
		break;
	    case 1:
		ec.outTx++;
		break;
	    case 2:
		ec.idle++;
		break;
	    case 3:
		ec.passive++;
		break;
	    }
	}
	return ec;
    }

}
