/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 * Dean Jennings : optimization
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container; 

import java.util.Enumeration;
import java.util.Vector;
import java.util.List;
import java.rmi.RemoteException;
import javax.ejb.RemoveException;
import javax.ejb.SessionBean;
import javax.naming.NamingException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionRolledbackException;

import org.objectweb.jonas.thds.ThreadData;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;

/**
 * Generic part of the Home for Session beans
 */
public abstract class JSessionHome extends JBeanHome {

    // pooling of EJBObjects
    // objects in free_session_pool are cleaned (sc=null, ...)
    // XXX Try to use a Stack (more efficient?)
    // All pool accesses are synchronized with this monitor
    Vector use_session_pool = new Vector();  // EJBObjects in use (referenced by clients)
    Vector free_session_pool = new Vector(); // EjbObjects free

    boolean isStateful = false;	// true if stateful session bean
    int timeout = 0;		// session timeout

    /**
     * constructor
     */
    public JSessionHome(SessionDesc desc, String beanName, JContainerImpl cont) throws RemoteException {

	super(desc, beanName, cont);
	isStateful = (desc instanceof SessionStatefulDesc);
	TxBeanManaged = (desc.getTransactionType() == SessionDesc.BEAN_TRANSACTION_TYPE);
	JBeanMetaData bbmd = new JBeanMetaData(this,
					       desc.getHomeClass(), desc.getRemoteClass(), null,
					       true, !isStateful);
	this.setEJBMetaData(bbmd);
	this.timeout = desc.getSessionTimeout();
    }
    
    /**
     * return true if bean session stateful
     */
    public boolean isStateful() {
	return isStateful;
    }

    /**
     * Removes an EJB object identified by its primary key.
     * not allowed for a session.
     */
    public void remove(Object primaryKey) throws RemoteException, RemoveException {
	throw new RemoveException("There is no primary key on a sessionBean");
    }

    /**
     * Create a new couple Context+Instance
     * @params ejbobject The ejbObject owning this context
     *
     * @returns ctx The new Session Context
     */
    public JContextSession createSessionContext(JBeanSession ejbo, boolean sync) throws RemoteException {
	TraceEjb.debugInvoke("createSessionContext");

	JContextSession ctx = null;
	SessionBean sb = null;
	try {
	    sb = createNewInstance();
	    if (isStateful) {
		ctx = new JContextSessionSf(this, sb, sync);
	    } else {
		ctx = new JContextSessionSl(this, sb);
	    }
	} catch (Exception e) {
	    TraceEjb.error("createSessionContext: Failed to create new instance:", e);
	    throw new RemoteException("Container cannot create a new instance", e);
	}
	// Init the Context and give it to the instance.
	ctx.initSessionContext(ejbo);
	sb.setSessionContext(ctx);
	if (!isStateful) {
	    // for the stateless Session bean call to ejbCreate()
	    // We cannot simply code here "sb.ejbCreate()" because it's not part of
	    // the SessionBean standard interface.
	    callejbCreate(sb);
	}
	return ctx;
    }

    /**
     * preinvoke for session home methods
     * Don't check transaction attribute for session beans
     */
    protected RequestCtx preinvoke(int txattr, String methodSignature)  throws RemoteException {
	TraceEjb.debug("JSessionHome.preinvoke");
	try {
	    // security controls
	    checkSecurity(methodSignature);

	    // Build a RequestCtx to save information about this request
	    // until the postinvoke time.
	    RequestCtx rctx = new RequestCtx(txattr);

	    // Set bean context for JNDI
	    rctx.setOldJNDICtx(setComponentContext());

	    // Set the connectionList in threadLocal as null, since we don't know yet
	    // on which ejb instance we will work.
	    // save old value in RequestContext
	    List oldList = ThreadData.getInstance().setConnectionList(null);
	    rctx.setOldConnectionList(oldList);
	    return rctx;
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }

    /**
     * postinvoke for Session home methods
     */
    public void postinvoke(RequestCtx rctx) throws RemoteException {
	TraceEjb.debug("JSessionHome.postinvoke");
	try {
	    // We got a system Exception in a method from home:
	    // - Log system Exception
	    // - set client transaction rollback only
	    // - throw remote exception (will be done by generated code)
	    if (rctx.getSysExc() != null) {
		// Log system exception
		TraceEjb.error("postinvoke: System Exception in a home method:", rctx.getSysExc());
		
		// If client transaction: set it rollback only and throw TransactionRolledBackException
		try {
		    Transaction currentTx = tm.getTransaction();
		    if (currentTx != null) {
			currentTx.setRollbackOnly();
			throw new TransactionRolledbackException("System exception in business method");
		    }
		} catch (SystemException e) {
		    TraceEjb.error("JSessionHome.postinvoke: cannot set rollback only current tx:", e);
		}
	    }
	    // Reset bean context for JNDI
	    naming.setComponentContext(rctx.getOldJNDICtx());
	    
	    // Save connectionList in bean instance (create only) and reset the ThreadLocal
	    List oldList = rctx.getOldConnectionList();
	    List newList = ThreadData.getInstance().setConnectionList(oldList);
	    JContextSession ctx = (JContextSession) rctx.getEJBContext();
	    if (ctx != null) {
		ctx.setConnectionList(newList);
	    }
	} catch (RuntimeException e) {
	    // Log RuntimeException before rethrowing it
	    // -> Print Stack Trace to see where the problem is.
	    TraceEjb.error("Unexpected Runtime Exception: "+e);
	    e.printStackTrace();
	    throw e;
	}
    }
	
    /*
     * Get an instance of EJBObject
     * Try to get an instance from the free_session pool
     * -> called by generated classes
     */
    public synchronized JBeanSession reuseEJBObject() throws RemoteException {

	if (free_session_pool.size() > 0) {
	    JBeanSession sejb = (JBeanSession)free_session_pool.remove(0);
	    if (sejb.reUseObject() == false) {
		TraceEjb.debug("reuseEJBObject: Bad JBeanSession found in pool.");
		return null;
	    }
	    TraceEjb.debug("reuseEJBObject: JBeanSession found in pool.");
	    newEJBObject(sejb);
	    return sejb;
	} else {
	    TraceEjb.debug("reuseEJBObject: No JBeanSession available in pool");
	    return null;
	}
    }

    /*
     * Set this EJBObject as used (= session bean create)
     */
    public synchronized void newEJBObject(JBeanSession ejbobj) {
	TraceEjb.debug("JSessionHome.newEJBObject");

	// Put this EJBObject in the list of used
	use_session_pool.add(ejbobj);

	// Start a timer for this new session
	if (timeout > 0) {
	    ejbobj.startTimer(timeout);
	}
    }

    /*
     * Release  instance of EJBObject
     * -> called by JBeanSession.unreferenced() and remove()
     * In case of JEREMIE, only called in case of remove (no distributed GC)
     */
    public synchronized void freeEJBObject(JBeanSession sejb) {
	TraceEjb.debug("JSessionHome.freeEJBObject");

	// Stop timer
	if (timeout > 0) {
	    sejb.stopTimer();
	}

	// Move this EJBObject in free list
	use_session_pool.remove(sejb);
	free_session_pool.add(sejb);
	if (TraceEjb.isDebugPool) {
	    TraceEjb.debugPool("use_session_pool.size= " + use_session_pool.size());
	    TraceEjb.debugPool("free_session_pool.size= " + free_session_pool.size());
	}
    }
 
    /**
     * This home is removed by jonas admin
     */
    public synchronized void unregister() throws NamingException {
	TraceEjb.debug("JSessionHome.unregister");

	// remove all session beans
	Enumeration en = use_session_pool.elements();
	while (en.hasMoreElements()) {
	    JBeanSession sb = (JBeanSession)en.nextElement();
	    sb.removeInstance();
	}

	// If Session Stateless: call ejbRemove on all pooled instances
	if (!isStateful) {
	    try {
		JContextSessionSl sctx = null;
		while ((sctx = (JContextSessionSl) reuseEJBCtx()) != null) {
		    sctx.removeMyInstance();
		}
	    } catch (RemoveException e) {
		TraceEjb.verbose("unregister: RemoveException");
	    }
	}

	// release all objects in pools
	use_session_pool = null;
	free_session_pool = null;

	// unregister Home in JNDI
	super.unregister();
    }


    /**
     * MBean method: get the session time out for this session ejb
     * @return int session time out
     */
    public int getCurrentSessionTimeOut() {
	return timeout;
    }

    /**
     * MBean method: set the session time out for this session bean
     * @param int new session time out
     */ 
    public void setCurrentSessionTimeOut(int timeOut) {
	this.timeout=timeOut;
    }


    /**
     * MBean method: return the size for free ejb object pool
     * @return Integer pool size
     */ 
    public int getCurrentEJBObjectFreePoolSize() {
	return free_session_pool.size();
    }

    /**
     * MBean method:return the size for in use ejb object pool
     * @return Integer pool size
     */ 
    public int getCurrentEJBObjectInUsePoolSize(){
	return use_session_pool.size();
    } 



    /* 
     * must be generated in the interposition class
     */
    protected abstract void callejbCreate(SessionBean sb) throws RemoteException;
    protected abstract SessionBean createNewInstance() throws Exception;
}
