/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s):
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.LinkedList;
import java.util.NoSuchElementException;

/**
 * This class is responsible for swapping bean instances to keep
 * memory usage low enough, according to entity max-cache-size value
 * and to passivate instances on timeout when used outside transactions.
 * There is 1 swapper per Container.
 */
class Swapper extends Thread {

    private boolean valid;
    private boolean syncallbeans;
    private JContainerImpl container;
    private LinkedList homeList = new LinkedList();
    private LinkedList entityList = new LinkedList();
    private final int DEFAULT_TIMEOUT = 15 * 60 * 1000; // 15 mn
    private int swapTimeout = DEFAULT_TIMEOUT;

    /**
     * constructor
     */
    public Swapper(JContainerImpl cont) {
	super("JonasSwapper");
	valid = true;
	syncallbeans = false;
	container = cont;
    }

    public void run() {

	JBeanEntity ejb  = null;
	JEntityHome home = null;

	while (valid) {
	     // Don't keep lock while accessing beans.
	    synchronized(this) {
		if (entityList.size() == 0 && homeList.size() == 0) {
		    try {
			wait(swapTimeout);
		    } catch (InterruptedException e) {
			TraceEjb.error(getName()+": swapper interrupted");
		    } catch (Exception e) {
			TraceEjb.error(getName()+"swapper:", e);
		    }
		}
		try {
		    ejb = (JBeanEntity) entityList.removeFirst();
		} catch (NoSuchElementException e) {
		    ejb = null;
		}
		try {
		    home = (JEntityHome) homeList.removeFirst();
		} catch (NoSuchElementException e) {
		    home = null;
		}
	    }
	    if (!valid) {
		break;
	    }
	    if (ejb != null) {
		// Synchronizes this entity bean instance.
		if (ejb.storeState(false) == false) {
		    TraceEjb.error(getName()+"swapper: instance busy");
		}
	    }
	    if (home != null) {
		home.reduceCache();
	    }
	    if (ejb == null && home == null && syncallbeans) {
		// synchonize all beans on timeout if someone asked for this.
		container.syncAll();
	    }
	}
    }
	
    /**
     * stop the swapper.
     */
    public synchronized void stopIt() {
	valid = false;
	notify();
    }

    /**
     * add an entity bean to the list of beans to synchronize
     */
    public synchronized void addEntity(JBeanEntity ejb) {
	entityList.addLast(ejb);
	notify();
    }

    /**
     * add a home to the list of homes
     */
    public synchronized void addHome(JBeanHome home) {
	homeList.addLast(home);
	notify();
    }

    /**
     * set timeout value
     * @args timeout in milliseconds
     */
    public synchronized void setTimeout(int t) {
	syncallbeans = true;
	if (t < swapTimeout) {
	    swapTimeout = t;
	}
	notify();
    }

    /**
     * return the timeout for this container.
     * 0 means "no timeout".
     */
    public int getTimeout() {
	if (syncallbeans) {
	    return swapTimeout;
	} else {
	    return 0;
	}
    }
}
