/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 * 01/05/30 Christophe Ney - Lutris Technologies
 *          Fix to handle arguments containing white spaces.
 * 01/11/23 Guillaume Riviere
 *          Fix the addClassesInJar() method (on David, the Stub/Skel classes names are different).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */



package org.objectweb.jonas_ejb.tools;

import java.io.File;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmpDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;

/**
 *  This class allows to generate the sources of:
 * <ul>
 * <li>
 *  the class that implements the Enterprise bean's remote interface,
 * <li>
 *  the class that implements the Enterprise bean's home interface,
 * <li>
 *  the class of the Entity Handle in case of entity,
 * <li>
 *  and the extended class of the Bean for persistence in case of entity with CMP,
 * </ul>
 * of a given Enterprise Java Bean.
 */

public class GenIC {

    // Names of the implementation classes (Full names and base names)
    private String wrpRemoteName     = null;
    private String wrpHomeName       = null;
    private String wrpRemoteBaseName = null;
    private String wrpHomeBaseName   = null;
    private String wrpHandleName     = null;
    private String wrpHandleBaseName = null;
    private String wrpBeanName       = null;
    private String wrpBeanBaseName   = null;

    // Path of the java files of the implementation classes
    private String wrpRemoteFileName = null;
    private String wrpHomeFileName   = null;
    private String wrpHandleFileName = null;
    private String wrpBeanFileName   = null;

    // Pseudo path of the java files of the Home and Remote interface
    private String ejbRemoteFileName = null;
    private String ejbHomeFileName   = null;

    // Sources of the Implementation Remote and Home classes
    private GenICHome   srcICHome   = null;
    private GenICRemote srcICRemote = null;
    private GenICHandle srcICHandle = null;      // Warning: Stay to null in case of Session !
    private GenICBean   srcICBean   = null;      // Warning: Stay to null in case of no CMP Entity!

    // Name of the directory where to place the generated file
    private String directoryOutputName = null;

    // Bean Deployment Description
    private BeanDesc beanDD = null;

    /**
     * GenIC Constructor
     *
     * @param     beanDescr      deployment descriptor of the bean
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @exception GenICException In error case
     */
    public GenIC(BeanDesc beanDesc, String dirOutputName) throws GenICException {
	
	beanDD = beanDesc;
	directoryOutputName = dirOutputName;

	// Classes Names building
	BeanNaming bn     = new BeanNaming(beanDesc);
	wrpRemoteBaseName = bn.getWrpRemoteName();
	wrpRemoteName     = bn.getFullWrpRemoteName();
	wrpHomeBaseName   = bn.getWrpHomeName();
	wrpHomeName       = bn.getFullWrpHomeName();
	wrpHandleBaseName = bn.getWrpHandleName();
	wrpHandleName     = bn.getFullWrpHandleName();
	wrpBeanBaseName   = bn.getDerivedBeanName();
	wrpBeanName       = bn.getFullDerivedBeanName();
	
	// Files Names building
	if ("".equals(directoryOutputName)) {
	    // FileName = wrpXxBaseName + ".java"
	    wrpRemoteFileName = wrpRemoteBaseName + ".java";
	    wrpHomeFileName   = wrpHomeBaseName + ".java";
	    wrpHandleFileName = wrpHandleBaseName + ".java";
	    wrpBeanFileName   = wrpBeanBaseName + ".java";

	    ejbRemoteFileName = beanDD.getRemoteClass().getName().replace('.', File.separatorChar)
		                + ".java";
	    ejbHomeFileName = beanDD.getHomeClass().getName().replace('.', File.separatorChar)
		              + ".java";
	} else {
	    // FileName = directoryOutputName + '/' + packageName + '/' + wrpXxBaseName + ".java"
	    String sDir;
	    sDir = BeanNaming.getPackageName(wrpRemoteName);
	    if (! "".equals(sDir)) {
		sDir = sDir.replace('.', File.separatorChar) + File.separatorChar;
	    }
	    wrpRemoteFileName = directoryOutputName + File.separatorChar + sDir
		                + wrpRemoteBaseName + ".java";
	    sDir = BeanNaming.getPackageName(wrpHomeName);
	    if (! "".equals(sDir)) {
		sDir = sDir.replace('.', File.separatorChar) + File.separatorChar;
	    }
	    wrpHomeFileName   = directoryOutputName + File.separatorChar + sDir
		                + wrpHomeBaseName + ".java";
	    sDir = BeanNaming.getPackageName(wrpHandleName);
	    if (! "".equals(sDir)) {
		sDir = sDir.replace('.', File.separatorChar) + File.separatorChar;
	    }
	    wrpHandleFileName   = directoryOutputName + File.separatorChar + sDir
				  + wrpHandleBaseName + ".java";
	    sDir = BeanNaming.getPackageName(wrpBeanName);
	    if (! "".equals(sDir)) {
		sDir = sDir.replace('.', File.separatorChar) + File.separatorChar;
	    }
	    wrpBeanFileName   = directoryOutputName + File.separatorChar + sDir
				+ wrpBeanBaseName + ".java";

	    ejbRemoteFileName = directoryOutputName + File.separatorChar
		              + beanDD.getRemoteClass().getName().replace('.', File.separatorChar)
		              + ".java";
	    ejbHomeFileName = directoryOutputName + File.separatorChar
		            + beanDD.getHomeClass().getName().replace('.', File.separatorChar)
		            + ".java";
	}

	// Implementation Remote and Home classes
	if (beanDesc instanceof SessionDesc) {
	    // Session
	    srcICHome   = new GenICSessionHome((SessionDesc)beanDesc, wrpHomeFileName);
	    srcICRemote = new GenICSessionRemote((SessionDesc)beanDesc, wrpRemoteFileName);
	} else {
	    // Entity
	    srcICHandle = new GenICEntityHandle((EntityDesc)beanDesc, wrpHandleFileName);
	    srcICHome   = new GenICEntityHome((EntityDesc)beanDesc, wrpHomeFileName);
	    srcICRemote = new GenICEntityRemote((EntityDesc)beanDesc, wrpRemoteFileName);
	    if (beanDesc instanceof EntityJdbcCmpDesc) {
		// Entity Persistence Container Managed
		srcICBean   = new GenICEntityCMbyJDBCBean((EntityJdbcCmpDesc)beanDesc,
							  wrpBeanFileName);
	    }
	}
    }

    /**
     * Return the bean's name
     */
    public String getEjbName() {
	return(beanDD.getEjbName());
    }

    /**
     * Return the file name of the generated source for the Home
     */
    public String getWrpHomeFileName() {
	return(wrpHomeFileName);
    }

    /**
     * Return the file name of the generated source for the Remote
     */
    public String getWrpRemoteFileName() {
	return(wrpRemoteFileName);
    }

    /**
     * Return the file name of the generated source for the Entity Handle
     * (null if none)
     */
    public String getWrpHandleFileName() {
	return( srcICHandle==null ? null : wrpHandleFileName );
    }

    /**
     * Return the file name of the generated source for the derived class of the CMP Entity
     * (null if none)
     */
    public String getWrpBeanFileName() {
	return( srcICBean==null ? null : wrpBeanFileName );
    }

    /**
     * Return the 'pseudo' source file name of the Remote interface
     */
    public String getEjbRemoteFileName() {
	return ejbRemoteFileName;
    }

    /**
     * Return the 'pseudo' source file name of the Home interface
     */
    public String getEjbHomeFileName() {
	return ejbHomeFileName;
    }

    /**
     * Return the class name of the generated source for the Remote (package included)
     */
    public String getWrpRemoteClassName() {
	return(wrpRemoteName);
    }

    /**
     * Return the class name of the generated source for the Home (package included)
     */
    public String getWrpHomeClassName() {
	return(wrpHomeName);
    }

    /**
     * Generates the java sources
     * - that implements the enterprise Bean's remote interface,
     * - that implements the enterprise Bean's home interface,
     * - of the Handle class of the entity bean
     * - of the derived class of the CMP entity bean
     */
    public void generate(boolean withError) throws GenICException {
	genImplRemoteClass(withError);
	genImplHomeClass(withError);
	if (srcICHandle != null) {
	    genImplHandleClass(withError);
	}
	if (srcICBean != null) {
	    genDerivedBeanClass(withError);
	}
    }

    /**
     * Generates the java source that implements the enterprise Bean's remote interface,
     * (EJB Object class).
     * @param     withError to generate the java source with compilation errors
     *                      (Useful for test only!).
     * @exception GenICException In error case
     */
    void genImplRemoteClass(boolean withError) throws GenICException {
	srcICRemote.generate(withError);
    }

    /**
     * Generates the java source that implements the enterprise Bean's home interface,
     * (EJB Home class).
     * @param     withError to generate the java source with compilation errors
     *                      (Useful for test only!).
     * @exception GenICException In error case
     */
    void genImplHomeClass(boolean withError) throws GenICException {
	srcICHome.generate(withError);
    }

    /**
     * Generates the java source of the Handle class of the entity bean,
     * @param     withError to generate the java source with compilation errors
     *                      (Useful for test only!).
     * @exception GenICException In error case
     */
    void genImplHandleClass(boolean withError) throws GenICException {
	if (srcICHandle != null) {
	    srcICHandle.generate(withError);
	}
    }

    /**
     * Generates the java source of the derived class of the CMP entity bean,
     * @param     withError to generate the java source with compilation errors
     *                      (Useful for test only!).
     * @exception GenICException In error case
     */
    void genDerivedBeanClass(boolean withError) throws GenICException {
	if (srcICBean != null) {
	    srcICBean.generate(withError);
	}
    }

}
