/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.tools;


import java.lang.String;
import java.lang.reflect.Method;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatefulDesc;
import org.objectweb.jonas_ejb.lib.JavaType;


/**
 *  This class allows to generate the source of the class that implements
 *  the Enterprise bean's Home interface of a given Session Enterprise Java Bean.
 *  It extends GenICHome class.
 */

class GenICSessionHome extends GenICHome {

    boolean isStateful;

    /**
     * GenICSessionHome Constructor
     *
     * @param dd       deployment descriptor of the Session Enterprise Java Bean
     * @param fileName name of the source file of the class to generate
     * @exception GenICException in error case
     */
    GenICSessionHome(SessionDesc dd, String fileName) throws GenICException {
	super(dd, fileName);
	isStateful = (dd instanceof SessionStatefulDesc);
	isSynchronized = false;	// do not synchronize methods in Home
    }

    /**
     * Name of the superclass of the generated class
     */
    String getSuperClass() {
	return(new String("JSessionHome"));
    }

    /**
     * Source generation of the import clauses
     * @exception GenICException in error case
     */
    void genImport() throws GenICException {
	src.println("import java.lang.Error;");
	src.println("import java.lang.Object;");
	src.println("import java.lang.RuntimeException;");
	src.println("import java.rmi.RemoteException;");
	src.println("import javax.ejb.RemoveException;");
	src.println("import javax.ejb.SessionBean;");
	src.println("import org.objectweb.jonas_ejb.container.RequestCtx;");
	src.println("import org.objectweb.jonas_ejb.container.JContainerImpl;");
	src.println("import org.objectweb.jonas_ejb.container.JContextSession;");
	src.println("import org.objectweb.jonas_ejb.container.JSessionHome;");
	src.println("import org.objectweb.jonas_ejb.container.TraceEjb;");
	src.println("import org.objectweb.jonas_ejb.deployment.api.SessionDesc;");
	src.println("import "+wrpRemoteName+";");
    }

    /**
     * Source generation of the constructor method
     * @exception GenICException in error case
     */
    void genConstructor() throws GenICException {
	src.println("public " + wrpHomeBaseName +
		    "(SessionDesc desc, String BeanName, JContainerImpl cont) throws RemoteException {");
	src.indentPlus();
	src.println("super(desc, BeanName, cont);");
	src.println("TraceEjb.debugGenIC(\"" + wrpHomeBaseName + "(\"+BeanName+\")\");");
	src.indentMinus();
	src.println("}");
    }

    /**
     * Source generation for the given create() method described in the Home interface
     * @exception GenICException in error case
     */
    void genCreate(Method method) throws GenICException {

	Class paramsTable[] = method.getParameterTypes();

	genHeaderMethod(method, wrpHomeBaseName);

	src.println(wrpRemoteBaseName + " ejbobj = null;");
	src.println(ejbBeanName + " sb = null;");
	genSecuritySignature(method);
	src.println("RequestCtx rctx = preinvoke("+ddesc.getMethodDesc(method).getTxAttribute()+", methodSignature) ;");
	src.println("JContextSession ctx = null;");
	src.println("try {");
	src.indentPlus();
	src.println("ejbobj = ("+wrpRemoteBaseName+")reuseEJBObject();");
	src.println("if (ejbobj == null) {");
	src.indentPlus();
	src.println("ejbobj = new " + wrpRemoteBaseName + "(this);");
	src.println("newEJBObject(ejbobj);");
	src.indentMinus();
	src.println("}");
	if (isStateful) {
	    src.println("ctx = ejbobj.getNewSessionContext();");
	    src.println("rctx.setEJBContext(ctx);");
	    src.println("sb = ("+ejbBeanName +")ctx.getInstance();");
	    src.print("sb.ejbCreate(");
	    for (int p=1; p <= paramsTable.length; p++) {
		src.print("p" + p, false);
		if (p < paramsTable.length) {
		    src.print(", ", false);
		}
	    }
	    src.println(");", false);
	}
	genCatchBean(method, false);
	src.println("} finally {");
	src.indentPlus();
	src.println("postinvoke(rctx);");
	src.indentMinus();
	src.println("}");
	src.println("return(ejbobj);");
	src.indentMinus();
	src.println("}");
    }

    /**
     * Source generation for the remove methods
     * @exception GenICException in error case
     */
    void genRemove(Method method) throws GenICException {

	Class params[] = method.getParameterTypes();

	// remove(pk) is generic. 
	// Only remove(Handle) must be generated for SessionBeans
	boolean isHandle = JavaType.getName(params[0]).equals("javax.ejb.Handle");
	if (!isHandle) {
	    return;
	}

	genHeaderMethod(method, wrpHomeBaseName);
	src.println("javax.ejb.EJBObject ejbobj = null;");
	genSecuritySignature(method);
	src.println("RequestCtx rctx = preinvoke("+ddesc.getMethodDesc(method).getTxAttribute()+", methodSignature) ;");
	src.println("try {");
	src.indentPlus();
	src.println("ejbobj = p1.getEJBObject();");
	src.println("ejbobj.remove();");
	genCatchBean(method, true);
	src.println("} finally {");
	src.indentPlus();
	src.println("postinvoke(rctx);");
	src.indentMinus();
	src.println("}");
	src.indentMinus();
	src.println("}");
    }

    /**
     * Source generation for the given find method described in the Home interface
     * @exception GenICException in error case
     */
    void genFind(Method method) throws GenICException {
	// Nothing to do !!!
	// No find method in case of Session
    }

    /**
     * Source generation for callejbCreate method
     * Only stateless session beans use this callejbCreate().
     * Stateful beans may have several ejbCreate methods with args.
     * @exception GenICException in error case
     */
    void genCallejbCreate() {
	src.println("protected void callejbCreate(SessionBean sb) throws java.rmi.RemoteException {");
	if (!isStateful) {
	    src.indentPlus();
	    src.println(ejbBeanName + " b = ("+ejbBeanName+")sb;");
	    src.println("try {");
	    src.indentPlus();
	    src.println("b.ejbCreate();");
	    src.indentMinus();
	    src.println("} catch (Exception e) {");
	    src.indentPlus();
	    src.println("throw new RemoteException(\"Exception thrown in ejbCreate()\");");
	    src.indentMinus();
	    src.println("}");
	    src.indentMinus();
	}
	src.println("}");
    }

    /**
     * Source generation for createNewInstance method
     */
    void genCreateNewInstance() {
	src.println("protected SessionBean createNewInstance() throws Exception {");
	src.indentPlus();
	src.println("return (new " + wrpBeanName + "());");
	src.indentMinus();
	src.println("}");
    }
}

