/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 * 01/05/30 Christophe Ney - Lutris Technologies
 *          Fix to handle arguments containing white spaces.
 * 01/11/23 Guillaume Riviere
 *          Fix the addClassesInJar() method (on David, the Stub/Skel classes names are different).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */



package org.objectweb.jonas_ejb.tools;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.StringTokenizer;
import java.util.Vector;

import org.objectweb.common.Env;
import org.objectweb.common.Cmd;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.Version;


/**
 *  This class allows to generate:
 * <ul>
 * <li>
 *  the classes that implements the Enterprise bean's remote interface,
 * <li>
 *  the classes that implements the Enterprise bean's home interface,
 * <li>
 *  the classes that implements the Handles of the Entity beans,
 * <li>
 *  the classes that implements the persistence of the Entity beans with CMP,
 * </ul>
 * of all the Enterprise Java Beans defined in the given Deployment Descriptor.
 */

public class GenWholeIC {

    // Is the command is verbose ?
    private boolean verbose = false;
    
    // Generate with or without the  propagation of the security  context
    private boolean withSecPropag;

    // Generate the sources with errors (used for test only !)
    private boolean withError = false;

    // Name of the directory where to place the generated file
    private String directoryOutputName = null;

    // A GenIC for each bean
    private Vector vICSources = null;

    // Path of the generated files to delete
    private Vector fileNamesToDelete = null;


    /**
     * GenWholeIC allows to generate the classes that implement
     * the EJB Home and the EJB remote interfaces.
     * <p>
     *
     * Usage: java org.objectweb.jonas_ejb.tools.GenWholeIC -help
     * <br>
     *         to print this help message
     * <p>
     *
     * or    java org.objectweb.jonas_ejb.tools.GenWholeIC <Options> <Input_File>
     * <br>
     *         to generate the classes that implement
     * <ul>
     * <li>
     *           - the EJB home interface, and
     * <li>
     *           - the EJB remote interface,
     * </ul>
     *         for a given EJB.
     * <p>
     *           Options include:
     * <ul>
     * <li>
     *             -d <output_dir>  specify where to place the generated files
     * <li>
     *             -keepgenerated   do not delete intermediate generated source files
     * <li>
     *             -nocompil        do not compile the generated source files via
     *                              the java and rmi compilers    
     * <li>
     *             -noaddinjar      do not add the generated classes in the given
     *                              ejb-jar file
     * <li>
     *             -novalidation    parse the XML deployment descriptors without validation
     * <li>
     *             -javac     <opt> specify the name of the java compiler to use
     * <li>
     *             -javacopts <opt> specify the options to pass to the java compiler
     * <li>
     *             -rmicopts <opt>  specify the options to pass to the rmi compiler
     * <li>
     *             -verbose
     * <li>
     *             -secpropag       modify the RMI Skel. and Stub. to implement
     *                              the implicit propagation of the transactionnal
     *                              context and security context
     * </ul>
     * <p>
     *           Input_File         file name of the standard deployment descriptor
     *                              (.xml ended), or
     *                              file name of the EJB-jar (.jar ended).
     */

    public static void main(String args[]) {

	boolean  error = false;

	boolean  isHelp          = false;
	boolean  isVerbose       = false;
	boolean  isKeepGenerated = false;
	boolean  isSecPropag     = false;
	boolean  isCompil        = true;
	boolean  isAddInJar      = true;
	boolean  isWithError     = false;
	boolean  parseWithValidation = true;

	String   fileInputName = null;
	String   dirOutputName = null;
	String   nameJavaC     = new String("javac");
	Vector   optionsJavaC  = new Vector();
	Vector   optionsRmiC   = new Vector();

	// Get args
	for (int argn=0; argn<args.length; argn++) {
	    String arg = args[argn];
	    if (arg.equals("-help") || arg.equals("-?")) {
		isHelp = true;
		continue;
	    }
	    if (arg.equals("-verbose")) {
		isVerbose = true;
		continue;
	    }
	    if (arg.equals("-keepgenerated")) {
		isKeepGenerated = true;
		optionsRmiC.addElement(args[argn]);
		continue;
	    }
	    if (arg.equals("-secpropag")) {
		isSecPropag = true; 
		continue;
	    }
	    if (arg.equals("-nocompil")) {
		isCompil = false;
		isKeepGenerated = true;
		continue;
	    }
	    if (arg.equals("-noaddinjar")) {
		isAddInJar = false;
		continue;
	    }
	    if (arg.equals("-novalidation")) {
		parseWithValidation = false;
		continue;
	    }
	    if (arg.equals("-witherror")) {
		// hidden option for test only !!!
		// Generation of classes with errors to get compilation errors
		isWithError = true;
		continue;
	    }
	    if (arg.equals("-javac")) {
		argn++;
		if (argn<args.length) {
		    nameJavaC = args[argn];
		    if (Env.isJeremie() || Env.isDavid()) {
			optionsRmiC.addElement("-c");
			optionsRmiC.addElement(args[argn]);
		    }
		} else {
		    error = true;
		}
		continue;
	    }
	    if (arg.equals("-javacopts")) {
		argn++;
		if (argn<args.length) {
		    StringTokenizer st = new StringTokenizer(args[argn]);
		    while (st.hasMoreTokens()) {
			optionsJavaC.addElement(st.nextToken());
		    }
		} else {
		    error = true;
		}
		continue;
	    }
	    if (arg.equals("-rmicopts")) {
		argn++;
		if (argn<args.length) {
		    StringTokenizer st = new StringTokenizer(args[argn]);
		    while (st.hasMoreTokens()) {
			optionsRmiC.addElement(st.nextToken());
		    }
		} else {
		    error = true;
		}
		continue;
	    }
	    if (arg.equals("-d")) {
		argn++;
		if (argn<args.length) {
		    dirOutputName = args[argn];
		} else {
		    error = true;
		}
		continue;
	    }
	    fileInputName = args[argn];
	}

	// Usage ?
	if (isHelp) {
	    usage();
	    System.exit(0);
	}

	// Check args
	if (error || (fileInputName == null)) {
	    usage();
	    System.exit(2);
	}
	if (dirOutputName == null) {
	    dirOutputName = new String("");
	}

	// The dirOutputName must be initialize in case of an ejb-jar file
	// (The file names of the generated classes must contain the name of the package)
	if (fileInputName.endsWith(".jar")
	    && isAddInJar
	    && "".equals(dirOutputName)) {
	    dirOutputName = new String(".");
	}

	// Calculate the classpath for javac, rmic, JRMICompiler
	String classpath;
	if (fileInputName.endsWith(".jar")) {
	    classpath = fileInputName +
		System.getProperty("path.separator", "") +
		System.getProperty("java.class.path", "");
	} else {
	    classpath = System.getProperty("java.class.path", "");
	}

	// Set the parsing mode (with or without validation)
	if (!parseWithValidation) {
	    DeploymentDesc.setParsingWithValidation(false);
	}

	// Generates the classes for the set of the beans
	try {
	    DeploymentDesc ejbJarDD = null;
	    if (fileInputName.endsWith(".jar")) {
		// ejb-jar file
		URL url[] = new URL[1];
		url[0] = (new File(fileInputName)).toURL();
		URLClassLoader loaderForCls = new URLClassLoader(url);
		URLClassLoader loaderForDD = new URLClassLoader(url, null);
		// The loader used to get the DDs has no parent, to avoid to get wrong DDs
		// (ie META-INF/ejb-jar.xml resource accessible via the classpath)
		ejbJarDD = DeploymentDesc.getInstance(loaderForDD, loaderForCls);
	    } else {
		// xml file
		ejbJarDD = DeploymentDesc.getInstance(fileInputName,
						      BeanNaming.getJonasXmlName(fileInputName));
	    }

	    GenWholeIC gwc = new GenWholeIC(ejbJarDD, dirOutputName, isSecPropag, isVerbose, isWithError);
	    gwc.genSources();
	    if (isCompil) {
		gwc.compilClasses(nameJavaC, optionsJavaC, optionsRmiC, classpath);
		if (fileInputName.endsWith(".jar") && isAddInJar) {
		    gwc.addClassesInJar(fileInputName);
		}
		if (!isKeepGenerated) {
		    gwc.clean();
		}
	    }

	} catch (MalformedURLException e) {
	    error = true;
	    System.err.println("GenIC ERROR: Invalid ejb-jar file name "
			       + e.getMessage());
	} catch (GenICException e) {
	    error = true;
	    System.err.println("GenIC ERROR: " + e.getMessage());
	} catch (DeploymentDescException e) {
	    error = true;
	    System.err.println("GenIC ERROR: When reading the Deployment Descriptors for "+fileInputName);	    
	    System.err.println(e.getMessage());
	}
	if (error) {
	    System.exit(2);
	}
	// End
    }

    /**
     * GenWholeIC Constructor
     *
     * @param     ejbJarDescr    deployment descriptors of the beans
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @param     isSecPropag    generation with or without the security propagation
     *                           of the transactionnal context
     * @param     isVerbose      if true, some traces are print
     * @param     genWithError      to generate the java source with compilation errors
     *                           (Useful for test only!).
     * @exception GenICException In error case
     */
    public GenWholeIC(DeploymentDesc ejbJarDesc, String dirOutputName,
		 boolean isSecPropag, boolean isVerbose, boolean genWithError)
	throws GenICException {
	
	verbose = isVerbose;
	withSecPropag = isSecPropag;
	withError = genWithError;
	directoryOutputName = dirOutputName;
	fileNamesToDelete = new Vector();

	vICSources = new Vector();
	BeanDesc[] beansDD = ejbJarDesc.getBeanDesc();
	for (int i=0; i<beansDD.length; i++) {
	    // For debug only
	    //System.out.println(beansDD[i].toString());
	    if (!(beansDD[i] instanceof MessageDrivenDesc)) {
		// Nothing to generate in case of MessageDriven Bean
		vICSources.addElement(new GenIC(beansDD[i], directoryOutputName));
	    }
	}

	if (vICSources.size()!=0) {
	    System.out.print("GenIC for JOnAS "+Version.NUMBER+": Beans ");
	    for (int i=0; i < vICSources.size(); i++) {
		GenIC ics = (GenIC)vICSources.elementAt(i);
		System.out.print((i==0?"":", ") + "'" + ics.getEjbName() + "'");
	    }
	    System.out.println(" generation ...");
	}
    }

    /**
     *
     */
    public void genSources() throws GenICException {
	for (int i=0; i < vICSources.size(); i++) {
	    GenIC ics = (GenIC)vICSources.elementAt(i);
	    ics.generate(withError);
	    if (verbose) {
		System.out.println("GenIC: Sources implementation classes successfully generated"
				   + " for '" + ics.getEjbName() + "'");
		System.out.println("       in " 
				   + ics.getWrpHomeFileName()
				   + ", " + ics.getWrpRemoteFileName()
				   + (ics.getWrpHandleFileName()==null?"":", " + ics.getWrpHandleFileName())
				   + (ics.getWrpBeanFileName()==null?"":", " + ics.getWrpBeanFileName())
);
	    }
	}
    }

    /**
     * Compiles the java sources generated by the genSources() method.
     *
     * I.e. :
     * <ul>
     * <li>
     *   compile the classes via javac tool,
     * <li>
     *   creates the RMI stubs and skeletons via rmic tool,
     * <li>
     *   modify the RMI stubs and skeletons in case the propagation
     *   of the transactionnal context is implicite, and
     *   recompile the RMI skubs and skeletons via the javac tool.
     * </ul>
     *
     * @param     nameJavaC      name the javac tool to use
     * @param     optionsJavaC   options for the javac tool
     * @param     optionsRmiC    options for the rmic tool
     * @param     classpath      classpath value
     * @exception GenICException In error case
     */
    public void compilClasses(String nameJavaC,
			      Vector optionsJavaC,
			      Vector optionsRmiC,
			      String classpath)
	throws GenICException {
	
	String optDirOutput;
	String sRmic;
	Cmd cmd;

	if (vICSources.size()==0) {
	    return;
	}

	// Compile the generated sources
	cmd = new Cmd(nameJavaC);
	cmd.addArgument("-classpath");
	cmd.addArgument(classpath);
	cmd.addArgument("-d");
	cmd.addArgument((directoryOutputName.length()==0)?".":directoryOutputName);
	cmd.addArguments(optionsJavaC);
	for (int i=0; i < vICSources.size(); i++) {
	    GenIC ics = (GenIC)vICSources.elementAt(i);
	    cmd.addArgument(ics.getWrpHomeFileName());
	    fileNamesToDelete.addElement(ics.getWrpHomeFileName());
	    cmd.addArgument(ics.getWrpRemoteFileName());
	    fileNamesToDelete.addElement(ics.getWrpRemoteFileName());
	    if (ics.getWrpHandleFileName() != null) {
		cmd.addArgument(ics.getWrpHandleFileName());
		fileNamesToDelete.addElement(ics.getWrpHandleFileName());
	    }
	    if (ics.getWrpBeanFileName() != null) {
		cmd.addArgument(ics.getWrpBeanFileName());
		fileNamesToDelete.addElement(ics.getWrpBeanFileName());        
	    }
	}
	
	if (verbose) {
	    System.out.println("GenIC: Running '" + cmd.toString() + "'");
	}
	if (cmd.run()) {
	    if (verbose) {
		System.out.println("GenIC: Sources implementation classes successfully compiled via java compiler.");
	    }
	} else {
	    throw new GenICException("Failed when compiling the implementation classes via java compiler");
	}

	// Generate the stub and skeletons of the home and remote implementations
	if (Env.isSunRMI()) {
	    sRmic = "rmic";
	    cmd = new Cmd("rmic");
	    cmd.addArgument("-classpath");
	    cmd.addArgument(classpath);
	    cmd.addArgument("-v1.1");
	    cmd.addArgument("-keepgenerated");
	} else if (Env.isJeremie() || Env.isDavid()) {
	    String rmicClassName;
	    if (Env.isJeremie()) {
		rmicClassName = "org.objectweb.jeremie.tools.jrmic.JRMICompiler";
		sRmic = "jrmic";
	    } else {
		rmicClassName = "org.objectweb.david.tools.jrmic.JRMICompiler";
		sRmic = "drmic";
	    }
	    try {
		Class rmicClass = Class.forName(rmicClassName);
	    } catch (ClassNotFoundException e) {
		throw new
		    GenICException("Failed when compiling the implementation classes: "+
				   "Jeremie/David stub compiler not found");
	    }
	    cmd = new Cmd("java");
	    cmd.addArgument("-classpath");
	    cmd.addArgument(classpath);
	    cmd.addArgument(rmicClassName);
	    cmd.addArgument("-opt");
	    if (Env.isDavid()) {
		cmd.addArgument("-iiop");
	    }
	} else {
	    throw new GenICException("ORB unknown (neither RMI, nor JEREMIE, nor DAVID)");
	}
	
	if (!"".equals(directoryOutputName)) {
	    cmd.addArgument("-d");
	    cmd.addArgument(directoryOutputName);
	}
	cmd.addArguments(optionsRmiC);
	for (int i=0; i < vICSources.size(); i++) {
	    GenIC ics = (GenIC)vICSources.elementAt(i);
	    cmd.addArgument(ics.getWrpHomeClassName());
	    cmd.addArgument(ics.getWrpRemoteClassName());
	}
	if (verbose) {
	    System.out.println( "GenIC: Running '" + cmd.toString() + "'");
	}
	if (cmd.run()) {
	    if (verbose) {
		System.out.println("GenIC: Stubs and Skels successfully generated via "+sRmic+".");
	    }
	} else {
	    throw new GenICException("Failed when generating the Stubs and Skels via '"+sRmic+"'");
	}
	if (Env.isSunRMI()) {
	    // Patch the stubs and skels in case of Sun RMI.
	    Vector namesSkelStub = new Vector();
	    for (int i=0; i < vICSources.size(); i++) {
		GenIC ics = (GenIC)vICSources.elementAt(i);
		namesSkelStub.addElement(convertNameJavaToSkelSrc(ics.getWrpHomeFileName()));
		namesSkelStub.addElement(convertNameJavaToStubSrc(ics.getWrpHomeFileName()));
		namesSkelStub.addElement(convertNameJavaToSkelSrc(ics.getWrpRemoteFileName()));
		namesSkelStub.addElement(convertNameJavaToStubSrc(ics.getWrpRemoteFileName()));
	    }

	    for (int i=0; i<namesSkelStub.size(); i++) {
		if (verbose) {
		    System.out.println("GenIC: Modifying the '"
				       + namesSkelStub.elementAt(i) + "' RMI file");
		}
		modifySkelStubRMI((String)namesSkelStub.elementAt(i));
		fileNamesToDelete.addElement(namesSkelStub.elementAt(i));
	    }
	    
	    cmd = new Cmd(nameJavaC);  
	    cmd.addArgument("-classpath");
	    cmd.addArgument(classpath);      
	    if (!"".equals(directoryOutputName)) {
		cmd.addArgument("-d");
		cmd.addArgument(directoryOutputName);
	    }
	    cmd.addArguments(optionsJavaC);
	    cmd.addArgument("-nowarn");
	    
	    // Add -nowarn to take off the 'deprecated' messages due to rmic -v1.1
	    for (int i=0; i<namesSkelStub.size(); i++) {
		cmd.addArgument((String)namesSkelStub.elementAt(i));
	    }
	    if (verbose) {
		System.out.println("GenIC: Running '" + cmd.toString()+ "'");
	    }
	    if (cmd.run()) {
		if (verbose) {
		    System.out.println("GenIC: The 'patched' RMI Skel. and Stub. successfully compiled.");
		}
	    } else {
		throw new GenICException("Failed when compiling the 'patched' RMI Skel. and Stub.");
	    }
	}
    }

    /*
     * Modify the Skel or Stub RMI for the implicite propagation of contexts.
     * The modification is different depending on we need to propagate the security 
     * context or not.
     * The modifications to do are Operating System dependent.
     */
    void modifySkelStubRMI(String fileName) throws GenICException {

	int javaVersion = Env.getJavaVersion();

	String fileSaveName = new String(fileName + ".save");
	fileNamesToDelete.addElement(fileSaveName);
	boolean isSkel = fileName.endsWith("_Skel.java");
	// Rename the original skel or stub file
	try {
	    File f = new File(fileName);
            File saveFile = new File(fileSaveName);
            if (saveFile.exists()) {
                saveFile.delete();
            }
	    if (!(f.renameTo(saveFile))) {
		throw new GenICException("Cannot rename the file '"+fileName+"' to '"+fileSaveName+"'");
	    }
	} catch (Exception e) {
	    throw new GenICException("Failed when renaming a RMI Skel or Stub ("+e+")");
	}
	// Read and Modify the skel or stub file
	// The modifications of the RMI Skels and Stubs are differents
	// on the JDK 1.1.6 and on the others JDK !!!
	BufferedReader  fIn;
	BufferedWriter fOut;
	try {
	    fIn = new BufferedReader(new InputStreamReader(new FileInputStream(fileSaveName)));
	    fOut = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fileName)));
	} catch (Exception e) {
	    throw new GenICException("Failed when reading or writing a RMI Skel or Stub ("+e+")");
	}

	String skeletonName = new String("Skeleton");
	String stubName = new String("RemoteStub");
	skeletonName = withSecPropag ? "Security"+skeletonName : skeletonName;
	stubName = withSecPropag ? "Security"+stubName : stubName;
	try {
	    String line;
	    line = fIn.readLine();
	    while (line != null)  {
		if (isSkel) {
		    // Modify a Skel RMI file
		    line = (new JString(line)).replace
			("implements",
			 "extends org.objectweb.jonas.rmifilters."+skeletonName+" implements",
			 false);
		    line = (new JString(line)).replace
			("call.releaseInputStream()",
			 "this.inRequest(call)",
			 false);
		    line = (new JString(line)).replace
			("call.getResultStream(true)",
			 "this.outReply(call)",
			 false);
		} else {
		    // Modify a Stub RMI file
		    line = (new JString(line)).replace
			("java.rmi.server.RemoteStub",
			 "org.objectweb.jonas.rmifilters."+stubName);
		    line = (new JString(line)).replace
			("ref.invoke",
			 "this.invoke");
		}
		fOut.write(line, 0, line.length());
		fOut.newLine();
		line = fIn.readLine();
	    }
	    fIn.close();
	    fOut.close();
	} catch (Exception e) {
	    throw new GenICException("Failed when reading or writing a RMI Skel. or Stub. ("+
				     e+")");
	}
    }

    /**
     * Add the generated classes in the given ejb-jar file.
     */
    public void addClassesInJar(String jarFileName) throws GenICException {

	if (vICSources.size()==0) {
	    return;
	}

	boolean isOsWindows = Env.isOsWindows();

	Cmd cmd = null;
	Vector classesFileName = new Vector();
	for (int i=0; i < vICSources.size(); i++) {
	    GenIC ics = (GenIC)vICSources.elementAt(i);
	    classesFileName.addElement(convertNameJavaToClass(ics.getWrpHomeFileName()));
	    classesFileName.addElement(convertNameJavaToClass(ics.getWrpRemoteFileName()));
	    if (Env.isDavid()) {
		classesFileName.addElement(convertNameJavaToServerClass(ics.getWrpHomeFileName()));
		classesFileName.addElement(convertNameJavaToClientClass(ics.getEjbHomeFileName()));
		classesFileName.addElement(convertNameJavaToServerClass(ics.getWrpRemoteFileName()));
		classesFileName.addElement(convertNameJavaToClientClass(ics.getEjbRemoteFileName()));
	    } else {
		classesFileName.addElement(convertNameJavaToSkelClass(ics.getWrpHomeFileName()));
		classesFileName.addElement(convertNameJavaToStubClass(ics.getWrpHomeFileName()));
		classesFileName.addElement(convertNameJavaToSkelClass(ics.getWrpRemoteFileName()));
		classesFileName.addElement(convertNameJavaToStubClass(ics.getWrpRemoteFileName()));
	    }
	    if (ics.getWrpHandleFileName()!=null) {
		classesFileName.addElement(convertNameJavaToClass(ics.getWrpHandleFileName()));
	    }
	    if (ics.getWrpBeanFileName()!=null) {
		classesFileName.addElement(convertNameJavaToClass(ics.getWrpBeanFileName()));
	    }
	}
	// Construct and launch the jar command
	if (isOsWindows) {
	    /*
	     * Needed to walk around a Windows problem :-((
	     * (See http://developer.java.sun.com/developer/bugParade/bugs/4386865.html)
	     * If a file in the 'META-INF' directory have been read,
	     * the update of the jar file fails.
	     * The work around is to generated a script called in GenIC.bat
	     * (Warning: This generated script must not have a line too long
	     *           (limitated to 255 characters))
	     */
	    String batFileName = "GenICTmpJarU";
	    try {
		PrintWriter outBat =
		    new PrintWriter(new BufferedWriter(new FileWriter(batFileName)));
		outBat.println("-uf " + jarFileName);
		for (int i=0; i < classesFileName.size(); i++) {
		    String classFileName = (String)classesFileName.elementAt(i);
		    outBat.println("-C " + directoryOutputName + " "
				   + classFileName.substring(directoryOutputName.length()+1,
							     classFileName.length()));
		}
		outBat.close();
	    } catch (IOException ioe) {
		throw new GenICException("Failed when creating the temporary bat file '"
					 + batFileName+"' needed to update the ejb-jar-file (."
					 + ioe + ").");
	    }
	    if (verbose) {
		System.out.println("GenIC: "+batFileName+" file created'");
	    }
	} else {
	    cmd = new Cmd("jar");
	    cmd.addArgument("-uf");
	    cmd.addArgument(jarFileName);
	    for (int i=0; i < classesFileName.size(); i++) {
		cmd.addArgument("-C");
		cmd.addArgument(directoryOutputName);
		String classFileName = (String)classesFileName.elementAt(i);
		cmd.addArgument(classFileName.substring(directoryOutputName.length()+1,
							classFileName.length()));
		fileNamesToDelete.addElement(classFileName);
	    }
	    if (verbose) {
		System.out.println("GenIC: Running '" + cmd.toString() + "'");
	    }
	    boolean exitCmd = cmd.run();
	    // Analyse the result of the jar command
	    if (exitCmd) {
		if (verbose) {
		    System.out.println("GenIC: The generated classes are successfully added "
				       + "in the given ejb-jar file '"+jarFileName+"'.");
		}
	    } else {
		throw new GenICException("Failed when adding the generated classes "
					 + "in the given ejb-jar file '"+jarFileName+"'.");
	    }
	}
    }

    /**
     * Clean the intermediate generated files.
     */
    public void clean() {
	if (verbose) {
	    System.out.println("GenIC: Deleting " + fileNamesToDelete.toString());
	}
	for (int i=0; i < fileNamesToDelete.size(); i++) {
	    File f = new File((String)fileNamesToDelete.elementAt(i));
	    f.delete();
	}
    }

    static String convertNameJavaToSkelSrc(String jSrc) {
	return((new JString(jSrc)).replaceEndWith(".java", "_Skel.java"));
    }

    static String convertNameJavaToStubSrc(String jSrc) {
	return((new JString(jSrc)).replaceEndWith(".java", "_Stub.java"));
    }

    static String convertNameJavaToSkelClass(String jSrc) {
	return((new JString(jSrc)).replaceEndWith(".java", "_Skel.class"));
    }

    static String convertNameJavaToStubClass(String jSrc) {
	return((new JString(jSrc)).replaceEndWith(".java", "_Stub.class"));
    }

    static String convertNameJavaToClass(String jSrc) {
	return((new JString(jSrc)).replaceEndWith(".java", ".class"));
    }


    static String convertNameJavaToServerClass(String jSrc) {
	// Ex: convert a/b/Foo.java to a/b/_Foo_Tie.class
	String jSrcR = (new JString(jSrc)).replaceEndWith(".java", "");
	String result = jSrcR.substring(0,jSrcR.lastIndexOf(File.separatorChar)+1) +
	    "_" + jSrcR.substring(jSrcR.lastIndexOf(File.separatorChar)+1) + "_Tie.class";
	return result;
    }

    static String convertNameJavaToClientClass(String jSrc) { 
	// Ex: convert a/b/Foo.java to a/b/_Foo_Stub.class
	String jSrcR = (new JString(jSrc)).replaceEndWith(".java", "");
	String result = jSrcR.substring(0,jSrcR.lastIndexOf(File.separatorChar)+1) +
	    "_" + jSrcR.substring(jSrcR.lastIndexOf(File.separatorChar)+1) + "_Stub.class";
	return result;
    }

    private static void usage() {
	System.out.println("");
	System.out.println("Usage: java org.objectweb.jonas_ejb.tools.GenWholeIC -help");
	System.out.println("         to print this help message");
	System.out.println("");
	System.out.println("");
	System.out.println(" or    java org.objectweb.jonas_ejb.tools.GenWholeIC <Options> <Input_File>");
	System.out.println("");
	System.out.println("         to generate the container-specific classes for given EJB(s).");
	System.out.println("");
	System.out.println("           Options include:");
        System.out.println("             -d <output_dir>  specify where to place the generated files");
	System.out.println("             -keepgenerated   do not delete intermediate generated source files");
	System.out.println("             -noaddinjar      do not add the generated classes in the given");
	System.out.println("                              ejb-jar file");
	System.out.println("             -nocompil        do not compile the generated source files via");
	System.out.println("                              the java and rmi compilers");
	System.out.println("             -novalidation    parse the XML deployment descriptors without");
	System.out.println("                              validation");      
	System.out.println("             -javac     <opt> specify the java compiler to use");
	System.out.println("             -javacopts <opt> specify the options to pass to the java compiler");
	System.out.println("             -rmicopts  <opt> specify the options to pass to the rmi compiler");
	System.out.println("             -secpropag       modify the RMI Skel. and Stub. to implement");
	System.out.println("                              the implicit propagation of the security context");
	System.out.println("             -verbose");
	System.out.println("");
	System.out.println("           Input_File         standard deployment descriptor file's name or");
	System.out.println("                              ejb-jar file's name");
	System.out.println("");
    }
}



