/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): ____________________________________.
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */



package org.objectweb.jonas_ejb.tools;

import java.io.File;
import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.StringTokenizer;
import java.util.Vector;

import org.objectweb.common.Cmd;
import org.objectweb.common.Env;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.MessageDrivenDesc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.Version;


/**
 *  This class allows to generate:
 * <ul>
 * <li>
 *  the IDLs of the Home and Remote interfaces
 * </ul>
 * of all the Enterprise Java Beans defined in the given Deployment Descriptor.
 */

public class GenWholeIDL {

    // Is the command is verbose ?
    private boolean verbose = false;
    
    // Name of the directory where to place the generated file
    private String directoryOutputName = null;

    // A BeanDesc for each bean for which IDL have to be generated
    private Vector vBeanDD = null;


    /**
     * GenWholeIDL allows to generate the IDLs of the Home and Remote interfaces of the
     * beans described in the given EJB deployment descriptor.
     * <p>
     *
     * Usage: java org.objectweb.jonas_ejb.tools.GenWholeIDL -help
     * <br>
     *         to print this help message
     * <p>
     *
     * or    java org.objectweb.jonas_ejb.tools.GenWholeIDL <Options> <Input_File>
     * <br>
     *         to generate the IDLs of the Home and Remote interfaces of the
     *         beans described in the given EJB deployment descriptor.
     * <p>
     *           Options include:
     * <ul>
     * <li>
     *             -d <output_dir>  specify where to place the generated files
     * <li>
     *             -novalidation    parse the XML deployment descriptors without validation
     * <li>
     *             -rmicopts <opt>  specify the options to pass to the rmic compiler
     * <li>
     *             -verbose
     * <li>
     * </ul>
     * <p>
     *           Input_File         file name of the standard deployment descriptor
     *                              (.xml ended), or
     *                              file name of the EJB-jar (.jar ended).
     */

    public static void main(String args[]) {

	// First of all, check if the used JDK is at least an JDK 1.3
	// Indeed, 'rmic -idl' only available from JDK 1.3
	if (Env.getJavaVersion() < Env.JAVA_1_3) {
	    System.err.println("GenIDL ERROR: 'rmic -idl' not available on a JDK version less than 1.3.");
	    System.exit(2);
	}

	boolean  error = false;

	boolean  isHelp          = false;
	boolean  isVerbose       = false; 
	boolean  parseWithValidation = true;

	String   fileInputName = null;
	String   dirOutputName = null;
	Vector   optionsRmiC   = new Vector();

	// Get args
	for (int argn=0; argn<args.length; argn++) {
	    String arg = args[argn];
	    if (arg.equals("-help") || arg.equals("-?")) {
		isHelp = true;
		continue;
	    }
	    if (arg.equals("-novalidation")) {
		parseWithValidation = false;
		continue;
	    }
	    if (arg.equals("-verbose")) {
		isVerbose = true;
		continue;
	    }
	    if (arg.equals("-rmicopts")) {
		argn++;
		if (argn<args.length) {
		    StringTokenizer st = new StringTokenizer(args[argn]);
		    while (st.hasMoreTokens()) {
			optionsRmiC.addElement(st.nextToken());
		    }
		} else {
		    error = true;
		}
		continue;
	    }
	    if (arg.equals("-d")) {
		argn++;
		if (argn<args.length) {
		    dirOutputName = args[argn];
		} else {
		    error = true;
		}
		continue;
	    }
	    fileInputName = args[argn];
	}

	// Usage ?
	if (isHelp) {
	    usage();
	    System.exit(0);
	}

	// Check args
	if (error || (fileInputName == null)) {
	    usage();
	    System.exit(2);
	}
	if (dirOutputName == null) {
	    dirOutputName = new String("");
	}

	// Calculate the classpath for javac, rmic, JRMICompiler
	String classpath;
	if (fileInputName.endsWith(".jar")) {
	    classpath = fileInputName +
		System.getProperty("path.separator", "") +
		System.getProperty("java.class.path", "");
	} else {
	    classpath = System.getProperty("java.class.path", "");
	}
	optionsRmiC.addElement("-classpath");
	optionsRmiC.addElement(classpath);

	// Set the parsing mode (with or without validation)
	if (!parseWithValidation) {
	    DeploymentDesc.setParsingWithValidation(false);
	}

	// Generates the IDL for the set of the beans
	try {
	    DeploymentDesc ejbJarDD = null;
	    if (fileInputName.endsWith(".jar")) {
		// ejb-jar file
		URL url[] = new URL[1];
		url[0] = (new File(fileInputName)).toURL();
		URLClassLoader loaderForCls = new URLClassLoader(url);
		URLClassLoader loaderForDD = new URLClassLoader(url, null);
		// The loader used to get the DDs has no parent, to avoid to get wrong DDs
		// (ie META-INF/ejb-jar.xml resource accessible via the classpath)
		ejbJarDD = DeploymentDesc.getInstance(loaderForDD, loaderForCls);
	    } else {
		// xml file
		ejbJarDD = DeploymentDesc.getInstance(fileInputName,
						      BeanNaming.getJonasXmlName(fileInputName));
	    }

	    GenWholeIDL gwc = new GenWholeIDL(ejbJarDD, dirOutputName, isVerbose);
	    gwc.genIDL(optionsRmiC);

	} catch (MalformedURLException e) {
	    error = true;
	    System.err.println("GenIDL ERROR: Invalid ejb-jar file name "
			       + e.getMessage());
	} catch (GenICException e) {
	    error = true;
	    System.err.println("GenIDL ERROR: " + e.getMessage());
	} catch (DeploymentDescException e) {
	    error = true;
	    System.err.println("GenIDL ERROR: When reading the Deployment Descriptors for "+fileInputName);	    
	    System.err.println(e.getMessage());
	}
	if (error) {
	    System.exit(2);
	}
	// End
    }

    /**
     * GenWholeIDL Constructor
     *
     * @param     ejbJarDescr    deployment descriptors of the beans
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @param     isVerbose      if true, some traces are print
     * @exception GenIDLException In error case
     */
    public GenWholeIDL(DeploymentDesc ejbJarDesc, String dirOutputName, boolean isVerbose)
	throws GenICException {
	
	verbose = isVerbose;
	directoryOutputName = dirOutputName;

	vBeanDD = new Vector();
	BeanDesc[] beansDD = ejbJarDesc.getBeanDesc();
	for (int i=0; i<beansDD.length; i++) {
	    // For debug only
	    //System.out.println(beansDD[i].toString());
	    if (!(beansDD[i] instanceof MessageDrivenDesc)) {
		// Nothing to generate in case of MessageDriven Bean
		vBeanDD.addElement(beansDD[i]);
	    }
	}

	if (vBeanDD.size()!=0) {
	    System.out.print("GenIDL for JOnAS "+Version.NUMBER+": IDL generation for beans ");
	    for (int i=0; i < vBeanDD.size(); i++) {
		BeanDesc dd = (BeanDesc)vBeanDD.elementAt(i);
		System.out.print((i==0?"":", ") + "'" + dd.getEjbName() + "'");
	    }
	    System.out.println(" ...");
	}
    }

    /**
     * Generate the IDLs for all Beans.
     *
     * @param     optionsRmiC    options for the rmic tool
     * @exception GenICException In error case
     */
    public void genIDL(Vector optionsRmiC)
	throws GenICException {
	
	String optDirOutput;
	Cmd cmd;

	if (vBeanDD.size()==0) {
	    return;
	}

	// Generate the IDLs by the way of the rmic -idl command
	cmd = new Cmd("rmic");
	cmd.addArgument("-idl");
	cmd.addArgument("-always");
	if (directoryOutputName.length() != 0) {
	    cmd.addArgument("-d");
	    cmd.addArgument(directoryOutputName);
	}
	cmd.addArguments(optionsRmiC);
	for (int i=0; i < vBeanDD.size(); i++) {
	    BeanDesc dd = (BeanDesc)vBeanDD.elementAt(i);
	    cmd.addArgument(dd.getHomeClass().getName());
	    cmd.addArgument(dd.getRemoteClass().getName());
	}
	
	if (verbose) {
	    System.out.println("GenIDL: Running '" + cmd.toString() + "'");
	}
	if (cmd.run()) {
	    if (verbose) {
		System.out.println("GenIDL: The IDL of the Home and Remote interfaces of the beans"
				   +" are successfully generated via rmic -idl.");
	    }
	} else {
	    throw new GenICException("Failed when generating the IDL of the Home and Remote interfaces of the beans via rmic -idl.");
	}

    }

    private static void usage() {
	System.out.println("");
	System.out.println("Usage: java org.objectweb.jonas_ejb.tools.GenWholeIDL -help");
	System.out.println("         to print this help message");
	System.out.println("");
	System.out.println("");
	System.out.println(" or    java org.objectweb.jonas_ejb.tools.GenWholeIDL <Options> <Input_File>");
	System.out.println("");
	System.out.println("         to generate the IDLs for given EJB(s).");
	System.out.println("");
	System.out.println("           Options include:");
        System.out.println("             -d <output_dir>  specify where to place the generated files");
	System.out.println("             -novalidation    parse the XML deployment descriptors without");
	System.out.println("                              validation");      
	System.out.println("             -rmicopts  <opt> specify the options to pass to the rmi compiler");
	System.out.println("             -verbose");
	System.out.println("");
	System.out.println("           Input_File         standard deployment descriptor file's name or");
	System.out.println("                              ejb-jar file's name");
	System.out.println("");
    }
}

